/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

/** "Hello, world." program for HyperCast 2.0
 *
 * @author HyperCast Team 
 * @version Hello World (Callback Version) 1.0, Nov 2001
 */


import java.util.*;
import java.io.*;
import java.net.*;
import java.lang.*;
import hypercast.*;
import hypercast.events.*;

public class HelloWorld_NoCallBack_Notification 
{ 

/******************************************************
                     Constructor 
******************************************************/
public HelloWorld_NoCallBack_Notification()
{
} 


/******************************************************
   This is the main function:
        1. Creates an overlay socket
        2. Sends a "Hello World" to all members
        3. Receives and prints "Hello World"s received 
        from other members in the overlay 
******************************************************/
public static void main(String[] args)throws InterruptedIOException
{
/******************************************************
                    Data Members 
******************************************************/

     /* This is the string we will sent */ 
     String MyString = new String("Hello World");

     /* MySocket is the name of the overlay socket */
     I_OverlaySocket MySocket = null;

     /* The configuration object contains configuration  parameters */
     HyperCastConfig ConfObj = null;

	
	/* Create an overlay configuration object */
	ConfObj = HyperCastConfig.createConfig("hypercast.xml");
	//ConfObj = HyperCastConfig.createOLConfig("hypercast.prop");

	/* Create notification handler */
	NoCallBack_notification_handler cnh = new NoCallBack_notification_handler();

	/* Finally, the overlay socket is created with the configuration object 
		"this" is the object that implements the callback 
	*/
	MySocket=ConfObj.createOverlaySocket(null, cnh);

	/* Once the overlay socket is create, it can join the overlay */ 
	MySocket.joinOverlay();


	/* Now we wait for 3 seconds. This is enough 
		time to have the overlay stabilize 
	*/
	try {
        Thread.sleep(3000);
	} catch (InterruptedException e) {}

	/* wait for the overlay socket to be stable for sending and receiving data */
	//cnh.waitUntilStable();

	/* Print the logical address */ 
	System.out.println("This node's logical address is:" +MySocket.getLogicalAddress().toString());

	nReceiver messageReceiver = new nReceiver(MySocket);

	while(true){

		/* Convert the string to a byte array and create a message */ 

		byte[] MyData = MyString.getBytes(); 
		I_OverlayMessage MyMessage = MySocket.createMessage(MyData);


		/* The message is sent to all members of the overlay */
		MySocket.sendToAll(MyMessage);
		System.out.println("<HelloWorld> message sent to other members ...");

			 try {
				Thread.sleep(2000);
			 } catch (InterruptedException e) {}
	 }
}

};

class nReceiver implements Runnable
{
    I_OverlaySocket ms = null;
	
    I_OverlayMessage RecvdMessage;

    public nReceiver(I_OverlaySocket os)
    {
        ms = os;
		
        Thread rt = new Thread(this);
        rt.start();
    }

    final public void run ()
    {
        while (true)
        {
            /* get the received message */ 
            RecvdMessage = ms.receive(); 

            /* get the logical address of the sender */ 
            String SenderSrc = RecvdMessage.getSourceAddress().toString();

            /* Next we transfor the received message into a string */
            String sMsg=new String(RecvdMessage.getPayload(), 0, 
                                   RecvdMessage.getPayloadLength());

            /* ... and print the payload */ 
            //if(!RecvdMessage.getSrcAddress().equals(ms.getLogicalAddress()))
            //{
                System.out.println("Received <"+ sMsg + ">from node with logical address: " +SenderSrc +" . ");
            //}
        }
    }
}

class NoCallBack_notification_handler extends NotificationHandler 
{
	/** The object used for synchronization that makes application to wait
	 * for the overlay socket to be stable before sending and receiving data.
	 */
	 Object StableGuard;

	/* Flag indicating if overlay socket is stable */
	boolean isSTABLE = false;


	NoCallBack_notification_handler() {
		super();

		/* create synchronization object */
		StableGuard = new Object();

	}
	
	public void waitUntilStable() {
		synchronized (StableGuard)
		{
			while (!isSTABLE) {
				try 
				{
				   System.out.println("Wait for the overlay socket being stable.");
				   StableGuard.wait();
				}
				catch (InterruptedException ie) 
				{
				}
			}
		}
	}
	
	
	public void handler (NOTIFICATION_EVENT obj)
	{
		if (obj instanceof NODE_ISSTABLE) {
			NODE_ISSTABLE ne = (NODE_ISSTABLE) obj;
			System.out.println("event NODE_ISSTABLE happens!");

			synchronized (StableGuard) 
			{
				isSTABLE = true;
				StableGuard.notify();
			}
		}
		else if (obj instanceof NODE_LEAVEOVERLAY) {
			NODE_LEAVEOVERLAY ne = (NODE_LEAVEOVERLAY) obj;
			System.out.println("event NODE_LEAVEGROUP happens!");
		}
		else if (obj instanceof NODE_LOGICALADDRESSCHANGED) {
			NODE_LOGICALADDRESSCHANGED ne = (NODE_LOGICALADDRESSCHANGED) obj;
			System.out.println("event NODE_LOGICALADDRESSCHANGED happens!");
		}
		else if (obj instanceof NODE_NEIGHBORHOODCHANGED) {
			NODE_NEIGHBORHOODCHANGED ne = (NODE_NEIGHBORHOODCHANGED) obj;
			System.out.println("event NODE_NEIGHBORHOODCHANGED happens!");
		}
		else
		   throw new IllegalArgumentException("unknown event happens!");
		   
	}
}
