/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.DT;

import hypercast.*;

/** This class defines the messages used for determining the logical
 * coordinates based on measurement.
 * <p>
 *
 * @author HyperCast Team
 * @author Haiyong Wang
 * @version 2.0, May 15, 2002
 */


class GNP_Message implements I_Message{

    /** A message type */
    public static final byte InitProbe = 0;
    /** A message type */
    public static final byte Ping = 1;
    /** A message type */
    public static final byte Pong = 2;
    /** A message type */
    public static final byte EndProbe = 3;

    /** Message Type */
    private byte type;

    /**Source IP Address + Port# */
	I_PhysicalAddress srcAddr;

    /** Delay measurement in terms of mili-second */
    private float[]delays;
    /* note: the value is multiplied 100 times in the message to save the conversion
    between int and float */


    /** Construct a GNP message
     *
     * @param mtype The type of the GNP_Message
     * @param data This field is only used in EndProbe message, which
     * transmits the measurements back to the initiator
     */
	public GNP_Message( byte mtype, I_PhysicalAddress src, float[] data){
        type = mtype;
        srcAddr = src;
        if( EndProbe == mtype ) delays = data;
    }

	/** Return the type this GNP message. */
    public byte getType() { return type; }

	/** Return the physical address of the source of the message. */
	public I_PhysicalAddress getSrcAddress() { return srcAddr; }

	/** return the delay measurement results. */
    public float[] getDelay() { return delays; }

	/** Convert a GNP message into byte array. */
    public byte[] toByteArray(){
    	if (srcAddr == null) {
            throw new HyperCastFatalRuntimeException ("The source physical address of this GNP message is null. Exiting...");
    	}
    	
        byte [] results;

        if( EndProbe == type ) {
			results = new byte [1+srcAddr.getSize()+4*delays.length];
            for( int i=0; i<delays.length; i++)
				System.arraycopy( ByteArrayUtility.toByteArray((int)(delays[i]*100)), 0, results, 1+srcAddr.getSize()+4*i, 4 );
        }
        else {
			results = new byte [1+srcAddr.getSize()];
        }

        results[0] = type;
		System.arraycopy(srcAddr.toByteArray(), 0, results, 1, srcAddr.getSize());

        return results;
    }

	/** Reconstruct GNP message from byte array. */
	public static GNP_Message restoreMessage(byte[] receiveBuffer, int[] validBytesStart, int validBytesEnd, I_UnicastAdapter adapt){

		final int PAsize = adapt.getPhysicalAddressSize();

        //restore type
        byte mType;
        int bufPtr;

        bufPtr = validBytesStart[0];
        mType = receiveBuffer[bufPtr++];

        //restore physical address
		byte[] phyAddr = new byte[6];
		I_PhysicalAddress src;
		System.arraycopy(receiveBuffer, bufPtr, phyAddr, 0, PAsize);
		src = adapt.createPhysicalAddress(phyAddr, 0);
		bufPtr += PAsize;

        //restore the data if they are there
        float[] data=null;
        if( mType == EndProbe ) {
            data = new float[ (validBytesEnd-bufPtr)/4 ];
            for( int i=0; i<data.length; i++){
                data[i] = ((float) ByteArrayUtility.toInteger( receiveBuffer, bufPtr+4*i))/100;
            }
        }
        return new GNP_Message( mType, src, data );
    }
}
