/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/
package hypercast.HC;

import hypercast.HyperCastStatsException;
import hypercast.I_AddressPair;
import hypercast.I_LogicalAddress;
import hypercast.I_PhysicalAddress;
import hypercast.NeighborhoodStats;
import hypercast.StatsElement;
import hypercast.StatsProcessor;
import hypercast.SimpleStats;
import hypercast.util.XmlUtil;
import hypercast.I_Stats;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * The stats table for HC neighborhood table
 *
 * @author HyperCast Team
 * @author Guangyu Dong
 * @version 2005 (version 3.0)
 */
class HC_NeighborhoodStats extends NeighborhoodStats {
	
	/** address pair array*/
	HC_AddressPair[] neighbors;
	int index;
	
	/**
	 * Constructor.
	 */
	public HC_NeighborhoodStats(HC_AddressPair[] neighbors, int index)
	{
		super(null);
		this.neighbors=neighbors;
		this.index=index;
	}
	
	
	/**
	 * Initialize the statistics processor instance contained in this object.
	 */
	protected void InitStatisticsStructure()
	{
		super.InitStatisticsStructure();
		
		statsPro.addStatsElement("Age"/* name */, new AgeQuery()/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);
		
		
   }
	

/** 
 * Return the result of query for the statistics specified by the given xpath.
 * 
 * @param doc   Document used for create new elements or nodes.
 * @param xpath XPath instance represents the statistics to be queried.
 * @see I_Stats#getStats(Document, XPath)
 */
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException
	{
		if(neighbors!=null)
		  addressPair=neighbors[index];
		
		return statsPro.getStatsResult(doc, xpath);
	}

	
	/** 
	 * Set the statistics specified by the given xpath. The value actually set is
	 * returned.
	 * 
	 * @param doc   	Document used for create new elements or nodes.
	 * @param xpath 	XPath instance represents the statistics to be queried.
	 * @param newValue	Element representing the value or sub-tree to be set.
	 * @see I_Stats#setStats(Document, XPath, Element)
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	{
		if(neighbors!=null)
		  addressPair=neighbors[index];
		
		return statsPro.setStatsResult(doc, xpath, newValue);
	}

   /**
	* This class manages simple statistic "Age".
	*/
   class AgeQuery extends SimpleStats
   {
	   /**
		* Gets the value of a simple statistic managed by the implementation of this class.
		* 
		* @see SimpleStats#getStats()
		*/
	   protected String getStats()
	   {
			if(neighbors!=null)
			  addressPair=neighbors[index];
	   	  
			if(addressPair!=null)
			  return ""+((HC_AddressPair)addressPair).getAge();
			else 
			  return null;
	   }

	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
		 throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
   }
	
}