/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;


import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This class includes all the functions related to an Internet Address/Port pair.
 * The port is always valid (i.e. in the range 0 to (2^16)-1).
 * <p>
 *
 * @author HyperCast Team
 * @author Michael Nahas
 * @author Dongwen Wang
 * @version	2.0, Jan. 19, 2001
 */


public class IP_Address implements I_PhysicalAddress {

    // Minimum and maximum port numbers. 
    protected static final int MIN_PORT = 0;
    protected static final int MAX_PORT = 256*256-1;  // = 2^16 - 1

	/**
	 * The size of IP_Address type physical address when 
	 * carried in the packet.
	 */
	public static final int ADDRESS_SIZE = 6;

    /**
     * The IP address of the node. 
     */
    protected InetAddress address = null;

    // We chose to implement using an int (rather than a short) because the java.net Socket
    // classes use an int for the port number.
	/**
	 * The port number.
	 */
	protected int port;

	
  /** 
	* Constructs a new <code>IP_Address</code> object from an <code>IP_Address</code> object. 
	* 
	* @param	ipAddr	any object of type <code>IP_Address</code>
	*/
	public IP_Address(IP_Address ipAddr) {
		address = ipAddr.address;
		port = ipAddr.port;
        	// No port check needed.
	}


   /** 
	* Constructs the <code>IP_Address</code> object cooresponding to the 
	* <code>java.net.DatagramSocket</code>.
	* 
	* @param	udpSocket	any object of type <code>DatagramSocket</code>
	*/
	public IP_Address(DatagramSocket udpSocket) {
		// Unfortunately, we cannot do: address = udpSocket.getLocalAddress();    
        // because it always returns 0.0.0.0!!!
        try 
        {
			//Some Unix hosts return 127.0.0.1 as local IP addresses that 
			//cause problem in the communication between OL Sockets. In this case, 
			//a DNS inquire command is used to get local IP address.
            if (InetAddress.getLocalHost().getHostAddress().equalsIgnoreCase("127.0.0.1"))
            {
               String LocalIPStr = getLocalIPbyDNS();
               if (!(LocalIPStr == null))
               {
                  address = InetAddress.getByName(LocalIPStr);
               } else {
                   throw new HyperCastFatalRuntimeException ("UDP_UnicastAdapter: Could not get local IP by DNS query!");
               }
            } else {
               address = InetAddress.getLocalHost();
            }
        }
        catch (UnknownHostException uhe) 
        {
            System.err.println("InetAddress.getLocalHost() threw an UnknownHostException.");
            address = null;
        }
        
        port = udpSocket.getLocalPort();

        // No port check needed.
	}


	//added by wjp 01/29/2004.
	/** 
	 * Constructs the <code>IP_Address</code> object which has the specified
	 * port number.
	 * @param	portnum		a port number.
	 */
	 public IP_Address(int portnum) {
		 try 
		 {
			//Some Unix hosts return 127.0.0.1 as local IP addresses that 
			//cause problem in the communication between OL Sockets. In this case, 
			//a DNS inquire command is used to get local IP address.
			 if (InetAddress.getLocalHost().getHostAddress().equalsIgnoreCase("127.0.0.1"))
			 {
				String LocalIPStr = getLocalIPbyDNS();
				if (!(LocalIPStr == null))
				{
				   address = InetAddress.getByName(LocalIPStr);
				} else {
                                    throw new HyperCastFatalRuntimeException ("UDP_UnicastAdapter: Could not get local IP by DNS query!");
				}
			 } else {
				address = InetAddress.getLocalHost();
			 }
		 }
		 catch (UnknownHostException uhe) 
		 {
			 System.err.println("IP_Address: InetAddress.getLocalHost() threw an UnknownHostException.");
			 address = null;
		 }
        
		 port = portnum;
	 }


   /** 
	* Constructs a <code>IP_Address</code> object from an <code>InetAddress</code> instance and port number.
	* 
	* @param	a an instance of <code>InetAddress</code>
	* @param	p			a port number
	* @exception IllegalArgumentException when p is not in 0 to (2^16)-1
	*/
	public IP_Address(InetAddress a, int p) {
		//updated by wjp 05/11/2004
		port = p;

		//System.out.println("IP_Address: InetAddress a = " + getIPString(a) +".");
		if ((a != null) && (!getIPString(a).equals("0.0.0.0"))) {
			address = a;
			//System.out.println(" (1) IP_Address: InetAddress a = " + getIPString(address) + ",port=" + port);
		}
		else {
			try 
			{
			   //Some Unix hosts return 127.0.0.1 as local IP addresses that 
			   //cause problem in the communication between OL Sockets. In this case, 
			   //a DNS inquire command is used to get local IP address.
				if (InetAddress.getLocalHost().getHostAddress().equalsIgnoreCase("127.0.0.1"))
				{
				   String LocalIPStr = getLocalIPbyDNS();
				   if (!(LocalIPStr == null))
				   {
					  address = InetAddress.getByName(LocalIPStr);
				   } else {
                                       throw new HyperCastFatalRuntimeException ("UDP_UnicastAdapter: Could not get local IP by DNS query!");
				   }
				} else {
				   address = InetAddress.getLocalHost();
				}
			}
			catch (UnknownHostException uhe) 
			{
				System.err.println("IP_Address: InetAddress.getLocalHost() threw an UnknownHostException.");
				address = null;
			}
			//System.out.println(" (2) IP_Address: InetAddress a = " + getIPString(address) +",port=" + port);
		}

        if (port < MIN_PORT || port > MAX_PORT)
            throw new IllegalArgumentException("IP_Address - Port value outside of range.");
	}


	//added by wj p05/11/2004
	/**
	 * Get string representation of IP address in a InetAddress object.
	 * @param	addr	An InetAddress object. 
	 */
	String getIPString(InetAddress addr) {
		if (addr == null) return new String("None");
		
		String addrstr = addr.toString();
		int slash_loc = addrstr.indexOf('/');
		return addrstr.substring(slash_loc+1, addrstr.length());
	}


   
   /** 
	* Constructs a <code>IP_Address</code> object from a byte array address and a port number.
	* 
	* @param	byteAddress	an Internet address as a byte array 
	* @param	p		port number
	* @exception IllegalArgumentException when p is not in 0 to (2^16)-1
	*/
	public IP_Address(byte[] byteAddress, int p) {
        address = createInetAddress(byteAddress);
		port = p;

        if (port < MIN_PORT || port > MAX_PORT)
            throw new IllegalArgumentException("IP_Address - Port value outside of range.");
	}

	
	/**
	 *  Constructs an instance of InetAddress from a byte array.
	 */
	private static InetAddress createInetAddress(byte[] byteAddress) {
        String stringAddress = null;
        try 
        {
            stringAddress = (byteAddress[0] & 0xff) + "." +
                (byteAddress[1] & 0xff) + "." +
                (byteAddress[2] & 0xff) + "." +
                (byteAddress[3] & 0xff);
            return InetAddress.getByName(stringAddress);
        }
        catch (UnknownHostException uhe)
        {
            throw new HyperCastFatalRuntimeException ("IP_Address - Unknown Host Exception thrown by new InetAddress on \"" + stringAddress + "\"", uhe);
        }
    }


   /** 
	* Constructs a <code>IP_Address</code> object from an 6 byte array.
	* This uses 2 more bytes than necessary.  This is as Tyler Beam initially implemented it.
	* 
	* @param	byteAddressAndPort	an Internet address and port stored in a byte array 
	* @exception IllegalArgumentException when p is not in 0 to (2^16)-1
	*/
	public IP_Address(byte[] byteAddressAndPort) {
        if (byteAddressAndPort.length < 6)
            throw new IllegalArgumentException("IP_Address - length of the byte array is too short");

        address = createInetAddress(byteAddressAndPort);
		port = (ByteArrayUtility.toShort(byteAddressAndPort, 4) & 0xFFFF);

        if (port < MIN_PORT || port > MAX_PORT)
            throw new IllegalArgumentException("IP_Address - Port value outside of range.");
	}

	
	//Updated by wjp 01/29/2004
	/* In Hypercast2.0, this method throws UnknownHostException exception, any call 
	 * to this method needs to exist in a "try-catch" statement. It is not
	 * a problem in Hypercast2.0 because IP_Address is a final class or not extended 
	 * by other classes. In Hypercast3.0, the class IP_Address is extended by
	 * other classes. In the constructors of a extended class, the first statement is 
	 * super() which cannot be put into a "try-catch" structure. So the exception
	 * is captured in the constructor of IP_Address now.
	 */
   /** 
	* Constructs a <code>IP_Address</code> object from a string address and a port number.
	* 
	* @param	addrString	a string representing the IP address (eg. "128.143.71.50")
	* @param p     The port.
	* @exception IllegalArgumentException when p is not in 0 to (2^16)-1
	* @exception UnknownHostException thrown by <code>InetAddress.getByName</code>
	*/
	//public IP_Address(String addrString, int p) throws UnknownHostException {
	public IP_Address(String addrString, int p) {
		//updated by wjp 01/28/2004
        // can throw UnknownHostException
		try
		{
           address = InetAddress.getByName(addrString);
		}
		catch (UnknownHostException e)
		{
		   System.err.println("IP_Address: Could not create the local IP address.-" + e);
		   address = null;
		}
		
		port = p;

        if (port < MIN_PORT || port > MAX_PORT)
            throw new IllegalArgumentException("IP_Address - Port value outside of range.");
	}

	
	//Updated by wjp 01/29/2004
	/* 
	 * Reason see the explanantion in method IP_Address(String addrString, int p).
	 */
   /** 
	* Constructs a <code>IP_Address</code> object from a string containing the address and port
	* separated by a '/' or ':'.
	* 
	* @param	addrAndPortString	a string representing the IP address and port (eg. "128.143.71.50/4747")
	* @exception IllegalArgumentException when p is not in 0 to (2^16)-1
	* @exception UnknownHostException thrown by <code>InetAddress.getByName</code>
	*/
	//public IP_Address(String addrAndPortString) throws UnknownHostException {
	public IP_Address(String addrAndPortString) {
        int slash_location = addrAndPortString.indexOf('/');
        if (slash_location < 0)
            slash_location = addrAndPortString.indexOf(':');
		if (slash_location < 0)
			slash_location = -1;
 
		if (slash_location < 0)
		{
			//the string only contains a port number.
		  try 
		  {
			if (InetAddress.getLocalHost().getHostAddress().equalsIgnoreCase("127.0.0.1"))
			{
			   String LocalIPStr = getLocalIPbyDNS();
			   if (!(LocalIPStr == null))
			   {
				  address = InetAddress.getByName(LocalIPStr);
			   } else {
                               throw new HyperCastFatalRuntimeException ("UDP_UnicastAdapter: Could not get local IP by DNS query!");
			   }
			} else {
			   address = InetAddress.getLocalHost();
			}
		  }
		  catch (UnknownHostException uhe) 
		  {
			System.err.println("IP_Address: InetAddress.getLocalHost() threw an UnknownHostException.");
			address = null;
		  }
		}
		else
		{
          String addrString = addrAndPortString.substring(0, slash_location);

		  //updated by wjp 01/28/2004
          // can throw UnknownHostException
          //address = InetAddress.getByName(addrString);
		  try
		  {
		     address = InetAddress.getByName(addrString);
		  }
		  catch (UnknownHostException e)
		  {
			 System.err.println("IP_Address: " + addrAndPortString + " is a bad address string.**" + e);
			 address = null;
		  }
		}

        String portString = addrAndPortString.substring(slash_location+1, addrAndPortString.length());
        try 
        {
            port = Integer.parseInt(portString);
        }
        catch (NumberFormatException e) 
        {
            throw new IllegalArgumentException("IP_Address - Unable to parse  \"" + portString + "\" as a port number.");
        }
        
        if (port < MIN_PORT || port > MAX_PORT)
            throw new IllegalArgumentException("IP_Address - Port value outside of range.");
	}


   /** 
	* Gets IP address of this object.
	*/
	public InetAddress getInetAddress() {
		return address;
	}
	

   /** 
	* Gets byte address of this object.
	* @return	A byte[] as returned by <code>InetAddress.getAddress</code>.
	*/
	public byte[] getByteAddress() {
		return address.getAddress();
	}

	
   /** 
	* Gets port number of this IP_Address object.
	*/	
	public int getPort() {
		return port;
	}

   //added by wjp 11/13/2003
   /** 
	* Sets port number of this IP_Address object.
	*/	
	public void setPort(int Port) {
		port = Port;
	}


	/**

	 * Returns the byte array size

	 */
	public int getSize() {
        return IP_Address.ADDRESS_SIZE;
    }
	
   /** 
	* Converts the Address/Port pair to a 6 bytes array. The format is
	 * |-------4------|----2----|
	 * |      IP      |node port|
	 * |              |         |
	 * |--------------|---------|
	* @return	a 6 byte array where the first 4 bytes holds the address, 
	* the next 2 holds the port, the following 1 holds index and the last 
	* 1 holds node protocol and application protocol
	*/	
	public byte[] toByteArray() {
		byte[] byteAddressAndPort = new byte[6];
		
		byte[] byteAddress = address.getAddress();
		byte[] bytePort = ByteArrayUtility.toByteArray((short) port);
			
		System.arraycopy(byteAddress, 0, byteAddressAndPort, 0, 4);
		System.arraycopy(bytePort, 0, byteAddressAndPort, 4, 2);

		return byteAddressAndPort;
	}

	
   /** 
	* Test if two IP addresses are identical.
	* It is important that the argument is of type <code>Object</code> so that 
	* it can be used in a <code>Hashtable</code>.
	* 
	* @param	obj an object (should be of type <code>IP_Address</code>.)
	* @exception IllegalArgumentException if <code>obj</code> is not of type <code>IP_Address</code>.
	*/
    public boolean equals(Object obj) {
        if (obj == null)
        {
            return false;
        }
        else if (!(obj instanceof IP_Address))  
        {
            throw new IllegalArgumentException("The passed object is not of type IP_Address!");
        }
		else 
        {
            IP_Address ipAddr = (IP_Address) obj;

            //Using the method InetAddress.equals does not seem to work
            byte[] remoteByteAddress = ipAddr.getByteAddress();
            byte[] localByteAddress = address.getAddress();

            return ((remoteByteAddress[0]==localByteAddress[0]) && (remoteByteAddress[1]==localByteAddress[1]) && 
                    (remoteByteAddress[2]==localByteAddress[2]) && (remoteByteAddress[3]==localByteAddress[3]) && 
                    (port == ipAddr.getPort()));

        }
	}

	
	/** 
	 * Test if two IP addresses are identical.
	 * It is important that the argument is of type <code>Object</code> so that 
	 * it can be used in a <code>Hashtable</code>.
	 * 
	 * @param	obj an object (should be of type <code>IP_Address</code>.)
	 * @exception IllegalArgumentException if <code>obj</code> is not of type <code>IP_Address</code>.
	 */
	public boolean addressCheck(Object obj) {
	   return this.equals(obj);
	}
	
	
	
   /** 
	* Compares two IP address objects. 
	* 
	* @param	addr	an object of type <code>I_Address</code>
	* @return	true if (this > addr), otherwise false 
	* @exception IllegalArgumentException if <code>obj</code> is not of type <code>IP_Address</code>.
	*/
	public boolean greaterThan(I_Address addr) {
        if (addr == null || !(addr instanceof IP_Address)) 
            throw new IllegalArgumentException("The I_Address instance passed to IP_Address.greaterThan is not of type IP_Address!");
        
		IP_Address ipAddr = (IP_Address) addr;

		byte[] localByteAddress = address.getAddress();
		byte[] remoteByteAddress = ipAddr.getByteAddress();

		for (int i = 0; i < 4; i++) 
        {
			if (localByteAddress[i] == remoteByteAddress[i]) 
                continue;
			else
				return ((localByteAddress[i]&0xff) > (remoteByteAddress[i]&0xff));
        }

		return (port > ipAddr.getPort());
	}
	
	
   /** 
    * Converts the IP address object to string format "xxx.xxx.xxx.xxx/port" (e.g. "128.143.71.50/4747")
	*/
	public String toString() {
		byte[] byteAddress = address.getAddress();
		return ((byteAddress[0]&0xff) + "." + (byteAddress[1]&0xff) +"."+
			(byteAddress[2]&0xff) + "." + (byteAddress[3]&0xff) + "/" + port);
	}
	
	
	/**
	 * Returns a hashcode of this object
	 * It is important that the argument is of type <code>Object</code> so that 
	 * it can be used in a <code>Hashtable</code>.
	 */
	public int hashCode() {
		byte[] byteAddress = address.getAddress();

        int i = port;
		i ^= (byteAddress[0] << 0);
		i ^= (byteAddress[1] << 8);
		i ^= (byteAddress[2] << 16);
		i ^= (byteAddress[3] << 24);
		return i;
    }
	

        /**
         * Obtain local IP address by a DNS query and return local IP address in String form.
         */
     public String getLocalIPbyDNS() {
        try
        {
           String LocalIP = InetAddress.getLocalHost().getHostAddress();
           String LocalHostname = InetAddress.getLocalHost().getHostName();
              System.out.println("The local IP address is tested : " + LocalIP);


                boolean bDebug=false;
                String sOs=System.getProperty("os.name");

                Process p=null;
               	Process q=null; //for ifconfig command

                BufferedReader r=null;
                String sLine;
                String Local_IP_DNS = null;

                if (sOs.equalsIgnoreCase("Linux") || sOs.equalsIgnoreCase("SunOS")) {
                        System.out.println("The local Host Name is tested : " + LocalHostname);

                        try {
                                p=Runtime.getRuntime().exec(new String[] {"host", LocalHostname});
                        } catch (IOException e) {
                                System.out.println(e);
                        }
                        r=new BufferedReader(new InputStreamReader(new BufferedInputStream(p.getInputStream())));
                        try {
                                int sig = 0;
                                while ((sLine=r.readLine())!=null) {
                                    if (sLine.trim().startsWith(LocalHostname))
                                    {
                                        System.out.println("The result is : " + sLine);
                                        System.out.println("The local HostName is : " + LocalHostname);
                                        Local_IP_DNS = sLine.substring(sLine.indexOf("address")+7).trim(); 
                                        System.out.println("The result DNS IP is : *****" + Local_IP_DNS + "****");
                                        return Local_IP_DNS;
                                    } else if (sLine.trim().startsWith("Name"))
                                    {   
                                        sig++;
                                    } else if ((sig == 1) && (sLine.trim().startsWith("Address")))
                                    {   
                                        sig++;
                                        Local_IP_DNS = sLine.substring(sLine.indexOf("Address")+8).trim(); 
                                        System.out.println("The result DNS IP is : *****" + Local_IP_DNS + "****");
                                        return Local_IP_DNS;
                                    }
                                }
                        } catch (IOException e) {
                                System.out.println(e);
                        }


                        System.out.println("The DNS query is failed");



                        try {
                                q=Runtime.getRuntime().exec(new String[] {"/sbin/ifconfig"});
                        } catch (IOException e) {
                                System.out.println(e);
                        }
                        r=new BufferedReader(new InputStreamReader(new BufferedInputStream(q.getInputStream())));
                        try {
                                System.out.println("Start ifconfig try");
                                while ((sLine=r.readLine())!=null) {
                                    if (sLine.trim().startsWith("inet"))
                                    {
                                        //System.out.println("The result is : " + sLine);
                                        //System.out.println("The local HostName is : " + LocalHostname);
                                        int start_ind = sLine.trim().indexOf("inet");
                                        int end_ind = sLine.trim().indexOf("Bcast");
                                        Local_IP_DNS = sLine.trim().substring(start_ind+10, end_ind-1).trim();
                                      //System.out.println("start ind:"+start_ind+"end ind is:"+end_ind+"result is:"+Local_IP_DN S);
                                        System.out.println("The result DNS IP is : *****" + Local_IP_DNS + "****");
                                        return Local_IP_DNS;
                                    }
                                }
                        } catch (IOException e) {
                                System.out.println(e);
                        }
                        System.out.println("The ifconfig command is failed");


                }
        
        } catch (UnknownHostException ee) { 
           System.err.println("Unknown Host :" + ee);
        }
        
        return null;
     }

}
