/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import hypercast.events.NOTIFICATION_EVENT;
import hypercast.util.XmlUtil;

/**
 * This class is a finite state machine which eliminate the duplicate
 * message forwarding. This service applies to messages with
 * multicast, unicast and flood delivery mode.
 * 
 * @author HyperCast Team
 * @author Guimin Zhang
 * @version 2.0, Jul 10,2004
 */
public class MessageStoreFSM_DuplicateElimination extends I_MessageStoreFSM
{
    /** OL_Socket object*/
    private OL_Socket socket;
    /** MessageStore*/
    private MessageStore MStore;
    /** logical address*/
    private I_LogicalAddress mylogicaladdress;
    /**extension of ADF message */
    FSM_Extension e = null;	
    /** The state information of the message */
    private byte MyState;
    /** The state of the message,Init */
    public static final byte Init = 0;
    /** The state of the message,have message ID */
    public static final byte HaveMsgID = 1;
    /** The timestamp id, delete message */
    public static final byte DELETE = 0;
    /** The timeout const,delete(in millisecond) */
    private static long delete ;
    
    private static FiniteStateMachineMetaOperations metaOps;
    protected static FiniteStateMachineMetaOperations registerMetaOperations ()
    {
        if (metaOps == null)
        {
            metaOps = new DuplicateEliminationMetaOperations();
        }
        return metaOps;
    }

    /**Construct a MessageStoreFSM_DuplicateElimination object
     * with OL_Socket and MessageStore
     * @param  Socket			OL_Socket
     * @param  mstore			MessageStore
     * returns MessageStoreFSM_DuplicateElimination
     */
    public MessageStoreFSM_DuplicateElimination(OL_Socket Socket,MessageStore mstore, HyperCastConfig config)
    {
        socket = Socket;
        MStore = mstore;
        MyState = Init;
        mylogicaladdress = socket.getLogicalAddress();

		delete = config.getPositiveLongAttribute
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/DuplicateElimination/Delete"));
    }

    /**
     * The purpose of this finite state machine is to eliminate
     * duplicate messages. If the forwarding engine is allowed to
     * forward messages then duplicates cannot be eliminated so this
     * finite state machine handles all message forwarding.
     */
    protected static boolean messageStoreWillForward ()
    {
        return true;
    }

    /** check the change of the neighborhood
     */	
    public void changingNeighbor()
    {
        //nothing 
    }  
	
    /**
     * @return the finite state machine id
     */
    public short getFSMID()
    {
        return MessageStore.FSM_DUPLICATE_ELIMINATION;
    }

    /// TODO!!!!!!!!!!!!!! THIS IS DUPLICATED FROM OL_SOCKET!!!!!!!

    /**process new ADF_Data message
     * @param	msg			OL_Message
     */
    public void newADMsg(OL_Message msg)
    {
        /*
         * extract the formated message extension 
         */
        e = (FSM_Extension) msg.getFirstExtensionByType(Extension.FSM);
        if(e==null) throw new HyperCastWarningRuntimeException("Message format is not supported ");
        MyState = HaveMsgID;
        MStore.setTimer(this,DELETE,delete);
        /*
         * From this point, act just as forwarding engine
         */
        byte deliveryMode = msg.getDeliveryMode();
        /*
         * The hop limit field is interpreted to mean the maximum number of
         * hops that a message is permitted to make in the overlay before it is
         * dropped. This can also be described as the maximum number of times
         * that a message is permitted to be forwarded. It is useful to think
         * in terms of forwarding rather than dropping because there is no code
         * that explicitly drops a message due to the value of its hop limit
         * field, rather the hop limit field is used to decide if the
         * message should be forwarded.
         *
         * The hop limit field of a message is decremented here because when
         * a message arrives from the adapter the message is considered to have
         * completed one hop.
         *
         * A check is done to ensure that the contents of the hop limit
         * field are valid.
         */
        short hopLimit = 0;
        try
        {
            hopLimit = msg.decrementHopLimit();
        }
        catch (IllegalArgumentException iae)
        {
            String warningMessage =
                "Dropping message due to invalid hop limit: " +
                "hopLimit = " + (msg.getHopLimit() + 1) + ", " +
                "src = " + msg.getSourceAddress().toString();
            throw new HyperCastWarningRuntimeException (warningMessage, iae);
        }

        /*
         * After the hop limit field has been decremented a
         * decision can be made about whether or not it is possible to
         * forward this message. Note that simply because this node
         * *can* forward this message based on the hop limit field
         * does not mean that the message *will* be forwarded,
         * e.g. perhaps the message is destined for this node so it is
         * delivered locally, the message store will assume the
         * responsibility of forwarding, or the interception callback
         * will cancel forwarding.
         */
        boolean isPossibleToForwardThisMessage = hopLimit != 0;
		
        switch (deliveryMode)
        {
            /*
             * A multicast message continues to propagate through the overlay
             * by being forwarded to children in a multicast tree defined by
             * the overlay topology. The message is forwarded to the current application.
             */
            case OL_Message.DELIVERY_MODE_MULTICAST:
            {
                if (isPossibleToForwardThisMessage) socket.forwardToChildren (msg);               
                socket.appmsgArrived(msg,socket.callback);

            }
            break;

            /*
             * A unicast message continues to be routed through the
             * network until one of two things occurs:
             *
             * (1) Its destination logical address matches the logical
             * address of the current node, in which case the message
             * is received by the local node.
             *
             * (2) It has exceeded its maximum hop count, or
             * "hop limit" value, before reaching its
             * destination. This means that it can no longer be
             * forwarded to other nodes in an effort to reach its
             * destination - the message must be dropped.
             *
             * Thus, if it has not reached its destination and it has
             * not exceeded its maximum hop count, it will be
             * forwarded to another node.
             */
            case OL_Message.DELIVERY_MODE_UNICAST:
            {
                I_LogicalAddress destinationLogicalAddress = msg.getDestinationAddress();
                boolean messageHasReachedDestination = mylogicaladdress.equals (destinationLogicalAddress);
                /*
                 * Case where the message has NOT reached its
                 * destination, but can be forwarded to another node.
                 */
                if ((! messageHasReachedDestination)&&isPossibleToForwardThisMessage)
                {
                    socket.forwardToParent (msg);
                }
                /*
                 * Case where the message has reached its destination
                 */
                else if (messageHasReachedDestination)
                {
                    socket.appmsgArrived(msg,socket.callback);;
                }
                /*
                 * Case where the message has NOT reached its
                 * destination and it CANNOT be forwarded to another
                 * node - this means that the message is dropped
                 * because its maximum hop count has been exceeded.
                 */
                else
                {
                    // message is dropped silently.
                }
            }
            break;

            /*
             * A message that is in the overlay with flood mode delivery has no
             * destination or route, it is simply forwarded to all neighboring
             * nodes (save the node that sent it) at each hop until it reaches
             * its maximum hop count. In a flood, messages are received
             * unconditionally at each hop.
             *
             * NB: If the message store is active, it is responsible for
             * continuing to forward messages in the flood. However, stream
             * store will never forward in a flood.
             */
            case OL_Message.DELIVERY_MODE_FLOOD:
            {            
                if (isPossibleToForwardThisMessage)	socket.forwardFlood (msg);
                socket.appmsgArrived(msg,socket.callback);;
            }
            break;

            case OL_Message.DELIVERY_MODE_INVALID:
                throw new HyperCastFatalRuntimeException ("Previous check for unknown delivery mode failed");

            default:
                throw new HyperCastWarningRuntimeException ("Message with undefined delivery mode detected: " + deliveryMode);
        }

    }	

    /**process new ADF_Control message
     * @param	msg			OL_Message
     */
    public void newACMsg(OL_Message msg)
    {
        //nothing
    }	

    /**Update new ADF_Data message
     * @param	msg			OL_Message
     */		
    public void updateADMsg(OL_Message msg)
    {
        switch(MyState){
            case Init:
                /*
                 * for a new message, forward as the forwarding engine
                 */
                newADMsg(msg);
                break;
            case HaveMsgID:
                /*
                 * for an message with known message id, no action is taken
                 * to eliminate the duplicate forwarding
                 */
                break;
			
        }
    }

    /**process new ADF_Control message
     * @param	msg			OL_Message
     */
    public void updateACMsg(OL_Message msg)
    {
        //no body needed
    }
	
    /** process the timer expired events
     */
    public void timerExpired(int timerIndex)
    {
        /*
         * the message id is deleted at the configured time 
         */
        if(timerIndex==DELETE && MyState==HaveMsgID){
            MStore.remove(e.getUniqueID());
        }
    }
}

class DuplicateEliminationMetaOperations implements FiniteStateMachineMetaOperations
{
    public boolean messageStoreWillForwardMessage (OL_Message message) { return true; }
    public boolean isValidMessage (OL_Message message) { return true; }
	public boolean processIntermediateUnicastMessage(OL_Message message){return true;}
    public NOTIFICATION_EVENT notifyOnSend (OL_Message message) { return null; }
}
