/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import hypercast.events.NOTIFICATION_EVENT;
import hypercast.events.E2EACK_RECEIVED;
import hypercast.events.E2EPARTIALACK_RECEIVED;
import hypercast.util.XmlUtil;

import java.util.*;

/**
 * This class is a finite state machine which handles end to end ack.
 *
 * @author HyperCast Team
 * @author Guimin Zhang
 * @version 2.0, May. 23, 2001
 */
public class MessageStoreFSM_E2EACK extends I_MessageStoreFSM
{
	/** debug flag */
	boolean debug = false;
	
		/** OL_Socket object*/
	private OL_Socket socket;
	
		/** MessageStore*/
	private MessageStore MStore;
	
	/** The message handled by this FSM */
	private OL_Message message = null;
	
		/** logical address*/
	private I_LogicalAddress mylogicaladdress;
		/** root logical address*/
	private I_LogicalAddress root=null;
		/** logical address of the source of message*/
	private I_LogicalAddress sourceaddress = null;

	/** delivery mode*/
	private byte deliveryMode=OL_Message.DELIVERY_MODE_MULTICAST;
		/** Children information */
	private Vector ACKExpected;
	/** ACK Received*/
	private Vector ACKReceived;
		/**ADF_Payload for E2E ACK*/
	public static final byte PAYLOAD					= -1;
    public static final byte PAYLOAD_RETRAN				= -2;
    	
    		/**ADF_Control Types for E2E ACK*/
	public static final byte E2E_NACK					= 5;
	public static final byte FULL_E2E_ACK				= 6;
	public static final byte FINAL_PARTIAL_E2E_ACK		= 7;
	public static final byte E2E_REQUEST				= 8;
	public static final byte E2E_RESET					= 9;
	
	

	/** the message id */
	private byte[] messageid = null;	
	
		/** ACK sent*/
	private boolean  ACKSent;
		/** partial done*/
	private boolean  PartialDone;
		/** number of ack request sent*/
	private int NumACKRequestsSent;
		/** The state information of the message */
	private byte MyState;
	
		/** the length of Children*/
	private int length;
	
		/** Max Nack number */
	static final byte MaxACKRequest = 20;
	
		/** The state of the message,Init */
	public static final byte Init = 0;
		/** The state of the message,Don't have payload */
	public static final byte NoPayLoad = 1;
		/** The state of the message,Wait for ACK*/
	public static final byte WaitforACK = 2;
		/** The state of the message,Done */
	public static final byte Done = 3;
		/** The state of the message,Reset */
	public static final byte Reset = 4;
		
		/** the ACK timeout id */
	private static final int ACK_TIMER_INDEX = 0;
	/** the max ACK timeout id */	
	private static final int MAXACK_TIMER_INDEX = 1;
	/**the NACK timeout id */
	private static final int NACK_TIMER_INDEX = 2;
	/**the max NACK timeout id */
	private static final int MAXNACK_TIMER_INDEX = 3;
	/**the RESET timeout id */
	private static final int RESET_TIMER_INDEX = 4;
	/**the max RESET timeout id */
	private static final int MAXRESET_TIMER_INDEX = 5;
	/**the delete timeout id */
	private static final int DELETE_TIMER_INDEX = 6;
	/**the final delete timeout id */
	private static final int FINALDELETE_TIMER_INDEX = 7;
	/**the routing change timeout id */
	private static final int ROUTINGCHANGE_TIMER_INDEX = 8;

	
	
		/** The timeout const,ACK(in millisecond) */
	private static long timeout_ACK = -1;
		/** The timeout const,max ACK(in millisecond) */
	private static long max_ACK;
		/** The timeout const,ACK(in millisecond) */
	private static long timeout_NACK;
		/** The timeout const,max NACK(in millisecond) */
	private static long max_NACK;	
		/** The timeout const,RESET(in millisecond) */
	private static long timeout_RESET;
		/** The timeout const,max RESET(in millisecond) */
	private static long max_RESET;
		/** The timeout const,delete(in millisecond) */
	private static long delete;
	private static long final_Delete;
		/** The timeout const,routing change(in millisecond) */
	private static long routing_CHANGE;

    private static FiniteStateMachineMetaOperations metaOps;
    protected static FiniteStateMachineMetaOperations registerMetaOperations ()
    {
        if (metaOps == null)
        {
            metaOps = new E2EMetaOperations();
        }
        return metaOps;
    }

	/**
	 * Constructs a MessageStoreFSM_E2EACK object with OL_Socket and MessageStore. 
	 * @param	Socket			OL_Socket
	 * @param   mstore			MessageStore
	 */
	public MessageStoreFSM_E2EACK(OL_Socket Socket,MessageStore mstore,HyperCastConfig config){
		/*
		 * Initialization
		 */
		socket = Socket;
		MStore = mstore;
		MyState = Init;
		NumACKRequestsSent = 0;
		ACKReceived = new Vector();
		ACKExpected = new Vector();
		ACKSent = false;
		PartialDone = false;
		mylogicaladdress = socket.getLogicalAddress();
		/*
		 * Obtain the value of configurable parameter from the config object
		 */

	//The configuration file is accessed only one time.
	if (timeout_ACK < 0)
	{
		//System.out.println("MessageStoreFSM_E2EACK: read configuration file.");

		timeout_ACK    = config.getPositiveLongAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/GlobalACK/TimeoutACK"));
		max_ACK    = config.getPositiveLongAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/GlobalACK/MaxTimeoutACK"));
		timeout_NACK   = config.getPositiveLongAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/GlobalACK/TimeoutNACK"));
		max_NACK       = config.getPositiveLongAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/GlobalACK/MaxTimeoutNACK"));
		timeout_RESET  = config.getPositiveLongAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/GlobalACK/TimeoutReset"));
		max_RESET      = config.getPositiveLongAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/GlobalACK/MaxTimeoutReset"));
		delete         = config.getPositiveLongAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/GlobalACK/Delete"));
		final_Delete   = config.getPositiveLongAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/GlobalACK/FinalDelete"));
		routing_CHANGE = config.getPositiveLongAttribute 
				(XmlUtil.createXPath("/Public/NetworkService/EnhancedDelivery/GlobalACK/RoutingCheck"));
	}
          
		/*
		 * In case that the finite state machine stuck at the Init state,
		 * it will delete itself when the final_Delete timeout occurs.
		 */
		MStore.setTimer(this,FINALDELETE_TIMER_INDEX,final_Delete);
	}
	/** 
	 * This method is used to adapt the dynamically changing tree
	 * structure of the overlay topology. It will take different actions
	 * depending on the current state of the finite state machine.
	 */	
	public void changingNeighbor()
	{
		 mylogicaladdress = socket.getLogicalAddress();
		 switch (MyState)
		 {
			case Init:
				break;
			case NoPayLoad:
			{
			 	setNeighbor();
			}
			 break;
			case WaitforACK:
			{
				/*
				 * if leaf node, change to Done state
				 */
				if(length == 0)
				{
					MyState = Done;
					MStore.setTimer(this,DELETE_TIMER_INDEX,delete);
				}
				/*
				 * update the children nodes
				 */
				else 
				{
					if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
					{
						I_LogicalAddress[] childrenArray = socket.getChildren(root);
						for(int i =0; i<childrenArray.length; i++){
							if(!ACKReceived.contains(childrenArray[i])&& !ACKExpected.contains(childrenArray[i]))
							{
								ACKExpected.addElement(childrenArray[i]);
							}
								
						}		
					}
				}
			}
			 break;
			 /*
			  * no action if in Done or Reset State
			  */
			case Done:
			case Reset:
			 break;
		 }
	}  
	
	/**
	 * @return the finite state machine id
	 */
	public short getFSMID()
	{
		return MessageStore.FSM_E2E_ACK;
	}


	/**Set the new ADF_Data message entry
	 * the information buffered includes the message id, the root of the tree
	 * and the payload data 
	 * @param	msg			OL_Message
	 */
	private void setADMsg(OL_Message msg)
	{

		if (message == null) message = msg;

		messageid = msg.getMessageIdentifier();
		deliveryMode = msg.getDeliveryMode();
		switch (deliveryMode)
		{
			case OL_Message.DELIVERY_MODE_MULTICAST:
			{
				root = msg.getSourceAddress();
			}
				break;
			case OL_Message.DELIVERY_MODE_UNICAST:
			{
				root = msg.getDestinationAddress();	

				/*
				 * The source address is used by the destination to send back acknowledge.
				 */
				sourceaddress = msg.getSourceAddress();
			}
				break;

			default:
			throw new HyperCastWarningRuntimeException ("E2E ACK applies to unicast message and multicast message only");
					 
		}
	}
	
	/**Set the new ADF_Control message entry
	 * the information buffered includes the message id, the root of the tree
	 * @param	msg			OL_Message
	 */
	private void setACMsg(OL_Message msg)
	{
		messageid = msg.getMessageIdentifier();
		deliveryMode = ((FSM_Extension) msg.getFirstExtensionByType(Extension.FSM)).getDeliveryMode();
		/*
		 * each control message carries the root logical address
		 * this is necessary when the control message arrives earlier than the payload message
		 */
		root = socket.createLogicalAddress(((FSM_Extension) msg.getFirstExtensionByType(Extension.FSM)).getSenderAddress(),0);

	}
	
	/**
	 * Set the neighbor information
	 * This is calculated based on the root of the tree
	 */
	private void setNeighbor()
	{
		if (debug)
			System.out.println("MessageStoreFSM_E2EACK: setNeighbor: root="+root.toString()+", mylogicaladdress="+mylogicaladdress.toString());

		if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
		{
			I_LogicalAddress[] childrenArray = socket.getChildren(root);
			
			/*
			 * leaf node, no children node
			 */
			if(childrenArray==null)
			{
				length = 0;
			}
			/*
			 * none leaf node, update the children nodes
			 */
			else
			{
				length = childrenArray.length;
				for(int i =0; i<childrenArray.length; i++)
				{
					ACKExpected.addElement(childrenArray[i]);
				}
			
			}
		}
		else if(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
		{
			/*
			 * root (or destination) node
			 */
			 if(root.equals(mylogicaladdress))
			 {
				if (debug)
					System.out.println("MessageStoreFSM_E2EACK: setNeighbor: root == mylogicaladdress");

			 	length = 0;
			 }
			 /*
			  * Otherwise, it is the source node. All intermediate nodes don't
			  * build FSM for unicast E2E message.
			  */
			 else
			 {
				if (debug)
					System.out.println("MessageStoreFSM_E2EACK: setNeighbor: root != mylogicaladdress");

				/*
				 * For unicast e2e ack message, the source is waiting the ack from
				 * destination.
				 */

				length = 1;
				ACKExpected.addElement(root);
			 }
		}
	
	}
	
	/**
	 * This is the method used to send ack or nack message to parent node.
	 * It can only be called when the root, message id are initiated.
	 * @param type the type of the message.
	 * @param destination the destination of the message.
	 */
	private void sendToParent(byte type, I_LogicalAddress destination)
	{
		OL_Message msg = (OL_Message) socket.createMessage(); //no payload
		
		FSM_Extension ee = new FSM_Extension(MessageStore.FSM_E2E_ACK, type, messageid, deliveryMode, root);
		msg.addExtension(ee);
		
		socket.sendToParent(msg, destination);
	}

	/**
	 * This is the method used by the destination of a unicast E2E message to 
	 * send ack or nack message to the source.
	 * It can only be called when the root, message id are initiated.
	 * @param type the type of the message.
	 * @param destination the node to which the message is sent.
	 */
	private void sendToNode(byte type, I_LogicalAddress destination)
	{
		if (debug)
			System.out.println("MessageStoreFSM_E2EACK: sendToNode: type="+type+", destination="+destination);

		OL_Message msg = (OL_Message) socket.createMessage(); //no payload

		FSM_Extension ee = new FSM_Extension(MessageStore.FSM_E2E_ACK, type, messageid, deliveryMode, root);
		msg.addExtension(ee);
		
		socket.sendToNode(msg, destination);
	}
	
	/**
	 * This is the method used to update the information about which node received the payload.
	 */
	private synchronized void processAck(I_LogicalAddress la)
	{
		if (debug)
			System.out.println("MessageStoreFSM_E2EACK: processAck: ack is received from "+la.toString());
		
		
			ACKReceived.addElement(la);
			if(ACKExpected.contains(la))
			{
				ACKExpected.removeElement(la);
			}

		if (debug)
			System.out.println("MessageStoreFSM_E2EACK: processAck: ACKExpected.isEmpty() is "+ACKExpected.isEmpty());

	}
	
	/**
	 * This is the operation to notify the arriving of full or partial ack to the application.
	 */
	private void notifyApplication()
	{

		if (debug)
			System.out.println("MessageStoreFSM_E2EACK: notifyApplication(): Report to application event happens.");

		if (MStore.notificationHandler != null) 
		{
			if(PartialDone)
			{
				E2EPARTIALACK_RECEIVED se = new E2EPARTIALACK_RECEIVED(System.currentTimeMillis(), messageid);
				MStore.notificationHandler.eventOccurred(se);
			}
			else
			{
				E2EACK_RECEIVED se = new E2EACK_RECEIVED(System.currentTimeMillis(), messageid);
				MStore.notificationHandler.eventOccurred(se);
			}
		}
		
	}
	
	/**process a new formated application message
	 * @param	msg			OL_Message
	 */
	public void newADMsg(OL_Message msg)
	{
		if (debug)
			System.out.println("MessageStoreFSM_E2EACK: newADMsg: received a new message.");

		if (debug)
			System.out.println("MessageStoreFSM_E2EACK: newADMsg: pass the new message to application.");


		/*
		 * Buffer the essential information such as message id, root of the tree
		 * and the payload.
		 */
		setADMsg(msg);
		
		/*
		 * calculate the topology tree based on the root logical address
		 */
		setNeighbor();

		/*
		 * For E2EACK service, if it is multicast, each nodes receives the message and passes it
		 * to application; if it is unicast, only the destination node passes
		 * it to application.
		 */
		/*
		 * Firstly, the receiver forward the message to the application.
		 */
		if ((deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST) ||
			((deliveryMode == OL_Message.DELIVERY_MODE_UNICAST) && (mylogicaladdress.equals(root))))
		{
			socket.appmsgArrived(msg,socket.callback);
		}

		
		if (debug)
			System.out.println("MessageStoreFSM_E2EACK: newADMsg: ACKExpected.isEmpty() is "+ACKExpected.isEmpty());

		if (debug)
		{
			String logMessage =
				"MessageStoreFSM_E2EACK: newADMsg: The received message: " +
				"src = " + msg.getSourceAddress() + ", " +
				"dst = " + msg.getDestinationAddress() + ", " +
				"prevhop = " + msg.getPreviousHopAddress();
			System.out.println(logMessage);
		}
		
		/*
		 * leaf node, change to Done state
		 */
		
		if(ACKExpected.isEmpty())
		{
			MyState = Done;
			MStore.setTimer(this,DELETE_TIMER_INDEX,delete);

			if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
			{
				/*
				 * not root, send full ack to parent node
				 */
				if(!mylogicaladdress.equals(root))
				{
					if (debug)
						System.out.println("MessageStoreFSM_E2EACK: newADMsg: send ack to "+msg.getPreviousHopAddress().toString());

					/*
					 * This is wrong, the destination is not root, but the parent with
					 * respect to root.
					 */
					sendToParent(FULL_E2E_ACK, root);
					ACKSent = true;						
				
				}
				/*
				 * if the current node is the root, notify the application the arrival of the full ack
				 */
				else
				{
					if (debug)
						System.out.println("MessageStoreFSM_E2EACK: newADMsg: call notifyApplication()");

					PartialDone = false;
					notifyApplication();
				} 
			}
			else if(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
			{
				/*
				 * if the current node is not the root (it must be the source), notify the application the arrival of the full ack.
				 */
				if(!mylogicaladdress.equals(root))
				{
					if (debug)
						System.out.println("MessageStoreFSM_E2EACK: newADMsg: call notifyApplication()");

					PartialDone = false;
					notifyApplication();
				}
				/*
				 * if the current node is the root (or destination), send full ack directly to the source of the message.
				 */
				else
				{
					if (debug)
						System.out.println("MessageStoreFSM_E2EACK: newADMsg: send ack to "+msg.getPreviousHopAddress().toString());

					sendToNode(FULL_E2E_ACK,sourceaddress);
					ACKSent = true;						
				} 
			}
		}
		
		/*
		 * not leaf node, change to WaitforACK state
		 */
		else
		{
			MyState = WaitforACK;
			MStore.setTimer(this,ACK_TIMER_INDEX,timeout_ACK);
			MStore.setTimer(this,MAXACK_TIMER_INDEX,max_ACK);
		}
				
	}
	/**process new ADF_Control message
	 * @param	msg			OL_Message
	 */
	public void newACMsg(OL_Message msg)
	{


		/*
		 * store the root, the message id
		 * update the children nodes
		 */
		setACMsg(msg);
		setNeighbor();
		/*
		 * extract the format message header information
		 */
		FSM_Extension e = ((FSM_Extension) msg.getFirstExtensionByType(Extension.FSM));
		
		
		/*
		 * state transit based on the message type
		 */
		/*
		 * state transit based on the message type
		 */
		switch(e.getType()){
			case FULL_E2E_ACK:
			case FINAL_PARTIAL_E2E_ACK:	
			case E2E_NACK:
				break;
			case E2E_REQUEST:
			{

				if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
				{
					if(!mylogicaladdress.equals(root))
					{
						/*
						 * send nack to parent node
						 */
						sendToParent(E2E_NACK,msg.getSourceAddress());
						/*
						 * change the state to NoPayLoad
						 */
						MyState = NoPayLoad;
						MStore.setTimer(this,NACK_TIMER_INDEX,timeout_NACK);
						MStore.setTimer(this,MAXNACK_TIMER_INDEX,max_NACK);
					}
					/*
					 * if the current node is the root, then some thing inconsistency happened,
					 * change the state to Reset.
					 */
					else if(mylogicaladdress.equals(root))
					{
						MyState = Reset;
						MStore.setTimer(this,RESET_TIMER_INDEX,timeout_RESET);
						MStore.setTimer(this,MAXRESET_TIMER_INDEX,max_RESET);
					}
				}
				else if(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
				{
					if(mylogicaladdress.equals(root)) //i.e., it is the destination.
					{
						sourceaddress = msg.getSourceAddress();

						/*
						 * send nack to parent node
						 */
						//sendToParent(E2E_NACK,msg.getSourceAddress());
						sendToNode(E2E_NACK,sourceaddress);
						/*
						 * change the state to NoPayLoad
						 */
						MyState = NoPayLoad;
						MStore.setTimer(this,NACK_TIMER_INDEX,timeout_NACK);
						MStore.setTimer(this,MAXNACK_TIMER_INDEX,max_NACK);
					}
					/*
					 * if the current node is the source, then some thing inconsistency happened,
					 * change the state to Reset.
					 */
					else {
						MyState = Reset;
						MStore.setTimer(this,RESET_TIMER_INDEX,timeout_RESET);
						MStore.setTimer(this,MAXRESET_TIMER_INDEX,max_RESET);
					}
				}				
			}
				break;
			/*
			 * all other types of messages
			 * some thing inconsistency happened,change the state to Reset.
			 */	
			default:
			{
				MyState = Reset;
				MStore.setTimer(this,RESET_TIMER_INDEX,timeout_RESET);
				MStore.setTimer(this,MAXRESET_TIMER_INDEX,max_RESET);
			}
				break;	
							
		}

	
	}	
	
	/**Update new ADF_Data message
	 * @param	msg			OL_Message
	 */
	public void updateADMsg(OL_Message msg) 
	{
		if (message == null)
		{
			message = msg;
		}
		/*
		 * If the message is received before, don't process the comming packet. 
		 */
		else
		{
			return;
		}

		/*
		 * extract the format message header information
		 */
		FSM_Extension e = ((FSM_Extension) msg.getFirstExtensionByType(Extension.FSM));
		

		if (debug)
			System.out.println("MessageStoreFSM_E2EACK:updateADMsg: MyState is "+ MyState);
		
		/*
		 * take different actions at different state
		 */
		switch (MyState)
		{
			case Init:
			{
				/*
				 * forward the message to the application
				 */
				socket.appmsgArrived(msg,socket.callback);
				/*
				 * if all ack received, change to Done state
				 */
				if(ACKExpected.isEmpty())
				{
					MyState = Done;
					MStore.setTimer(this,DELETE_TIMER_INDEX,delete);

					if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
					{
						/*
						 * send ack if not the root and has not sent the ack before
						 */
						if(!mylogicaladdress.equals(root) && !ACKSent)
						{
							sendToParent(FULL_E2E_ACK,root);
							ACKSent = true;
						}
						/*
						 * if the current node is the root, notify the application the arrival of the full ack
						 */
						else if(mylogicaladdress.equals(root))
						{
							if (MStore.notificationHandler != null) 
							{
								notifyApplication();
							}
						} 
					}
					else if(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
					{
						/*
						 * if the current node is not the root (it must be the source), notify the application the arrival of the full ack.
						 */
						if(!mylogicaladdress.equals(root))
						{
							PartialDone = false;
							notifyApplication();
						}
						/*
						 * if the current node is the root (or destination), send full ack directly to the source of the message.
						 */
						else
						{
							sendToNode(FULL_E2E_ACK,sourceaddress);
							ACKSent = true;						
						} 
					}
				}
				/*
				 * pending ack, change to WaitforACK state
				 */
				else
				{
					MyState = WaitforACK;
					MStore.setTimer(this,ACK_TIMER_INDEX,timeout_ACK);
					MStore.setTimer(this,MAXACK_TIMER_INDEX,max_ACK);
				}

			}

			break;
				
			case NoPayLoad:	
			{

				/*
				 * forward the message to the application
				 */
				socket.appmsgArrived(msg,socket.callback);
				
				/*
				 * if all ack received, change to Done state
				 */
				if(ACKExpected.isEmpty())
				{
					MyState = Done;
					MStore.setTimer(this,DELETE_TIMER_INDEX,delete);


					if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
					{
						/*
						 * not root, send full ack to parent node
						 */
						if(!mylogicaladdress.equals(root) && !ACKSent)
						{
							sendToParent(FULL_E2E_ACK, root);
							ACKSent = true;						
				
						}
						/*
						 * if the current node is the root, notify the application the arrival of the full ack
						 */
						else if (mylogicaladdress.equals(root))
						{
							PartialDone = false;
							notifyApplication();
						} 
					}
					else if(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
					{
						/*
						 * if the current node is not the root (it must be the source), notify the application the arrival of the full ack.
						 */
						if(!mylogicaladdress.equals(root))
						{
							PartialDone = false;
							notifyApplication();
						}
						/*
						 * if the current node is the root (or destination), send full ack directly to the source of the message.
						 */
						else
						{
							sendToNode(FULL_E2E_ACK,sourceaddress);
							ACKSent = true;						
						} 
					}
					
				}
				else
				{
					if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
					{
						OL_Message retrans = new OL_Message (message);
						FSM_Extension ee = ((FSM_Extension) msg.getFirstExtensionByType(Extension.FSM));	
						ee.setType(PAYLOAD);

						retrans.setDeliveryMode(OL_Message.DELIVERY_MODE_MULTICAST);
						retrans.setDestinationAddress(null);
						retrans.setHopLimit((short) 255);						
						socket.forwardToChildren(retrans);
					}
					
					MyState = WaitforACK;
					MStore.setTimer(this,ACK_TIMER_INDEX,timeout_ACK);
					MStore.setTimer(this,MAXACK_TIMER_INDEX,max_ACK);
				}
	
			}

				break;

				/*
				 * no action if the current node has already obtained the payload
				 */	
				case(WaitforACK):
				case(Done):
					break;
		}
	}
	
	/**process new ADF_Control message
	 * @param	msg			OL_Message
	 */
	public void updateACMsg(OL_Message msg)
	{
		/*
		 * extract the format message header information
		 */
		FSM_Extension e = ((FSM_Extension) msg.getFirstExtensionByType(Extension.FSM));	

		if (debug)
			System.out.println("MessageStoreFSM_E2EACK:updateACMsg: MyState is "+ MyState);

		/*
		 * take different actions at different state
		 */
		switch(MyState)
		{
			case Init:
				break;
			case NoPayLoad:
			{
				switch(e.getType()){
					case E2E_REQUEST:
					{
						if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
						{
							if(!mylogicaladdress.equals(root))
							{
								/*
								 * send nack to parent node
								 */
								sendToParent(E2E_NACK,msg.getSourceAddress());
							}
						}
						else if(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
						{
							if(mylogicaladdress.equals(root)) //i.e., it is the destination.
							{
								/*
								 * send nack to parent node
								 */
								sendToNode(E2E_NACK,sourceaddress);
							}
						}				

					}
						break;
					case E2E_NACK:
					break;
					case E2E_RESET:
					{
						MyState = Reset;
						MStore.setTimer(this,RESET_TIMER_INDEX,timeout_RESET);
						MStore.setTimer(this,MAXRESET_TIMER_INDEX,max_RESET);
					}
						break;
				}
			}
		
		case WaitforACK:
		{
			if (debug)
				System.out.println("MessageStoreFSM_E2EACK:updateACMsg: process WaitforACK");

			switch(e.getType()){
				case FULL_E2E_ACK:
				{
					if (debug)
						System.out.println("MessageStoreFSM_E2EACK:updateACMsg: process WaitforACK_FULL_E2E_ACK");

					/*
					 * update the nodes that send ack
					 */
					processAck(msg.getSourceAddress());
					
					/*
					 * all ack received, change to Done state 
					 */
					if(ACKExpected.isEmpty())
					{
						MyState = Done;	
						MStore.setTimer(this,DELETE_TIMER_INDEX,delete);
							
						if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
						{
							if (debug)
								System.out.println("MessageStoreFSM_E2EACK:updateACMsg: mylogicaladdress="+mylogicaladdress+", root="+root);
	
							/*
							 * the current node is root. notify the application
							 */
							if(mylogicaladdress.equals(root))
							{
								PartialDone = false;
								notifyApplication();
							}
							/*
							 * send ack if not the root and has not sent the ack before
							 */
							else if(!mylogicaladdress.equals(root) && !ACKSent){
								sendToParent(FULL_E2E_ACK, root);
								ACKSent = true;
							}
						}
						else if(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
						{
							/*
							 * If current node is the source, notify the application.
							 */
							if(mylogicaladdress.equals(sourceaddress))
							{
								PartialDone = false;
								notifyApplication();
							}
							/*
							 * send ack if not the source and has not sent the ack before
							 */
							else if((sourceaddress != null) && !mylogicaladdress.equals(sourceaddress) && !ACKSent){
								sendToNode(FULL_E2E_ACK, sourceaddress);
								ACKSent = true;
							}
						}						
					}
				}
					break;
					
				case FINAL_PARTIAL_E2E_ACK:
				{
					/*
					 * update the node that send ack
					 */
					processAck(msg.getSourceAddress());
					PartialDone = true;
					/*
					 * change the state to Dane state
					 */
					MyState = Done;
					MStore.setTimer(this,DELETE_TIMER_INDEX,delete);

					/*
					 * send partial ack
					 */
					if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
					{
						if(!mylogicaladdress.equals(root))
						{
							/*
							 * send nack to parent node
							 */
							sendToParent(FINAL_PARTIAL_E2E_ACK, root);
						}
						else
						{
							notifyApplication();
						}
					}
					else if(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
					{
						if(mylogicaladdress.equals(root)) //i.e., it is the destination.
						{
							/*
							 * send nack to source node
							 */
							sendToNode(FINAL_PARTIAL_E2E_ACK,sourceaddress);
						}
						else
						{
							notifyApplication();
						}
					}				

				}
					break;
				/*
				 * retransmit the payload
				 */
				case E2E_NACK:
				{
					OL_Message retrans = new OL_Message (message);
					FSM_Extension ee = ((FSM_Extension) msg.getFirstExtensionByType(Extension.FSM));	
					ee.setType(PAYLOAD_RETRAN);

					System.out.println("MessageStoreFSM_E2EACK: retransmit packets.");

					if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
					{
						retrans.setDeliveryMode(OL_Message.DELIVERY_MODE_UNICAST);
						retrans.setDestinationAddress(msg.getSourceAddress());
						retrans.setHopLimit((short) 1);						
						socket.forwardToParent(retrans);
					}
					else if(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
					{
						socket.forwardToParent(retrans);
					}						

					MStore.incrementRetransmissionCount ();
				}
					break;
					/*
					 * change to Reset state
					 */
				case E2E_RESET:
				{
					MyState = Reset;
					MStore.setTimer(this,RESET_TIMER_INDEX,timeout_RESET);
					MStore.setTimer(this,MAXRESET_TIMER_INDEX,max_RESET);
				}
					break;
			}
		}
			
		case Done:
		{
			switch(e.getType()){
				case E2E_REQUEST:
				{
					if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
					{
						if(!mylogicaladdress.equals(root))
						{
							if(PartialDone){
								sendToParent(FINAL_PARTIAL_E2E_ACK, msg.getSourceAddress());
							}
							else
							{
								sendToParent(FULL_E2E_ACK, msg.getSourceAddress());
							}
							ACKSent = true;
						}
					}
					else if(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
					{
						if(mylogicaladdress.equals(root)) //i.e., it is the destination.
						{
							if(PartialDone){
								sendToNode(FINAL_PARTIAL_E2E_ACK, msg.getSourceAddress());
							}
							else
							{
								sendToNode(FULL_E2E_ACK, msg.getSourceAddress());
							}
							ACKSent = true;
						}
					}				
				}
					break;

//TODO: It seems that we should ignore this E2E_NACK message. Is it possible?
					/*
					 * retransmit the payload
					 */
				case E2E_NACK:
				{
					OL_Message retrans = new OL_Message (message);
					FSM_Extension ee = ((FSM_Extension) msg.getFirstExtensionByType(Extension.FSM));	
					ee.setType(PAYLOAD_RETRAN);
					
					if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
					{
						retrans.setDeliveryMode(OL_Message.DELIVERY_MODE_UNICAST);
						retrans.setDestinationAddress(msg.getSourceAddress());
						retrans.setHopLimit((short) 1);						
						socket.forwardToParent(retrans);
					}
					else if(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
					{
						socket.forwardToParent(retrans);
					}						
					
					MStore.incrementRetransmissionCount ();
				}
					break;	
					/*
					 * change to Reset state
					 */
				case E2E_RESET:
				{
					MyState = Reset;
					MStore.setTimer(this,RESET_TIMER_INDEX,timeout_RESET);
					MStore.setTimer(this,MAXRESET_TIMER_INDEX,max_RESET);
				}	
			}
		} 
			break;	
		case Reset:	
			break;	
		}//end of switch
	}

	
	
	/** process the timer expired events
	 */
	public void timerExpired(int timerIndex)
	{
		/*
		 * periodically check the topology changing
		 */
		 if(timerIndex==ROUTINGCHANGE_TIMER_INDEX)
		 {
			changingNeighbor();
			MStore.setTimer(this,ROUTINGCHANGE_TIMER_INDEX,routing_CHANGE);
		 }
		 /*
		  * take action based on state
		  */
		switch(MyState){
			case Init:
			{
				if(timerIndex==FINALDELETE_TIMER_INDEX)
				{
					MStore.remove(messageid);
				}
			}
				break;
			 case(NoPayLoad):
			 {
			 	/*
			 	 * Nack timeout, send nack
			 	 */
				 if(timerIndex==NACK_TIMER_INDEX)
				 {
					if (deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
					{
						if ((!root.equals(null)) && (!mylogicaladdress.equals(root)))
							sendToParent(E2E_NACK,root);
					}
					else if (deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
					{
						if ((!sourceaddress.equals(null)) && (!mylogicaladdress.equals(sourceaddress)))
							sendToNode(E2E_NACK,sourceaddress);
					}
				 }
				 /*
				  * max nack time out, change to Reset state
				  */
				 else if(timerIndex==MAXNACK_TIMER_INDEX)
				 {
					 MyState = Reset;
					 MStore.setTimer(this,RESET_TIMER_INDEX,timeout_RESET);
					 MStore.setTimer(this,MAXRESET_TIMER_INDEX,max_RESET);
				 }
			 }
				 break;
			case WaitforACK:
			{
				if(timerIndex==ACK_TIMER_INDEX )
				{

					OL_Message request = (OL_Message) socket.createMessage(); //no payload
					request.setDeliveryMode(OL_Message.DELIVERY_MODE_UNICAST);
		
					FSM_Extension ee = new FSM_Extension(MessageStore.FSM_E2E_ACK, E2E_REQUEST, messageid, deliveryMode, root);
					request.addExtension(ee);

					for (int i=0;  i<ACKExpected.size();i++)
					{
						I_LogicalAddress la = (I_LogicalAddress)ACKExpected.elementAt(i);
						request.setDestinationAddress(la);
						socket.forwardToParent(request);
					}
					MStore.setTimer(this,ACK_TIMER_INDEX,timeout_ACK);

				}
				else if(timerIndex==MAXACK_TIMER_INDEX)
				{
					if (ACKExpected.size() > 0)
					{
						if((!mylogicaladdress.equals(root))&&(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST))
						{
							//sendToParent(FINAL_PARTIAL_E2E_ACK, socket.getParent(root));
							sendToParent(FINAL_PARTIAL_E2E_ACK, root);
						}
						else if ((mylogicaladdress.equals(root))&&(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST))
						{
							sendToNode(FINAL_PARTIAL_E2E_ACK, sourceaddress);
						}
					}
					
					/*
					 * change the state to Done state
					 */
					MyState = Done;
					MStore.setTimer(this,DELETE_TIMER_INDEX,delete);
				}
				
				
			}
			
		case Reset:
		{
			if((timerIndex==RESET_TIMER_INDEX) && (length!=0))
		 	{

				if(deliveryMode == OL_Message.DELIVERY_MODE_MULTICAST)
				{
					I_LogicalAddress[] childaddress = socket.getChildren(root);
					for(int i=0;i<childaddress.length;i++)
					{
						sendToNode(E2E_RESET,childaddress[i]);
					}
				}
				else if(deliveryMode == OL_Message.DELIVERY_MODE_UNICAST)
				{
					if (!mylogicaladdress.equals(root)) //i.e., it is source.
					{
						sendToNode(E2E_RESET,root);
					}
				}
			}
			else if(timerIndex==MAXRESET_TIMER_INDEX)
			{
				MStore.remove(messageid);
			}
		}
			break;
		case Done:
		 {
			 if(timerIndex==DELETE_TIMER_INDEX){
				MyState = Init;

				/*
				 * delete itself
				 */
				MStore.remove(messageid);
			 }
		 }
		 	break;
			
		}
		
	}
}

abstract class AcknowledgmentMetaOperations implements FiniteStateMachineMetaOperations
{
    public boolean messageStoreWillForwardMessage (OL_Message message)
    {
        return false;
    }

	public boolean processIntermediateUnicastMessage(OL_Message message)
	{
		return (((FSM_Extension) message.getFirstExtensionByType(Extension.FSM)).getType()==FSM_Extension.EXTENSION_TYPE_PAYLOAD);
	}

  /*
   * The condition for a FSM message to be Valid is not the same for different service ID:
   * 
   * H2HACK,   message-oriented,  unicast and broadcast
   * E2EACK,   message-oriented,  unicast and broadcast
   * SYNC,     message-oriented,  broadcast
   * INCAST,   message-oriented,  unicast
   * DUPELI,   message-oriented,  all
   * INORDER,  stream-oriented,   all
   * NAMING,   stream-oriented,   ?
   * 
   * So the meta operation defined in a FSM instance needs to override below method.
   */ 
    public boolean isValidMessage (OL_Message message)
    {
  
        return (message.getDeliveryMode() != OL_Message.DELIVERY_MODE_FLOOD);
        
    }
}

class E2EMetaOperations extends AcknowledgmentMetaOperations
{
	/*
	 * For an E2E unicast message, intermediate nodes just forward it in the 
	 * forward engine but don't build FSM for it in MessageStore.
	 */
	public boolean processIntermediateUnicastMessage(OL_Message message)
	{
		return false;
	}

    public NOTIFICATION_EVENT notifyOnSend (OL_Message message)
    {
		/*
		 * MSG_WITH_E2EACK_SENT event is deleted.
		 */

        //return new MSG_WITH_E2EACK_SENT (System.currentTimeMillis(), message.getMessageIdentifier());
        return null;
    }
}
