/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.NONE;

import hypercast.I_AddressPair;
import hypercast.I_LogicalAddress;
import hypercast.I_PhysicalAddress;

/**
 * In the NONE protocol, a logical address and physical address are
 * the same, thus an address pair need only have a single address.
 * 
 * @author HyperCast Team
 * @author Guangyu Dong
 * @version 2005 (version 3.0)
 */
public class NONE_AddressPair
    implements I_AddressPair
{
    private NONE_LogicalAddress logicalAddress;

    public NONE_AddressPair (NONE_LogicalAddress logicalAddress)
    {
        if (logicalAddress == null)
            throw new NullPointerException();

        this.logicalAddress = logicalAddress;
    }

    /**
     * @see hypercast.I_AddressPair#getPhysicalAddress()
     */
    public I_PhysicalAddress getPhysicalAddress()
    {
        return logicalAddress.getPhysicalAddress();
    }

    /**
     * @see hypercast.I_AddressPair#getLogicalAddress()
     */
    public I_LogicalAddress getLogicalAddress()
    {
        return logicalAddress;
    }

    public String toString()
    {
        return logicalAddress.toString();
    }

    public boolean equals(Object o){
    	if(o==null)return false;
    	if(!(o instanceof I_AddressPair))
    		return false;
    	I_AddressPair ap = (I_AddressPair)o;
    	
    	return logicalAddress.equals(ap.getLogicalAddress()) && logicalAddress.equals(ap.getPhysicalAddress());
    	
    }
}
