/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/
package hypercast.SPT;

import hypercast.I_UnicastAdapter;

/**
 * This class measure the link quality between nodes
 * 
 * @author HyperCast Team
 * @author Guangyu Dong
 * @version 2005 (version 3.0)
 */

class LinkQuality{
	
	/** last time a ping message has been received 
	 * here it's the beacon message.
	 */
	long lastPing;
	/**
	 * A buffer recording which beacons have ping messages have been received
	 * during the last recording period. The length of the buffer is decided
	 * when initalizing. And the length*pingInterval=a recording period
	 */
	boolean pingBuf[];
	/** the start index of the ping buffer for recording the ping message */
	int start;
	/** the preset interval between two ping messages. 
	 * Here it should be set to the beacon period
	 */ 
	long pingInterval;
	/** for getting time */
	I_UnicastAdapter adapter;
	
	/** the reliable threshold */
	float ReliableThreshold=0;
	
	/** my LA */
	SPT_LogicalAddress myLA;
	
	/** latest beacon */
	SPT_Message lastBeacon;

	public LinkQuality(int pingBufSize, long interval, float reliThreshold, SPT_LogicalAddress la, I_UnicastAdapter adpt){
		
		myLA=la;
		adapter=adpt;
		ReliableThreshold=reliThreshold;
		
		pingBuf = new boolean[pingBufSize];
		for (int i = 0; i < pingBuf.length; i++)
			pingBuf[i] = false;
		pingInterval = interval;
		lastPing = adapter.getCurrentTime();
		
		start = 0;
		
	}
	public void beaconReceived(SPT_Message beacon){
		lastBeacon=beacon;
		
		recordPing(adapter.getCurrentTime());
	}
	
	/**
	 * Get the bidirectional link quality
	 * @return	the reliability value
	 */
	public float getBiLinkQuality() {
		int reli=getPingNum();
		if(reli>getReversedPingNum())reli=getReversedPingNum();
		
		return (float)reli/pingBuf.length;
	}
	/**
	 * Get the link quality from this neighbor
	 */
	public float getLinkQuality(){
		return (float)getPingNum()/pingBuf.length;
	}
	/**
	 * Get the link quality to this neighbor
	 */
	public float getReversedLinkQuality(){
		return (float)getReversedPingNum()/pingBuf.length;
	}
	
	/**
	 * Check if this link is relaible
	 * @return true/false
	 */
	public boolean isReliable() {
		return getBiLinkQuality() >= ReliableThreshold ;
	}

	/**
	 * @return Returns the reversedReliability.
	 */
	int getReversedPingNum() {
		if(lastBeacon!=null)
			return lastBeacon.getReversedOneHopReliability(myLA);
		return 0;
	}

	private void recordNoPing(long ts){
		int temp_start=start;
		long interval = (ts - lastPing + pingInterval / 2) / pingInterval;
		//System.out.println("interval="+interval);

		for (int i = 0; i < interval-1; i++) {
			pingBuf[temp_start] = false;
			
			temp_start++;
			if (temp_start >= pingBuf.length)
				temp_start = 0;

		}
	}
	private void recordPing(long ts) {
		/* calculate how many intervals has elpased since last ping */
		long interval = (ts - lastPing + pingInterval / 2) / pingInterval;
		//System.out.println("interval="+interval);
		if(interval<1){
			int end=start-1;
			if(end<0)end=pingBuf.length-1;
		
			for(int i=0;i<pingBuf.length;i++){
				if(!pingBuf[end]){
					pingBuf[end]=true;
					break;
				}else{
					end--;
					if(end<0)end=pingBuf.length-1;
				}
			}
		}else for (int i = 0; i < interval; i++) {
			if (i < interval - 1)
				pingBuf[start] = false;
			else
				pingBuf[start] = true;
			start++;
			if (start >= pingBuf.length)
				start = 0;

		}
		lastPing = ts;
	}
    /**
	 * get how many pings have been received in the current recording period
	 * @return	the number of pings
	 */
	int getPingNum() {
		long now=adapter.getCurrentTime();
		
		this.recordNoPing(now);
		
		int i = start;
		int sum = 0;
		for (int j = 0; j < pingBuf.length; j++) {
			if (pingBuf[i])
				sum++;
			i++;
			if (i >= pingBuf.length)
				i = 0;
		}
		return sum;
	}
	
	
}