/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.SPT;

/**
 * This class represents a Spanning Tree ID
 *
 * @author HyperCast Team
 * @author Guangyu Dong
 * @version	0.1, May 25, 2003
 */

class SPT_ID
{
	/** the SPT ID */
	protected int spt_id;
    /**
     * Constructor
     * @param id	the ID in integer form
     */
    public SPT_ID(int id)
    {
        spt_id=id;
        if(spt_id==0)throw new IllegalArgumentException("Invalid SPT Id "+spt_id);
    }
	/**
	 * Constructor
	 * @param p		another SPT_ID object
	 */
    public SPT_ID(SPT_ID p)
    {
        spt_id=p.spt_id;
    }

	/**
	 * Constructor
	 * @param s		the ID in string form
	 */
	public SPT_ID(String s) 
    {
        try 
        {
            spt_id = Integer.parseInt(s);
        }
        catch (Exception e) 
        {
            throw new IllegalArgumentException("String did not match a integer");
        }
        if(spt_id==0)throw new IllegalArgumentException("Invalid SPT Id "+spt_id);
    }
    

	/**
	 * @return		the string form of the ID
	 */
	public String toString() 
    {
        return Integer.toString(spt_id);
    }

	/**
	 * Compares to another SPT_ID
	 * @param	o an SPT_ID object
	 * @return	whether they are equal, or false if the param is null
	 * @throws 	IllegalArgumentException, if the object is not an instance of SPT_ID
	 */
	public boolean equals(Object o) 
    {
        if (o == null)  
            return false;
        
        if (!(o instanceof SPT_ID))
            throw new IllegalArgumentException("SPT_ID.equals() pass a object that was not a SPT_ID!");
        
        SPT_ID p = (SPT_ID) o;
        return spt_id==p.spt_id;
    }
	/**
	 * @see Object#hashCode()
	 */
	public int hashCode() 
    {
        return (spt_id);
    }
    /**
     * @return		the ID in integer form
     */
	public int getId() 
    {
        return spt_id;
    }
	/**
	 * Set the id
	 * @param id
	 */
	public void setId(int id)
	{
		spt_id=id;
	}
    


/** 
 * @return true if SPT_ID <code>a</code> comes before <code>b</code> in
 * the total ordering of points.
 * */
public final static boolean greaterThan(SPT_ID a, SPT_ID b) {
    return a.spt_id>b.spt_id;
}



}










