/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.util;
import hypercast.util.XmlUtil;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;

import java.util.Enumeration;
import java.util.EventObject;
import java.util.HashMap;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.util.Vector;
import java.util.Map;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;

import javax.swing.border.EtchedBorder;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.JTextPane;
import javax.swing.JToggleButton; // how is this used ???
import javax.swing.ListSelectionModel;
import javax.swing.SpinnerNumberModel;
import javax.swing.text.JTextComponent; // how is this used ???

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import javax.swing.ScrollPaneLayout;
import java.awt.*; 
import javax.swing.JSplitPane;
import javax.swing.JScrollPane;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Text;

// For creating borders
import javax.swing.border.EmptyBorder;
import javax.swing.border.BevelBorder;
import javax.swing.border.CompoundBorder;
import javax.swing.JTree;
import javax.swing.JPopupMenu;
import javax.swing.tree.*;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.ImageIcon;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

import java.text.ParseException;

//For XML property file validation. It only works under JDk 1.5.0 and above.
/*
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Validator;

import javax.xml.validation.Schema; 
import javax.xml.validation.AbstractSchemaImpl;
import javax.xml.validation.SchemaFactory; 
*/

/**
 * This class defines a Hypercast configuration file generator. It can
 * generates a new Hypercast configuration file from a Hypercast
 * schema file, or from a existing configuration file. It contains a
 * GUI for the users to edit the contents of a configuration file.
 *
 * @author HyperCast Team
 * @version 2005 (version 3.0)
 */

public class HypercastConfigurator  extends JPanel
implements ActionListener, TreeSelectionListener, MouseListener
{
	/**
	 * Debug flag.
	 */
	public final static boolean debug = false;
	
	/**
	 * The menu bar of the configuration GUI.
	 */
	private JMenuBar menu;

	/** The items under the File menu	 */
	private JMenuItem loaditem, newitem, openitem, saveitem, saveasitem, exititem;

	/**
	 * A popup menu which is displayed when right clicking
	 * on the jtree.
	 */
	private JPopupMenu popupmenu;	 

	/**The item under the popup menu	 */
	private JMenuItem expanditem, collapseitem;

	/** The Document instance created from the opened XML configuration file */
	private Document openedXMLFile = null;

	/** The XML file. */
	private File XMLFile = null;

	/** The Document instance created from the opened schema file. */
	private Document openedSchemaFile = null;

	/** The schema file. */
	private File SchemaFile = null;

	/*
	 * When reading schema file from a source file in the JAR, SchamaFile is 
	 * null but openedSchemaFile and schemaFileName are still assigned.
	 */
	/** The schema file name. */
	private String schemaFileName = null;


	/*
	 * For XML property file validation.
	 */
	/** The items under the Validation menu	 */
	private JMenuItem xmlfileitem, schemafileitem, validateitem;

	/** The schema file name for validation. */
	private String validateSchemaFileName = null;

	/** The xml file name for validation. */
	private String validateXMLFileName = null;

	/**
	 * The item for help information.
	 */
	private JMenuItem aboutitem;
	

	/**
	 * The JScrollPane instance for displaying tree.
	 */
	private JScrollPane treePanel;

	/**
	 * The JPanel for editing a tree node when it is selected.
	 */
	private JPanel editPanel;

	/**
	 * The JTextField instance for displaying opreation related information.
	 */
	private JTextField infoField;

	/**
	 * The JTextField instance for displaying if the schema file is loaded.
	 */
	private JTextField schemafileLoadInfoField;

	/**
	 * The tree structure for displaying and editing properties.
	 */
	private JTree tree;

	/**
	 * Flag indicating if the JTree object is empty or not.
	 * True--empty; False--not empty or a valid tree.
	 */
	private boolean isTreeEmpty = true;

	/**
	 * Tree node which is selected currently.
	 */
	private dispAndEditor selectedNode = null;

	/**
	 * static constants for GUI components.
	 */
	public static final int FRAME_WIDTH = 620; 
	public static final int FRAME_HEIGHT = 620; 
	public static final int LABEL_WIDTH = 300; 
	public static final int LABEL_HEIGHT = 20; 
	public static final int COMBOX_WIDTH = 300; 
	public static final int COMBOX_HEIGHT = 20; 
	public static final int SPINNER_WIDTH = 300; 
	public static final int SPINNER_HEIGHT = 20; 
	public static final int SPINNERFORNUMBER_WIDTH = 60; 
	public static final int TEXTFIELD_WIDTH = 300; 
	public static final int TEXTFIELD_HEIGHT = 20; 

	/** The root element name in the schema file. */
	public static final String ROOT_ATTRIBUTE_ELEMENT_NAME = "Public";

	/** The name of the default configulation file. */
	public static final String DEFAULT_FILENAME = "hypercast.xml";

	/**
	 * Contructor.
	 */
    public HypercastConfigurator()
    {
    }

	/*
	 * This constructor is for obtaining default value properties.
	 */
	/**
	 * Contructor.
	 */
	public HypercastConfigurator(File schemafile)
	{
		String filename = schemafile.getName();

		if (filename.toLowerCase().endsWith(".xsd"))
		{
			/*
			 * Store the reference to the selected schema file.
			 */
			SchemaFile = schemafile;
			schemaFileName = filename;
					
			/*
			 * Open schema file.
			 */
			try
			{ 
				openedSchemaFile = XmlUtil.createDocument(schemafile);
			}
			catch (final IOException ioe)
			{
				infoField.setText("Cannot load schema file "+schemaFileName+".");
			}
			catch (final ParseException ioe)
			{
				infoField.setText("Failed in parsingschema  file "+schemaFileName+".");
			}
								
			/*
			 * Remove all comments in the schema file.
			 */
			XmlUtil.removeComments(openedSchemaFile, openedSchemaFile);
					
			isTreeEmpty = true;
					
			if (selectedNode != null)
			{
				selectedNode.hidden();
			}
			selectedNode = null;
		}
	}

	/**
	 * Initialize GUI parameters and display GUI.
	 */
	public void initializeDisplay()
	{
		/*
		 * Create a JMenuBar. It will be set to the main JFrame.
		 * See methos main() of this file.
		 */
		menu = new JMenuBar();

		/*
		 * For schema and xml file operation.
		 */
		JMenu filemenu = new JMenu("File");
		loaditem = new JMenuItem("Load");
		newitem = new JMenuItem("New");
		openitem = new JMenuItem("Open");
		saveitem = new JMenuItem("Save");
		saveasitem = new JMenuItem("Save As");
		exititem = new JMenuItem("Exit");

		filemenu.add(loaditem);
		filemenu.addSeparator();
		filemenu.add(newitem);
		filemenu.add(openitem);
		filemenu.add(saveitem);
		filemenu.add(saveasitem);
		filemenu.addSeparator();
		filemenu.add(exititem);
		menu.add(filemenu);
		
		loaditem.addActionListener(this);
		newitem.addActionListener(this);
		openitem.addActionListener(this);
		saveitem.addActionListener(this);
		saveasitem.addActionListener(this);
		exititem.addActionListener(this);

		/*
		 * For validation.
		 */
/*
		JMenu validatemenu = new JMenu("Validation");
		xmlfileitem = new JMenuItem("Input XML file name");
		schemafileitem = new JMenuItem("Input Schema file name");
		validateitem = new JMenuItem("Validation");

		validatemenu.add(xmlfileitem);
		validatemenu.add(schemafileitem);
		validatemenu.add(validateitem);
		menu.add(validatemenu);

		xmlfileitem.addActionListener(this);
		schemafileitem.addActionListener(this);
		validateitem.addActionListener(this);
*/

		/*
		 * For help.
		 */
		JMenu helpmenu = new JMenu("Help");
		aboutitem = new JMenuItem("About HyperCast configuration editor");
		helpmenu.add(aboutitem);
		menu.add(helpmenu);
		aboutitem.addActionListener(this);


		/*
		 * Create a JPopupMenu instance. It is displayed 
		 * when right clicking on the jtree.
		 */
		popupmenu = new JPopupMenu();
		expanditem = new JMenuItem("Expand Subtree");
		collapseitem = new JMenuItem("Collapse Subtree");
		popupmenu.add(expanditem);
		popupmenu.add(collapseitem);
		expanditem.addActionListener(this);
		collapseitem.addActionListener(this);

		/*
		 * Make a nice border for this JPanel.
		 */
		EmptyBorder eb = new EmptyBorder(5,5,5,5);
		BevelBorder bb = new BevelBorder(BevelBorder.LOWERED);
		CompoundBorder cb = new CompoundBorder(eb,bb);
		this.setBorder(new CompoundBorder(cb,eb));

		/*
		 * Create a initial root node for the tree. It will be 
		 * replaced when actual tree is created.
		 */
		DefaultMutableTreeNode m_root = new  DefaultMutableTreeNode(null);
		
		/*
		 * Create a JTree instance.
		 */
		DefaultTreeModel m_model = new DefaultTreeModel (m_root);
		tree = new JTree (m_model);

		tree.addTreeSelectionListener(this);
		tree.addMouseListener(this);

		//tree.setEditable (true);
		tree.setVisible(false);

		/* 
		 * Build top-side view. Set an invisible tree in the treePanel
		 * when the treePanel is created. If we don't pass this tree 
		 * when initializing treePanel, no way is found to display
		 * a tree added later in the treePanel.
		 */
		treePanel = new JScrollPane(tree);

		/*
		 * Build bottomt-side view. It's layout is set to null, so absolute
		 * position for displaying Swing units is supported.
		 */
		editPanel = new JPanel();
		editPanel.setLayout(null);

		/*
		 * Build split-pane view which contains tree panel and edit panel.
		 */
		JSplitPane splitPane = 
			new JSplitPane( JSplitPane.VERTICAL_SPLIT,
						   treePanel,
						   editPanel );
		splitPane.setContinuousLayout( true );
		splitPane.setDividerLocation( FRAME_HEIGHT*2/3 );

		//Add split panel to this panel.
		this.setLayout(new BorderLayout());
		this.add(splitPane, BorderLayout.CENTER);

		JPanel infoPanel = new JPanel();
		infoPanel.setLayout(new BorderLayout());
		this.add(infoPanel, BorderLayout.SOUTH);

		/*
		 * Create a JTextField instance which is located at the bottom of the
		 * this JPanel object and is used to show the process and status information.
		 */		
		schemafileLoadInfoField = new JTextField("Schema file loaded:");
		schemafileLoadInfoField.setHorizontalAlignment(JTextField.CENTER);
		schemafileLoadInfoField.setEditable(false);
		infoPanel.add(schemafileLoadInfoField, BorderLayout.CENTER);

		/*
		 * Create a JTextField instance which is located at the bottom of the
		 * this JPanel object and is used to show the process and status information.
		 */		
		infoField = new JTextField("");
		infoField.setHorizontalAlignment(JTextField.CENTER);
		infoField.setEditable(false);
		infoPanel.add(infoField, BorderLayout.SOUTH);
		

		/*
		 * By default the schema file hypercast.xsd contained in hypercast.jar is loaded.
		 */
		loadDefaultSchemaFile();		 
	}


    public static void main(String args[])
    {
    	/*
    	 * Create a default configulation file without GUI
    	 */
    	if((args.length==1) && (args[0].equals("-printdefault"))){
    		HypercastConfigurator hc = new HypercastConfigurator();
    		hc.printDefaultConfigulationFile();
    		System.exit(0);
    	}
		JFrame frame = new JFrame ("HyperCast configuration editor");
		
		frame.addWindowListener (new WindowAdapter() {
				public void windowClosing (WindowEvent ignored) { System.exit (0); } });

		/*
		 * Create a new HypercastConfigurator instance, which is
		 * an implementation of JPanel.
		 */
		HypercastConfigurator hc = new HypercastConfigurator();

		ImageIcon iIcon = hc.createIcon("hypercast_logo.jpg");
		frame.setIconImage(iIcon.getImage());

		hc.initializeDisplay();
		
		frame.getContentPane().add (hc, BorderLayout.CENTER);
		frame.setJMenuBar(hc.menu);
		frame.setSize(FRAME_WIDTH,FRAME_HEIGHT);
		frame.setVisible (true);

    }

    /**
  	 * Create and save a default configulation file to a specific file
     */
    private void printDefaultConfigulationFile(){
		initializeDisplay();
		buildTreeFromSchemaFile();
		Document result = createDocument(tree);

		if (result == null)
		{
			System.err.println("Printing is failed.");
			return;
		}
		try
		{
			XmlUtil.writeXml(result, System.out);
		}
		catch (IOException error)
		{
			System.err.println("Printing is failed.");
			return;
		}    	
    }

	/*
	 * It cannot not be done in main(), because this.getClass()
	 * is not static method.
	 */
	/**
	 * Create a ImageIcon instance from a jar file for given
	 * file name.
	 */
	private ImageIcon createIcon(String iconFileName)
	{
		URL url = this.getClass().getResource(iconFileName);
		return new ImageIcon(url);
	}

	/**
	 * <p>The callback for ToggleButtons; sets the appropriate ones visible.</p>
	 *
	 * @param e ChangeEvent for the button
	 */
	public void stateChanged(ChangeEvent e) 
	{	
	}

	/**
	 * Create a ImageIcon instance from a jar file for given
	 * file name.
	 */
	private void loadDefaultSchemaFile()
	{
		InputStream fis = getClass().getResourceAsStream( "hypercast.xsd" );

		if (fis == null)
		{
			infoField.setText("No schema file "+schemaFileName+" is loaded.");
			return;
		}
				
		/*
		 * Open schema file.
		 */
		try
		{ 
			openedSchemaFile = XmlUtil.createDocument(fis, new String("hypercast.xsd"));
		}
		catch (final IOException ioe)
		{
			infoField.setText("Cannot load schema file "+schemaFileName+".");
		}
				
		if (openedSchemaFile == null)
		{
			infoField.setText("Poor formated schema file: hypercast.xsd");
			return;
		}

		schemaFileName = "hypercast.xsd";

		isTreeEmpty = true;

		if (selectedNode != null)
		{
			selectedNode.hidden();
		}
		selectedNode = null;

		schemafileLoadInfoField.setText("Schema file loaded: hypercast.xsd");
		infoField.setText("");
	}

	/**
	 * <p>Callback for menu items.</p>
	 *
	 * @param e ActionEvent object to determine which item called this.
	 */
	public void actionPerformed(ActionEvent e) 
	{
		JFileChooser filechooser = new JFileChooser();

		File file = null;

		if (e.getSource().equals(loaditem)) 
		{
			TypeBased_FileFilter filter = new TypeBased_FileFilter();
			filter.addFileType("xsd");
			filter.setDescription("Schema file");
			filechooser.setFileFilter(filter);
			filechooser.setCurrentDirectory(new File("."));

			int returnval = filechooser.showOpenDialog(this);

			/*
			 * if open was selected, fire up a file chooser and load the specified file.
			 */
			if (returnval == JFileChooser.APPROVE_OPTION) 
			{
				file = filechooser.getSelectedFile();

				String filename = file.getName();

				if (filename.toLowerCase().endsWith(".xsd"))
				{
					/*
					 * Open schema file.
					 */
					//System.out.println("Schema file: " + filename + " is loaded.");
					try
					{ 
						openedSchemaFile = XmlUtil.createDocument(file);
					}
					catch (final IOException ioe)
					{
						infoField.setText("Cannot load schema file "+file.getName()+".");
					}
					catch (final ParseException ioe)
					{
						infoField.setText("Failed in parsingschema  file "+file.getName()+".");
					}
					
					if (openedSchemaFile == null)
					{
						infoField.setText("Poor formated schema file: " + filename);
						return;
					}

					/*
					 * Store the reference to the selected schema file.
					 */
					SchemaFile = file;
					schemaFileName = filename;
					
					/*
					 * Remove all comments in the schema file.
					 */
					XmlUtil.removeComments(openedSchemaFile, openedSchemaFile);
					
					isTreeEmpty = true;
					
					if (selectedNode != null)
					{
						selectedNode.hidden();
					}
					selectedNode = null;

					//infoField.setText("Schema file "+filename+" is loaded.");
					schemafileLoadInfoField.setText("Schema file loaded: "+filename);
					infoField.setText("");

				}
				else
				{
					infoField.setText("Unknown file type!");
					return;
				}


			}
		}
		else if (e.getSource().equals(newitem)) 
		{
			if (openedSchemaFile == null)
			{
				loadDefaultSchemaFile();		 
			}

			if (buildTreeFromSchemaFile())
			{
				infoField.setText("The new property tree is built from schema file "+schemaFileName+".");
			}
		}
		else if (e.getSource().equals(openitem)) 
		{
			TypeBased_FileFilter filter = new TypeBased_FileFilter();
			filter.addFileType("xml");
			filter.setDescription("XML file");
			filechooser.setFileFilter(filter);
			filechooser.setCurrentDirectory(new File("."));

			int returnval = filechooser.showOpenDialog(this);

			/*
			 * if open was selected, fire up a file chooser and load the specified file.
			 */
			if (returnval == JFileChooser.APPROVE_OPTION) 
			{
				file = filechooser.getSelectedFile();

				//System.out.println("open file path is "+file.getAbsolutePath()+", name is "+file.getName());

				String filename = file.getName();

				infoField.setText("File "+filename+" is opened.");
				
				if (filename.toLowerCase().endsWith(".xml"))
				{
					if (openedSchemaFile == null)
					{
						loadDefaultSchemaFile();		 
					}

					if (isTreeEmpty)
					{
						if (!buildTreeFromSchemaFile())
							return;
					}

					/*
					 * Open xml configuration file.
					 */
					//System.out.println("open Xml file: "+filename);
					try
					{ 
						openedXMLFile = XmlUtil.createDocument(file);
					}
					catch (final IOException ioe)
					{
						infoField.setText("Cannot load XML file "+file.getName()+".");
					}
					catch (final ParseException ioe)
					{
						infoField.setText("Failed in parsing XML  file "+file.getName()+".");
					}

					if (openedXMLFile == null)
					{
						infoField.setText("Poor formated XML file: "+filename);
						return;
					}

					XMLFile = file;
										
					/*
					 * Remove all comments in the schema file.
					 */
					XmlUtil.removeComments(openedXMLFile, openedXMLFile);

					/*
					 * Load the attributes in the XML file to existing jtree.
					 */
					Element propertyElement = (Element) openedXMLFile.getChildNodes().item(0);

					DefaultMutableTreeNode root = (DefaultMutableTreeNode) tree.getModel().getRoot();
					
					/*
					 * Reset the values of all tree nodes to initial values.
					 */
					resetTreeNodeRecursively(tree, root);

//TODO: Replace the check "isCompatible" with standard validation call, but it is only supported in Java 1.5 or uppper.
					//if (!validateXMLFile(filename, schemaFileName))					
					if (!isCompatible(tree, root, propertyElement))
					{
						infoField.setText("The XML file " + filename + " is not validated with " + schemaFileName);
						return;
					}

					if (!loadXMLElementRecursively(tree, root, propertyElement))
					{
						openedXMLFile = null;
						XMLFile = null;

						infoField.setText("Cannot rebuild the property tree from XML file: " + filename + "!");
						/*
						 * Rebuild the tree from schema file, it may be damaged by the call
						 * loadXMLElementRecursively().
						 */
						buildTreeFromSchemaFile();
						
					}
					else
					{
						infoField.setText("The property tree is rebuilt from XML file: " + filename);
						root = (DefaultMutableTreeNode) tree.getModel().getRoot();
						((DefaultTreeModel) tree.getModel()).reload(root);
					}

					if (selectedNode != null)
					{
						selectedNode.hidden();
					}
					selectedNode = null;
				}
				else
				{
					infoField.setText("Unknown file type!");
					return;
				}

			}
		} 
		else if (e.getSource().equals(saveitem)) 
		{
			if (isTreeEmpty)
			{
				infoField.setText("Nothing to save.");
				return;
			}


			file = null;

			if (XMLFile != null)
			{
				file = XMLFile;
			}
			else
			{
				TypeBased_FileFilter filter = new TypeBased_FileFilter();
				filter.addFileType("xml");
				filter.setDescription("XML file");
				filechooser.setFileFilter(filter);
				filechooser.setCurrentDirectory(new File("."));

				int returnval = filechooser.showSaveDialog(this);
				if (returnval == JFileChooser.APPROVE_OPTION) 
				{
					file = filechooser.getSelectedFile();
				}
			}

			if (file != null)
			{
				String filename = file.getName();
				
				Document result = createDocument(tree);

				try
				{
					XmlUtil.writeXml(result, new FileOutputStream(file));

					infoField.setText("The file " + filename + " is saved.");
				}
				catch (IOException error)
				{
					infoField.setText("Saving to file " + filename + " is failed.");
				}
			}
		} 
		else if (e.getSource().equals(saveasitem)) 
		{
			if (isTreeEmpty)
			{
				infoField.setText("Nothing to save.");
				return;
			}

			TypeBased_FileFilter filter = new TypeBased_FileFilter();
			filter.addFileType("xml");
			filter.setDescription("XML file");
			filechooser.setFileFilter(filter);
			filechooser.setCurrentDirectory(new File("."));

			int returnval = filechooser.showSaveDialog(this);
			if (returnval == JFileChooser.APPROVE_OPTION) 
			{
				file = filechooser.getSelectedFile();

				String filename = file.getName();
				
				Document result = createDocument(tree);

				if (result == null)
				{
					infoField.setText("Saving to file " + filename + " is failed.");
					return;
				}

				try
				{
					XmlUtil.writeXml(result, new FileOutputStream(file));

					infoField.setText("The file " + filename + " is saved.");
				}
				catch (IOException error)
				{
					infoField.setText("Saving to file " + filename + " is failed.");
				}
			}

		}
		else if (e.getSource().equals(expanditem)) 
		{
			if (tree.getLastSelectedPathComponent() != null)
			{
				/*
				 * Get the selected tree node.
				 */
				DefaultMutableTreeNode node = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();
		
				/*
				 * Only for non-leaf tree node, display popup menu.
				 */
				if (((DefaultTreeModel) tree.getModel()).isLeaf(node))
				{
					return;
				}
				
				expandAll(tree, node);
			}
		} 
		else if (e.getSource().equals(collapseitem)) 
		{
			if (tree.getLastSelectedPathComponent() != null)
			{
				/*
				 * Get the selected tree node.
				 */
				DefaultMutableTreeNode node = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();
		
				/*
				 * Only for non-leaf tree node, display popup menu.
				 */
				if (((DefaultTreeModel) tree.getModel()).isLeaf(node))
				{
					return;
				}
				
				collapseAll(tree, node);
			}
		} 
		else if (e.getSource().equals(exititem)) 
		{
			System.exit(0);	
		}
		else 
		{
			infoField.setText("Unhandled Event: " +
							   e.getSource().getClass().getName());
		}

	}

	/**
	 * Build the property tree from schema file.
	 */
	private boolean validateXMLFile(String XMLFile, String schemaFile)
	{
		/*
		 * The code here is used to validate the opened XML file against the
		 * XML schema. It works with Java1.5.
		 */
/*
		  try {
			  SchemaFactory dbf = SchemaFactory.newInstance("http://www.w3.org/2001/XMLSchema");
			  //AbstractSchemaImpl s = (AbstractSchemaImpl)dbf.newSchema(new File("hypercast.xsd"));
			  AbstractSchemaImpl s = (AbstractSchemaImpl)dbf.newSchema(new File(schemaFile));
			  Validator v = s.newValidator();
			  v.validate(new StreamSource(XMLFile));
			  
			  //System.out.println("This document is well formated and validate against its schema!");
			  return true;
	
		  } catch (Exception exml)
		  {
			//System.out.println("This document is validate against its schema!");
		  	return false;
		  }
*/
		return true;
	}

	/**
	 * Build the property tree from schema file.
	 */
	private boolean buildTreeFromSchemaFile()
	{
		if (openedSchemaFile == null)
		{
			return false;
		}
		
		//String filename = SchemaFile.getName();

		/*
		 * Find the property Element in schema file. The way to find
		 * property Element is hypercast specific.
		 */
		Element propertyElement = findPropertyElement(openedSchemaFile);

		if (propertyElement == null)
		{
			infoField.setText("Poor formated schema file: " + schemaFileName);
			SchemaFile = null;
			openedSchemaFile = null;
			return false;						
		}


		if (debug)
			System.out.println("buildTreeFromSchemaFile: find property Element: "+propertyElement.getAttribute("name"));

		DefaultMutableTreeNode root = constructDispTreeFromSchemaFile(propertyElement);

		if (root == null)
		{
			infoField.setText("Cannot construct property tree from schema file: " + schemaFileName);
			SchemaFile = null;
			openedSchemaFile = null;
			return false;
		}

		/*
		 * Rebuild the tree and display it in collapse-all-node way.
		 */
		((DefaultTreeModel) tree.getModel()).setRoot(root);
		((DefaultTreeModel) tree.getModel()).reload(root);
		tree.setVisible(true);
		
		isTreeEmpty = false;
		
		return true;
	}


	/**
	 * Build the defaultvalue properties object from schema file.
	 */
	public Properties getDefaultProperties()
	{
		
		
		return null;
	}
	
	/*
	 * Below are a set of static tree node operation functions.
	 */

	public static void collapseAll(JTree tree)
	{
		Object root = tree.getModel().getRoot();

		if (root != null)
			collapseAll(tree, new TreePath(root));
	}

	public static void collapseAll(JTree tree, Object node)
	{
		if (node != null)
			collapseAll(tree, new TreePath(((DefaultTreeModel) tree.getModel()).getPathToRoot((DefaultMutableTreeNode) node)));
	}

	public static void collapseAll(JTree tree, TreePath path)
	{
		TreeModel data = tree.getModel();
 
		if (data.isLeaf(path.getLastPathComponent()))
			return;

		collapseAllImpl(tree, tree.getModel(), path);
	}

 
	private static void collapseAllImpl(JTree tree, TreeModel data, TreePath path)
	{
		Object node = path.getLastPathComponent();

		for (int count = data.getChildCount(node), i = 0; i < count; i++)
		{
			Object child = data.getChild(node, i);

			if (!data.isLeaf(child))
				collapseAllImpl(tree, data, path.pathByAddingChild(child));
		}

		// safe guard because JTree.collapsePath() will expand all pa-
		// rents if the path is already collapsed. Argh!
		if (tree.isExpanded(path))
			tree.collapsePath(path);
	}

	public static void expandAll(JTree tree)
	{
		Object root = tree.getModel().getRoot();

		if (root != null)
			expandAll(tree, new TreePath(root));
	}

	public static void expandAll(JTree tree, Object node)
	{
		if (node != null)
			expandAll(tree, new TreePath(((DefaultTreeModel) tree.getModel()).getPathToRoot((DefaultMutableTreeNode) node)));
	}


	public static void expandAll(JTree tree, TreePath path)
	{
		for (Iterator i = extremalPaths(tree.getModel(), path, new HashSet()).iterator(); i.hasNext();)
			tree.expandPath((TreePath)i.next());
	}

	/** The "extremal paths" of the tree model's subtree starting at
		path. The extremal paths are those paths that a) are non-leaves
		and b) have only leaf children, if any. It suffices to know
		these to know all non-leaf paths in the subtree, and those are
		the ones that matter for expansion (since the concept of expan-
		sion only applies to non-leaves).
		The extremal paths of a leaves is an empty collection.
		This method uses the usual collection filling idiom, i.e.
		clear and then fill the collection that it is given, and
		for convenience return it. The extremal paths are stored
		in the order in which they appear in pre-order in the
		tree model.
	*/
	public static Collection extremalPaths(TreeModel data, TreePath path, Collection result)
	{
		result.clear();

		if (data.isLeaf(path.getLastPathComponent()))
		{
			return result; // should really be forbidden (?)
		}
        
		extremalPathsImpl(data, path, result);

		return result;
	}

	private static void extremalPathsImpl(TreeModel data, TreePath path, Collection result)
	{
		Object node = path.getLastPathComponent();
        
		boolean hasNonLeafChildren = false;

		int count = data.getChildCount(node);
        
		for (int i = 0; i < count; i++)
			if (!data.isLeaf(data.getChild(node, i)))
				hasNonLeafChildren = true;

		if (!hasNonLeafChildren)
			result.add(path);
		else
		{
			for (int i = 0; i < count; i++)
			{
				Object child = data.getChild(node, i);
                
				if (!data.isLeaf(child))
					extremalPathsImpl(data, path.pathByAddingChild(child), result);
			}
		}
	}


	/**
	 * <p>The callback for tree node selection.</p>
	 *
	 * @param e TreeSelectionEvent event from the tree.
	 */
	public void valueChanged(TreeSelectionEvent e)
	{
		if (tree.getLastSelectedPathComponent() != null)
		{
			infoField.setText("Node selected in the Tree: " +
			 tree.getLastSelectedPathComponent().toString());
		}
		else
		{
			return;
		}

		/*
		 * Get the selected tree node.
		 */
		DefaultMutableTreeNode node = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();
		
		/*
		 * When a DefaultMutableTreeNode instance is created, by default it contains an non-null 
		 * UserObject.
		 */
		if ((node.getUserObject() == null) || (((DefaultTreeModel) tree.getModel()).isLeaf(node)) || 
			(!(node.getUserObject() instanceof dispAndEditor)))
		{
			return;
		}
		 
		dispAndEditor manageNode = (dispAndEditor) node.getUserObject();

		/*
		 * Hidden the Swing units bound to the last selected tree node and display the Swing units 
		 * for the new selected one.
		 */
		if (selectedNode != null)
		{
			selectedNode.hidden();
		}
		selectedNode = manageNode;
		
		manageNode.display();
	}

	/*
	 * Below are a set of mouse event handling methods required by
	 * MouseListener. We only handle mouseClick event. All other
	 * mouse events on the jtree are ignored.
	 */
	/**
	 * <p>The callback for mouse click event on a tree node.</p>
	 *
	 * @param e a MouseEvent event from the tree.
	 */
	public void mouseClicked(MouseEvent e)
	{
		if ((e.getButton() == MouseEvent.BUTTON3) && (tree.getLastSelectedPathComponent() != null))
		{
			/*
			 * Get the selected tree node.
			 */
			DefaultMutableTreeNode node = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();
		
			/*
			 * Only for non-leaf tree node, display popup menu.
			 */
			if (((DefaultTreeModel) tree.getModel()).isLeaf(node))
			{
				return;
			}

			popupmenu.show(e.getComponent(), e.getX(), e.getY());
		}
	}

	/**
	 * <p>The callback for mouse enter event on a tree node.</p>
	 * It is ignored.
	 *
	 * @param e a MouseEvent event from the tree.
	 */
	public void mouseEntered(MouseEvent e)
	{
	}

	/**
	 * <p>The callback for mouse exit event on a tree node.</p>
	 * It is ignored.
	 *
	 * @param e a MouseEvent event from the tree.
	 */
	public void mouseExited(MouseEvent e)
	{
	}

	/**
	 * <p>The callback for mouse press event on a tree node.</p>
	 * It is ignored.
	 *
	 * @param e a MouseEvent event from the tree.
	 */
	public void mousePressed(MouseEvent e)
	{
	}

	/**
	 * <p>The callback for mouse release event on a tree node.</p>
	 * It is ignored.
	 *
	 * @param e a MouseEvent event from the tree.
	 */
	public void mouseReleased(MouseEvent e)
	{
	}


	/**
	 * Create a XML DOM Document from a JTree instance.
	 * @param jtree a JTree instance
	 * @return	a DOM Document instance.
	 */
	private Document createDocument(JTree jtree)
	{
		DefaultMutableTreeNode treeroot = (DefaultMutableTreeNode) ((DefaultTreeModel) jtree.getModel()).getRoot();
		
		Document doc = XmlUtil.createDocument();
		
		Element xmltreeRoot = (Element) CreateSubXmlTreeFromJTree(doc, jtree, treeroot);
		
		doc.appendChild(xmltreeRoot);
		return doc;
	}
	
	/**
	 * Create a subtree in the XML DOM Document which is 
	 * equivalent to the subtree rooted at the given node
	 * in the JTree.
	 * 
	 * @param doc		a DOM Document instance used for creating new Nodes.
	 * @param jtree		a JTree instance.
	 * @param treeNode	a node in the JTree instance.
	 * @return	a Node which represents a subtree rooted at this Node.
	 */
	private Node CreateSubXmlTreeFromJTree(Document doc, JTree jtree, DefaultMutableTreeNode treeNode)
	{
		DefaultTreeModel treeModel = (DefaultTreeModel) jtree.getModel();
		
		Node xmlNode = null;
		
		/*
		 * Here we use treeNode.toString() as the name of XML node name
		 * because in jtree each tree node is assigned a used-defined object,
		 * in this case treeNode.toString() represents the identifier of
		 * the node in the tree.
		 */
		if (treeModel.isLeaf(treeNode))
		{
			DefaultMutableTreeNode parent = null;
			if (treeNode.getParent() != null)
			{
				parent = (DefaultMutableTreeNode) treeNode.getParent();
			}

			dispAndEditor manageObj = (dispAndEditor) parent.getUserObject();
			
			/*
			 * If a treenode is leaf node and its parent represents a simpletype
			 * attribute, create a Text node for this tree node.
			 */
			if ((manageObj != null) && (manageObj.isSimpleType()))
			{
				xmlNode = (Text) doc.createTextNode(treeNode.toString());
				
				//System.out.println("CreateSubXmlTreeFromJTree: text node for parent "+parent.toString()+" has value "+xmlNode.getNodeValue()+".");

				//if (xmlNode == null)
				//	System.out.println("CreateSubXmlTreeFromJTree: text node for parent "+parent.toString()+" is null");
				//else
				//	System.out.println("CreateSubXmlTreeFromJTree: text node for parent "+parent.toString()+" is "+xmlNode.toString()+".");
				
			}
			/*
			 * If a treenode is leaf node and but its parent doesn't represent a 
			 * simpletype attribute, create an Element for it.
			 */
			else
			{
				xmlNode = (Element) doc.createElement(treeNode.toString());
			}
		}
		else
		{
			xmlNode = (Element) doc.createElement(treeNode.toString());

			/*
			 * Recursively create XMl subtree for each jtree child and add them
			 * to this XML node.
			 */
			for (int i = 0; i<treeNode.getChildCount(); i++)
			{
				DefaultMutableTreeNode child = (DefaultMutableTreeNode) treeNode.getChildAt(i);
				Node xmlChild = CreateSubXmlTreeFromJTree(doc, jtree, child);
				xmlNode.appendChild(xmlChild);
			}
		}

		return xmlNode;
	}	
	
	/**
	 * Construct a display tree node for given Element.
	 * 
	 * @param xmlElement	The Element corresponding to which the tree node is constructed.
	 * @return	A tree node.
	 */
	private DefaultMutableTreeNode constructDispTreeFromSchemaFile(Element xmlElement)
	{
		/*
		 * Handle complexType Element.
		 */
		if (isComplexElement(xmlElement))
		{
			if (debug)
				System.out.println("constructDispTreeFromSchemaFile: process compleType Element: "+xmlElement.getAttribute("name"));

			return handleComplexElement(xmlElement);
		}
		/*
		 * Otherwise it is simpleType Element
		 */
		else
		{
			if (debug)
				System.out.println("constructDispTreeFromSchemaFile: process simpleType Element: "+xmlElement.getAttribute("name"));

			return handleSimpleElement(xmlElement);
		}
	}


	/**
	 * Build a diaplay subtree rooted at the node which represents the given 
	 * Element e.
	 */
	private DefaultMutableTreeNode handleComplexElement(Element e)
	{
		String name = e.getAttribute("name");

		if (debug)
			System.out.println("This complexType node is "+name);

		/*
		 * Create tree node for this Element.
		 */		
		DefaultMutableTreeNode thisNode = new DefaultMutableTreeNode(name);

		/*
		 * Get all children of this Element in the schema file.
		 */
		Vector children = getChildrenOfComplexElement(e);

		DefaultMutableTreeNode childNode = null;
		
		/*
		 * If the given Element has choice on its children, only one
		 * child becomes display tree child and all other children become
		 * candidate display tree children. The dispAndEditor object 
		 * contained in this tree node stores the list of all children,
		 * including display child and candidate children.
		 */
		if (hasChoiceOnChildren(e))
		{
			if (debug)
				System.out.println("It has choice on its children");

			String defaultValue = e.getAttribute("default");

			/*
			 * Vector stores the children in JTree, including tree children and
			 * candidate tree children.
			 */
			Vector treeChildren = new Vector();
		
			/*
			 * Vector stores the names of children, including tree children and
			 * candidate tree children.
			 */
			Vector childrenNames = new Vector();

			DefaultMutableTreeNode selectedChild = null;

			/*
			 * This loop is to process this node's children recursively.
			 */
			for (int i=0; i<children.size(); i++)
			{
				Element child = (Element) children.elementAt(i);
				
				if (child == null)
				{
					continue;
				}

				/*
				 * Create the tree node for its child.
				 */
				childNode = constructDispTreeFromSchemaFile(child);
				
				if (childNode == null)
				{
					//continue;
					return null;
				}
				
				/*
				 * If no default value is defined, choose the first child as default child. Otherwise
				 * the child which match default value is chosen as default child.
				 */
				if ((((defaultValue == null) || (defaultValue.equals(""))) && (i == 0)) ||
					((defaultValue != null) && (!defaultValue.equals("")) && (defaultValue.equals(child.getAttribute("name")))))
				{
					//System.out.println("Choice: Add a display child: "+child.getAttribute("name"));
					
					/*
					 * Set the selected Node.
					 */
					selectedChild = childNode;
				}
				
				treeChildren.addElement(childNode);
				
				childrenNames.addElement(child.getAttribute("name"));
				
			}

			if (selectedChild != null)
			{
				thisNode.add(selectedChild);
			}

			/*
			 * Create a editor for this tree node.
			 */
			ChoiceComponent cc = new ChoiceComponent(this.editPanel, this.tree, thisNode, name, treeChildren, childrenNames, defaultValue);

			/*
			 * Stores a reference to the schema Element in the dispAndEditor object.
			 */
			cc.setSchemaElement(e);

			cc.setHyperCastConf(this);
			
			/*
			 * Set reference to user-defined manage object in this tree node.
			 */
			thisNode.setUserObject(cc);
		}
		/*
		 * Otherwise, all children in the schema document become display tree children.
		 */
		else
		{
			//System.out.println("It doesn't have choice on its children");

			for (int i=0; i<children.size(); i++)
			{
				Element child = (Element) children.elementAt(i);

				//System.out.println("Element child has name: "+child.getAttribute("name"));
				
				childNode = constructDispTreeFromSchemaFile(child);
				
				if (childNode == null)
				{
					//continue;
					return null;
				}

				/*
				 * If a Element's maximum occur times is larger than 1, it becomes a 
				 * dependent, i.e. its real occur times in the jtree is specified by 
				 * another tree node's value.
				 */
				String maxoccurs = child.getAttribute("maxOccurs");
				if ((maxoccurs.equals("unbounded")) || 
					((!maxoccurs.equals("")) && ((new Integer (maxoccurs)).intValue() > 1)))
				{
					dispAndEditor manageObj = (dispAndEditor) childNode.getUserObject();
					manageObj.setDependentFlag(true);
				}


				//System.out.println("Sequence: Add a display child: "+child.getAttribute("name"));
					
				thisNode.add(childNode);
			}

			/*
			 * Create a Label unit for this tree node.
			 */
			LabelComponent lc = new LabelComponent(this.editPanel, this.tree, thisNode, name);

			/*
			 * Stores a reference to the schema Element in the dispAndEditor object.
			 */
			lc.setSchemaElement(e);

			lc.setHyperCastConf(this);
			
			/*
			 * Set reference to user-defined manage object in this tree node.
			 */
			thisNode.setUserObject(lc);

			/*
			 * Handle dependence between children of thisNode. In hypercast 
			 * configuration document, some attributes can occur multiple time
			 * which is controled by another attribute. A example is HTTPServer,
			 * the number of HTTPServer is specified by HTTPServerNum.
			 */
			 handleDependence(this.editPanel, this.tree, thisNode);

		}

		/*
		 * Get help information for the operation on this tree node.
		 */
		String Tip = getTip(e);	
		//dc.setTip(Tip);

		return thisNode;
	}


	/**
	 * Build the diaplay tree node which represents the given simpleType Element e.
	 */
	private DefaultMutableTreeNode handleSimpleElement(Element e)
	{
		String name = e.getAttribute("name");

		if (debug)
			System.out.println("This simpleType node is "+name);

		String defaultValue = e.getAttribute("default");

		/*
		 * Create tree node for this Element.
		 */		
		DefaultMutableTreeNode thisNode = new DefaultMutableTreeNode (name);
		
		/*
		 * Get help information for the operation on this tree node.
		 */
		String Tip = getTip(e);	
		//dc.setTip(Tip);
		
		/*
		 * Each simpleType Element must has a child which defines
		 * restrion for this Element. Find it and take different 
		 * action based on it.
		 */
		Element restritionElement = getRestrictionElement(e);
	
		/*
		 * If no restrition Element is found for this Element, 
		 * there is no operation on the corresponding tree node.
		 */
		if (restritionElement == null)
		{
			/*
			 * Create a Label unit for this tree node.
			 */
			LabelComponent lc = new LabelComponent(this.editPanel, this.tree, thisNode, name);

			thisNode.setUserObject(lc);

			return thisNode;
		}
	
		/*
		 * If restrition Element is found, it contains an attribute 
		 * "base" which defines the basic data type of this simpleType
		 * Element.
		 */
		String base = restritionElement.getAttribute("base");
		
		NodeList nl = restritionElement.getChildNodes();

		/*
		 * If the given Element defines restriction, create a leaf tree node, which 
		 * shows the default value and add it to this tree node. The created leaf node 
		 * is changed when the value for this tree node is reset.
		 */
		DefaultMutableTreeNode valueNode = new DefaultMutableTreeNode(defaultValue);
		
		dispAndEditor de = null;
		if (nl.getLength() == 0)
		{
			de = handleSimpleRestrition(this.editPanel, this.tree, valueNode, name, base, defaultValue);
		}
		else
		{
			de = handleComplexRestriction(this.editPanel, this.tree, valueNode, name, base, nl, defaultValue);
		}

		if (de == null)
		{
			return null;
		}

		/*
		 * Stores a reference to the schema Element in the dispAndEditor object.
		 */
		de.setSchemaElement(e);

		de.setHyperCastConf(this);

		/*
		 * Set the flag which shows this SWing unit represents a simpleType Element.
		 */
		de.setIsSimpleType(true);

		/*
		 * Two nodes share a user-defined object. In this object 
		 * toString() method needs to distinguish them.
		 */
		thisNode.setUserObject(de);
		
		leafNodeDisplayHelper lp = new leafNodeDisplayHelper(de);
		de.setDispHelper(lp);
		valueNode.setUserObject(lp);
	
		thisNode.add(valueNode);

		return thisNode;	
	}



	/**
	 * Check if this Element has a sequence of children.
	 */
	public static boolean hasChoiceOnChildren(Element e)
	{
		NodeList firstnl = e.getChildNodes();
		if (firstnl.getLength() != 1)
		{
			return false;
		}
		
		NodeList secnl = firstnl.item(0).getChildNodes();
		if (secnl.getLength() != 1)
		{
			return false;
		}
		else
		{
			String name = secnl.item(0).getNodeName();
			if (name.indexOf("choice") >= 0)
			{
				return true;
			}
			else
			{
				return false;
			}
		}
	}
	
	/**
	 * Check if this Element has a "complex" child.
	 */
	public static boolean isComplexElement(Element e)
	{
		NodeList firstnl = e.getChildNodes();

		for (int i=0; i<firstnl.getLength(); i++)
		{
			Node node = firstnl.item(i);
			String name = firstnl.item(i).getNodeName();
		
			if (name.indexOf("annotation") >= 0)
			{
				continue;
			}
			else if (name.indexOf("complexType") >= 0)
			{
				return true;
			}
			else
			{
				return false;
			}
		}
		return false;
	}
	

	/**
	 * Get children of this Element.
	 */
	public static Vector getChildrenOfComplexElement(Element e)
	{
		Vector p = new Vector();
		
		NodeList firstnl = e.getChildNodes();

		Node complexTypeChild = null;
		for (int i=0; i<firstnl.getLength(); i++)
		{
			Node node = firstnl.item(i);
			String name = firstnl.item(i).getNodeName();
				
			if (name.indexOf("annotation") >= 0)
			{
				continue;
			}
			else if (name.indexOf("complexType") >= 0)
			{
				complexTypeChild = node;
				break;
			}
		}

		if (complexTypeChild == null)
		{
			return p;
		}

		NodeList secnl = complexTypeChild.getChildNodes();
		NodeList thirdnl = null;

		if (secnl.getLength() == 0)
		{
			return p;
		}
		else if ((secnl.getLength() == 1) &&
				  ((secnl.item(0).getNodeName().equals("xs:sequence"))
					|| (secnl.item(0).getNodeName().equals("xs:choice"))))
		{
			thirdnl = secnl.item(0).getChildNodes();
		}
		else
		{
			thirdnl = secnl;
		}

		/*
		 * It is the Element with name "sequence" or "choice".
		 */
		for (int i=0; i<thirdnl.getLength(); i++)
		{
			Node node = thirdnl.item(i);

			/*
			 * There are some comments line in the file which become Node
			 * not Element. Skip all Nodes which are not instances of Element.
			 */
			if (!(node instanceof Element))
			{
				continue;
			}

			p.addElement((Element) node);
		}

		return p;
	}
	

	/**
	 * Find the property Element in the given Document. This Element is
	 * the starting point for building displaying tree.
	 * This method assume the schema file has fixed format to reach the 
	 * property Element.
	 * 
	 * @param doc	A Document instance.
	 * @return	the Element which has attribute "name = ROOT_ATTRIBUTE_ELEMENT_NAME".
	 */
	public static Element findPropertyElement(Document doc)
	{
		return findPropertyElement(doc, ROOT_ATTRIBUTE_ELEMENT_NAME);
	}


	/**
	 * Find the property Element in the given Document. This Element is
	 * the starting point for building displaying tree.
	 * This method assume the schema file has fixed format to reach the 
	 * property Element.
	 * 
	 * @param doc	A Document instance.
	 * @param rootElementName	The Element name of the root for all attributes.
	 * 
	 * @return	the Element which has attribute "name = rootElemenName".
	 */
	public static Element findPropertyElement(Document doc, String rootElementName)
	{
		Element propertyElement = null;

		//get document's children
		NodeList nl = doc.getChildNodes();

		Element schemaElement = null;

		/*
		 * Find the Element whose name contains "schema".
		 */
		for (int i=0; i<nl.getLength(); i++)
		{
			Node node = nl.item(i);
			//System.out.println("node with name "+node.getNodeName());

			if (node.getNodeName().indexOf("schema") >= 0)
			{
				schemaElement = (Element) node;
				break;
			}
		}

		if (schemaElement == null)
		{
			//throw new IllegalArgumentException("Poorly structured schame file!");
			System.err.println("Poorly structured schame file!");
			return null;
		}

		//get schema Element's children
		nl = schemaElement.getChildNodes();

		for (int i=0; i<nl.getLength(); i++)
		{
			Node node = nl.item(i);
			//System.out.println("element with name "+node.getNodeName());
			
			/*
			 * There are some comments line in the file which become Node
			 * not Element. Skip all Nodes which are not instances of Element.
			 */
			if (!(node instanceof Element))
			{
				continue;
			}

			String attrValue =((Element) node).getAttribute("name");

			if ((attrValue != null) && (attrValue.equals(rootElementName)))
			{
				propertyElement = (Element) node;
				break;
			}
		}

		return propertyElement;
	}


	/**
	 * Get the tip contents for this Element. It is used as help
	 * information when editing the tree node which represents this 
	 * Element.
	 * 
	 * @param e	Element in the configuration document.
	 * @return	the tip information.
	 */
	private String getTip(Element e)
	{
		NodeList nl = e.getChildNodes();

		Node annotationElement = null;

		/*
		 * Find a child which has name "annotation".
		 */
		for (int i=0; i<nl.getLength(); i++)
		{
			Node node = nl.item(i);

			if (node.getNodeName().indexOf("annotation") >= 0)
			{
				annotationElement = node;
				break;
			}
		}
		
		if ((annotationElement == null) || (annotationElement.getChildNodes().getLength() != 1))
		{
			return "";
		}
		else
		{
			Node tipNode = annotationElement.getChildNodes().item(0);
			if (tipNode.getNodeType() != Node.TEXT_NODE)
			{
				return "";
			}
			else
			{
				return tipNode.getNodeValue();
			}
		}
	}

	/**
	 * Get the Element which defines restrition for given simpleType Element.
	 * 
	 * @param e	Element in the configuration document.
	 * @return	the restrition definition Element.
	 */
	public static Element getRestrictionElement(Element e)
	{
		NodeList nl = e.getChildNodes();

		Node simpleTypeElement = null;

		/*
		 * Find a child which has name "simpleType".
		 */
		for (int i=0; i<nl.getLength(); i++)
		{
			Node node = nl.item(i);

			if (node.getNodeName().indexOf("simpleType") >= 0)
			{
				simpleTypeElement = node;
				break;
			}
		}
		
		/*
		 * simpleTypeElement's solo child is restrition Element.
		 */
		if ((simpleTypeElement == null) || (simpleTypeElement.getChildNodes().getLength() != 1))
		{
			return null;
		}

		Element restritionElement = (Element) simpleTypeElement.getChildNodes().item(0);
		if (restritionElement.getNodeName().indexOf("restriction") < 0)
		{
			return null;
		}
		else
		{
			return restritionElement;
		}
	}


	/**
	 * Create a dispAndEditor which dislays and edits the value of a configuration
	 * Element.
	 * 
	 * @param jp	A JPanel instance for editing a tree node.
	 * @param jt	A JTree instance.
	 * @param tn	The tree node to which this Swing unit is bound.
	 * @param name	the label name of this Swing unit.
	 * @param base 	data type in string format.
	 * @param defaultvalue defeult value specified in the schema document and is used
	 * to initialize the displaying of Swing unit.
	 * 
	 * @return	a dispAndEditor instance.
	 */
	private dispAndEditor handleSimpleRestrition(JPanel jp, JTree jt, DefaultMutableTreeNode tn, 
										String name, String base, String defaultvalue)
	{
		dispAndEditor de = null;
		
		if (base.indexOf("string") >= 0)
		{
			de = new TextFieldComponent(jp, jt, tn, name, defaultvalue);
		}
		else
		{
			de = new GroupComponent(jp, jt, tn, name, defaultvalue);
		}

		return de;
	}



	/**
	 * Create a dispAndEditor which dislays and edits the value of a configuration
	 * Element.
	 * 
	 * @param jp	A JPanel instance for editing a tree node.
	 * @param jt	A JTree instance.
	 * @param tn	The tree node to which this Swing unit is bound.
	 * @param name	the label name of this Swing unit.
	 * @param base 	data type in string format.
	 * @param nl	list of restrition Element's children.
	 * @param defaultValue defeult value specified in the schema document and is used
	 * to initialize the displaying of Swing unit.
	 * 
	 * @return	a dispAndEditor instance.
	 */
	private dispAndEditor handleComplexRestriction(JPanel jp, JTree jt, DefaultMutableTreeNode tn,
										String name, String base, NodeList nl, String defaultValue)
	{
		dispAndEditor de = null;
		
		if (base.indexOf("integer") >= 0)
		{
			String maxBound="", minBound="";
			
			for (int i = 0; i< nl.getLength(); i++) 
			{
				Element minormax = (Element) nl.item(i);
				if(minormax.getNodeName().indexOf("minInclusive") >= 0)
				{
					minBound = minormax.getAttribute("value");
				}
				if(minormax.getNodeName().indexOf("maxInclusive") >= 0)
				{
					maxBound = minormax.getAttribute("value");
				}
			}
			
			if (debug)
				System.out.println("handleComplexRestriction: Create SpinnerComponent for "+name);

			de = new SpinnerComponent(jp, jt, tn, name, defaultValue, minBound, maxBound);
		}
		else if (base.indexOf("string") >= 0)
		{
			Vector valueList = new Vector();
			
			for (int i = 0; i< nl.getLength(); i++) 
			{
				Element enuElement = (Element) nl.item(i);
				String enuVal = enuElement.getAttribute("value");
				if (!enuVal.equals(""))
				{
					valueList.addElement(enuVal);

					if (debug)
						System.out.println("Add new value "+enuVal);
				}
			}
			
			if (debug)
				System.out.println("handleComplexRestriction: Create ComboxComponent for "+name);

			de = new ComboxComponent(jp, jt, tn, name, valueList, defaultValue);
		}

		return de;
	}


	/**
	 * Handle dependence between the children of the given tree node. It is called 
	 * when loading a schema file.
	 * 
	 * @param jp		A JPanel instance for editing a tree node.
	 * @param jtree		JTree instance which contains the given node.
	 * @param node		A tree node.
	 */
	private void handleDependence(JPanel jp, JTree jtree, DefaultMutableTreeNode node)
	{
		HashMap dependentNodeNames = new HashMap();
		
		TreeModel model = jtree.getModel();
 
		for (int i = 0; i < model.getChildCount(node); i++)
		{
			DefaultMutableTreeNode child = (DefaultMutableTreeNode) model.getChild(node, i);
			dispAndEditor manageObj = (dispAndEditor) child.getUserObject();

			if (manageObj.isDependent())
			{
				dependentNodeNames.put(manageObj.getName(), child);
			}
		}
		
		/*
		 * For each dependent node, clone it so that the number of same name children
		 * equals the value specified in the node which specify the occur number for dependent.
		 */
		for (Iterator it=dependentNodeNames.entrySet().iterator(); it.hasNext(); )
		{ 
			Map.Entry entry = (Map.Entry)it.next(); 
			String dependentName = (String) entry.getKey(); 

			/*
			 * This is hypercast specific part. An alternative of handling dependence is to
			 * define dependence in some configuration file.
			 */
			DefaultMutableTreeNode controlNode = findChildWithName(jtree, node, dependentName+"Num");
			
			if (controlNode == null)
			{
				continue;
			}
			
			String valueStr = getValueString(controlNode);
			if (valueStr.equals(""))
			{
				continue;
			}

			dispAndEditor controlManageObj = (dispAndEditor) controlNode.getUserObject();
			controlManageObj.setDependentStr(dependentName);
			
			int number = (new Integer (valueStr)).intValue();
			
			int dependentIndex = findIndexWithName(jtree, node, dependentName);
			
			//System.out.println("add new dependent: "+dependentName);

			DefaultMutableTreeNode dependentNode = (DefaultMutableTreeNode) entry.getValue();

			/*
			 * Clone new tree (dependent) nodes, the number of dependent nodes is specified by
			 * the value of control node.
			 */
			dispAndEditor depmanageObj = (dispAndEditor) dependentNode.getUserObject();
			Element e = depmanageObj.getSchemaElement();
			for (int j = 0; j < number-1; j++)
			{
				//DefaultMutableTreeNode newDependent = cloneSubtree(jp, jtree, dependentNode);

				DefaultMutableTreeNode newDependent = constructDispTreeFromSchemaFile(e);
				((dispAndEditor) newDependent.getUserObject()).setDependentFlag(true);

				node.insert(newDependent, dependentIndex+1);
				dependentIndex++;
			}
		}

		/*
		 * Tell the jtree about the changes on the tree nodes.
		 */		
		//((DefaultTreeModel) jtree.getModel()).reload(node);
	}

	/**
	 * Update the number and displaying of dependent nodes (they have the same name
	 * as the given parameter) in the jtree.
	 * 
	 * @param jp	A JPanel instance for editing a tree node.
	 * @param jtree			JTree instance which contains the given node.
	 * @param parentNode	The parent of the dependent nodes.
	 * @param dependentNum	The new number of dependent nodes.
	 * @param dependentName	The name of dependent nodes.
	 */
	public void updateDependent(JPanel jp, JTree jtree, DefaultMutableTreeNode parentNode, int dependentNum, String dependentName)
	{
		TreeModel model = jtree.getModel();
 
		//System.out.println("updateDependent: parent node nameis "+ parentNode.toString()+", it has "+model.getChildCount(parentNode)+" children.");

		Vector dependentNodeIndexVec = findChildrenWithName(jtree, parentNode, dependentName);
		
		int oldNumber = dependentNodeIndexVec.size();

		//System.out.println("updateDependent: oldnumber is "+ oldNumber+", new number is "+dependentNum);

		if ((oldNumber == 0) || (dependentNum == 0) || (oldNumber == dependentNum))
		{
			return;
		}
		/*
		 * If the number of dependents decreases, remove some dependents from the jtree.
		 */
		else if (oldNumber > dependentNum)
		{
			for (int i=0; i<oldNumber-dependentNum; i++)
			{
				/*
				 * Remove dependents in reverse order, so that index for the removed nodes are correct.
				 */
				parentNode.remove(((Integer) dependentNodeIndexVec.elementAt(oldNumber-1-i)).intValue());
			}
		}
		/*
		 * If the number of dependents increases, add new dependents in the jtree.
		 */
		else 
		{
			DefaultMutableTreeNode tempchild = (DefaultMutableTreeNode) model.getChild(parentNode, ((Integer) dependentNodeIndexVec.elementAt(0)).intValue());
			dispAndEditor depmanageObj = (dispAndEditor) tempchild.getUserObject();
			Element e = depmanageObj.getSchemaElement();

			int lastIndex = ((Integer) dependentNodeIndexVec.elementAt(dependentNodeIndexVec.size()-1)).intValue();
			
			for (int j = oldNumber; j < dependentNum; j++)
			{
				/*
				 * Clone a new tree node and add it as child.
				 */

				//DefaultMutableTreeNode newDependent = cloneSubtree(jp, jtree, tempchild);
				DefaultMutableTreeNode newDependent = constructDispTreeFromSchemaFile(e);

				((dispAndEditor) newDependent.getUserObject()).setDependentFlag(true);
				
				parentNode.insert(newDependent, lastIndex+1);
				lastIndex++;
			}
		}

		//System.out.println("updateDependent: parent node now has "+model.getChildCount(parentNode)+" children.");
		/*
		 * Tell the jtree about the changes on the tree nodes.
		 */		
		((DefaultTreeModel) jtree.getModel()).reload(parentNode);
	}


	/**
	 * Update the number and displaying of child nodes which are dependents under 
	 * given tree node in the jtree.
	 * 
	 * @param jp	A JPanel instance for editing a tree node.
	 * @param jtree			JTree instance which contains the given node.
	 * @param parentNode	The parent of the dependent nodes.
	 */
	public void updateDependent(JPanel jp, JTree jtree, DefaultMutableTreeNode parentNode)
	{
		TreeModel model = jtree.getModel();
 
		for (int i = 0; i < model.getChildCount(parentNode); i++)
		{
			DefaultMutableTreeNode child = (DefaultMutableTreeNode) model.getChild(parentNode, i);
			dispAndEditor manageObj = (dispAndEditor) child.getUserObject();
			
			if (manageObj.hasDependents())
			{
				String dependentName = manageObj.getDependentStr();
				int newNumber = (new Integer (manageObj.getValueString())).intValue();

				updateDependent(jp, jtree, parentNode, newNumber, dependentName);
			}
		}
	}


	/**
	 * Get the child of given tree node which has the given name.
	 * 
	 * @param jtree		JTree instance which contains the given node.
	 * @param node		A tree node.
	 * @param childName	The name of the child to be found.
	 * @return	the child which has the given name.
	 */
	private static DefaultMutableTreeNode findChildWithName(JTree jtree, DefaultMutableTreeNode node, String childName)
	{
		TreeModel model = jtree.getModel();
 
		for (int i = 0; i < model.getChildCount(node); i++)
		{
			DefaultMutableTreeNode child = (DefaultMutableTreeNode) model.getChild(node, i);
			dispAndEditor manageObj = (dispAndEditor) child.getUserObject();
			if (manageObj.getName().equals(childName))
			{
				return child;
			}
		}
		
		return null;
	}

	/**
	 * Get the index of a child of given tree node which has the given name.
	 * 
	 * @param jtree		JTree instance which contains the given node.
	 * @param node		A tree node.
	 * @param childName	The name of the child to be found.
	 * @return	the index of the node which has the given name.
	 */
	private int findIndexWithName(JTree jtree, DefaultMutableTreeNode node, String childName)
	{
		TreeModel model = jtree.getModel();
 
		for (int i = 0; i < model.getChildCount(node); i++)
		{
			DefaultMutableTreeNode child = (DefaultMutableTreeNode) model.getChild(node, i);
			dispAndEditor manageObj = (dispAndEditor) child.getUserObject();
			if (manageObj.getName().equals(childName))
			{
				return i;
			}
		}
		
		return -1;
	}

	/**
	 * Get the index list of children of the given tree node which has the given name.
	 * 
	 * @param jtree		JTree instance which contains the given node.
	 * @param node		A tree node.
	 * @param childName	The name of the child to be found.
	 * @return	the index vector of children which has the given name.
	 */
	private static Vector findChildrenWithName(JTree jtree, DefaultMutableTreeNode node, String childName)
	{
		Vector dependentNodeIndexVec = new Vector();
		
		TreeModel model = jtree.getModel();
 
		for (int i = 0; i < model.getChildCount(node); i++)
		{
			DefaultMutableTreeNode child = (DefaultMutableTreeNode) model.getChild(node, i);
			dispAndEditor manageObj = (dispAndEditor) child.getUserObject();

			if (manageObj.getName().equals(childName))
			{
				dependentNodeIndexVec.addElement(new Integer(i));
			}
		}
		
		return dependentNodeIndexVec;
	}

	/**
	 * Get the index list of children of the given Node in the XMl doeument 
	 * which has the given name.
	 * 
	 * @param ele	A Node in the XML Document.
	 * @param name	The name of the child to be found.
	 * @return	the index vector of children which has the given name.
	 */
	private static Vector findXMLChildrenWithName(Node ele, String name)
	{
		Vector dependentNodeIndexVec = new Vector();
 
 		NodeList nl = ele.getChildNodes();
 
		for (int i = 0; i < nl.getLength(); i++)
		{
			Node child = nl.item(i);
			if (child.getNodeName().equals(name))
			{
				dependentNodeIndexVec.addElement(new Integer(i));
			}
		}
		
		return dependentNodeIndexVec;
	}

	/**
	 * Get the value string specified for the given node. 
	 * 
	 * @param node		A tree node.
	 * @return	the value of the given node.
	 */
	private static String getValueString(DefaultMutableTreeNode node)
	{
		dispAndEditor manageObj = (dispAndEditor) node.getUserObject();
		return manageObj.getValueString();
	}

	/**
	 * Clone a tree node and the subtree rooted at it. The Swing unit contained 
	 * in each tree node is also cloned and set to corresponding tree node.
	 * 
	 * @param pane	A JPanel instance for editing a tree node.
	 * @param jtree		JTree instance which contains the given node.
	 * @param node		A tree node.
	 * @return	the cloned tree node which represents a subtree.
	 */
	private static DefaultMutableTreeNode cloneSubtree(JPanel pane, JTree jtree, DefaultMutableTreeNode node)
	{
		DefaultMutableTreeNode newNode = new DefaultMutableTreeNode(node.toString());
		
		dispAndEditor manageObj = (dispAndEditor) node.getUserObject();
		
		dispAndEditor newManageObj = cloneSwingUnit(pane, jtree, manageObj, newNode);
		
		newNode.setUserObject(newManageObj);

		if (manageObj.isSimpleType())
		{
			/*
			 * Set the flag which shows this SWing unit represents a simpleType Element.
			 */
			newManageObj.setIsSimpleType(true);

			DefaultMutableTreeNode valueNode = new DefaultMutableTreeNode(manageObj.getDefaultValueString());

			leafNodeDisplayHelper lp = new leafNodeDisplayHelper(newManageObj);
			newManageObj.setDispHelper(lp);
			valueNode.setUserObject(lp);
	
			newNode.add(valueNode);
		}
		else
		{
			if (manageObj instanceof ChoiceComponent)
			{
				Vector childrenVec = ((ChoiceComponent) manageObj).getChildren();

				Vector newChildrenVec = new Vector();
				
				for (int i = 0; i < childrenVec.size(); i++)
				{
					DefaultMutableTreeNode child = (DefaultMutableTreeNode) childrenVec.elementAt(i);
				
					DefaultMutableTreeNode newChild = cloneSubtree(pane, jtree, child);
				
					newChildrenVec.addElement(newChild);
					
					String defaultStr = ((dispAndEditor) node.getUserObject()).getDefaultValueString();
					if (defaultStr.equals(((dispAndEditor) newChild.getUserObject()).getName()))
					{
						newNode.add(newChild);
					}
				}
				
				((ChoiceComponent) newManageObj).setChildren(newChildrenVec);
				
			}
			else
			{
				if (manageObj.isDependent())
				{
					newManageObj.setDependentFlag(true);
				}
				else if (manageObj.hasDependents())
				{
					newManageObj.setDependentStr(manageObj.getDependentStr());
				}
				
				TreeModel model = jtree.getModel();
 
				for (int i = 0; i < model.getChildCount(node); i++)
				{
					DefaultMutableTreeNode child = (DefaultMutableTreeNode) model.getChild(node, i);
				
					DefaultMutableTreeNode newChild = cloneSubtree(pane, jtree, child);
				
					newNode.add(newChild);
				}
			}
		}

		return newNode;
	}

	/**
	 * Clone a Swing unit.
	 * 
	 * @param pane	A JPanel instance for editing a tree node.
	 * @param jt		JTree instance which contains the given node.
	 * @param dispUnit	A Swing unit.
	 * @param node		A tree node.
	 * 
	 * @return	the cloned Swing unit.
	 */
	private static dispAndEditor cloneSwingUnit(JPanel pane, JTree jt, dispAndEditor dispUnit, DefaultMutableTreeNode node)
	{
		dispAndEditor newDispUnit = null;

		if (dispUnit instanceof ChoiceComponent)
		{
			newDispUnit = new ChoiceComponent(pane, jt, node, dispUnit.getName(), ((ChoiceComponent) dispUnit).getChildren(),
								((ChoiceComponent) dispUnit).getChildrenNames(), dispUnit.getDefaultValueString());
		}
		else if (dispUnit instanceof TextFieldComponent)
		{
			newDispUnit = new TextFieldComponent(pane, jt, node, dispUnit.getName(), dispUnit.getDefaultValueString());
		}
		else if (dispUnit instanceof SpinnerComponent)
		{
			newDispUnit = new SpinnerComponent(pane, jt, node, dispUnit.getName(), dispUnit.getDefaultValueString(), 
											((SpinnerComponent) dispUnit).getMinValue(), ((SpinnerComponent) dispUnit).getMaxValue());
		}
		else if (dispUnit instanceof ComboxComponent)
		{
			newDispUnit = new ComboxComponent(pane, jt, node, dispUnit.getName(), ((ComboxComponent) dispUnit).getValueList(), dispUnit.getDefaultValueString());
		}
		else if (dispUnit instanceof GroupComponent)
		{
			newDispUnit = new GroupComponent(pane, jt, node, dispUnit.getName(), dispUnit.getDefaultValueString());
		}
		else if (dispUnit instanceof LabelComponent)
		{
			newDispUnit = new LabelComponent(pane, jt, node, dispUnit.getName());
		}

		return newDispUnit;
	}

	/**
	 * Recursively reset the values of the given tree node and its offsprings
	 * to default values.
	 * 
	 * @param tree	A JTree instance.
	 * @param node	The tree node which represents the given XML Element.
	 */
	private void resetTreeNodeRecursively(JTree tree, DefaultMutableTreeNode node)
	{

		TreeModel model = tree.getModel();

		dispAndEditor manageObj = (dispAndEditor) node.getUserObject();

		//System.out.println("Current loading node "+getTreeNodeName(node));

		if (model.isLeaf(node))
		{
			return;
		}
		
		if (manageObj.isSimpleType())
		{
			manageObj.resetValue();
		}
		else
		{
			if (manageObj instanceof ChoiceComponent)
			{
				Vector children = ((ChoiceComponent) manageObj).getChildren();

				for (int i=0; i<children.size(); i++)
				{
					DefaultMutableTreeNode candidateChild = (DefaultMutableTreeNode) children.elementAt(i);

					if (manageObj.getDefaultValueString().equals(getTreeNodeName(candidateChild)))
					{
					   if (node.getChildCount() == 1)
					   {
						   node.remove(0);
						   node.add(candidateChild);
						   manageObj.resetValue();
					   }
					   break;
					}
				}
								
				for (int i=0; i<children.size(); i++)
				{
					resetTreeNodeRecursively(tree, (DefaultMutableTreeNode) children.elementAt(i));
				}
			}
			else
			{
				for (int i = 0; i < model.getChildCount(node); i++)
				{
					DefaultMutableTreeNode child = (DefaultMutableTreeNode) model.getChild(node, i);
					
					resetTreeNodeRecursively(tree, child);
				}
				
				/*
				 * Reset the number and displaying of dependent childrens
				 * under this tree node.
				 */
				updateDependent(this.editPanel, tree, node);
			}
		}
	}


	/**
	 * Load a XML Element and the subtree represented by it. The values defined in the subtree
	 * are merged into the existing JTree.
	 * 
	 * @param tree	A JTree instance.
	 * @param node	The tree node which represents the given XML Element.
	 * @param ele	The Element to be loaded.
	 * @return	true if load is successful. False otherwise.
	 */
	private boolean loadXMLElementRecursively(JTree tree, DefaultMutableTreeNode node, Element ele)
	{

		TreeModel data = tree.getModel();

		dispAndEditor manageObj = (dispAndEditor) node.getUserObject();

		
		if (debug)
			System.out.println("Current loading node "+ele.getNodeName());

		if (data.isLeaf(node))
		{
			return true;
		}
		else if (!(getTreeNodeName(node).equals(ele.getNodeName())))
		{
			return false;
		}
		
		if (manageObj.isSimpleType())
		{
			String newValue = null;


			if (debug)
				System.out.println("simpleType node "+ele.getNodeName()+" has "+ele.getChildNodes().getLength()+" childs.");

			/*
			 * For a simpleType tree node, the corresponding XML Element 
			 * may have zero or one TEXT node child. Only do replacement 
			 * when it has a child which stores new value.
			 */
			if (ele.getChildNodes().getLength() == 1)
			{
				newValue = ele.getChildNodes().item(0).getNodeValue();

				if (debug)
					System.out.println("newValue="+newValue);

				manageObj.resetValue(newValue);
			}
		}
		else
		{
			/*
			 * For each child of the Element ele, find match tree node and check recursively.
			 */
			 NodeList nl = ele.getChildNodes();

			 for (int i=0; i<nl.getLength(); i++)
			 {
				 String nodeName;
				 if (nl.item(i).getNodeType() == Node.TEXT_NODE)
				 {
					nodeName = nl.item(i).getNodeValue();
				 }
				 else
				 {
					nodeName = nl.item(i).getNodeName();
				 }
			   
				 DefaultMutableTreeNode treenode = null;
				
				 if (manageObj instanceof ChoiceComponent)
				 {
					 Vector children = ((ChoiceComponent) manageObj).getChildren();
								
					 for (int j=0; j<children.size(); j++)
					 {
						 if (debug)
						 	System.out.println("ChoiceComponent Elemen name is "+nodeName+", treenode is "+getTreeNodeName((DefaultMutableTreeNode) children.elementAt(j)));

						 DefaultMutableTreeNode candidateChild = (DefaultMutableTreeNode) children.elementAt(j);
						 if (nodeName.equals(getTreeNodeName(candidateChild)))
						 {
							if (node.getChildCount() == 1)
							{
								node.remove(0);
								node.add(candidateChild);
								manageObj.resetValue(nodeName);
							}

							treenode = candidateChild;
							break;
						 }
					 }
				 }
				 else
				 {
					 treenode = HypercastConfigurator.findChildWithName(tree, node, nodeName);
					 
					 /*
					  * If a dependent node is found, don't process it now. They are handled later.
					  */
					 if ((treenode != null) && ((dispAndEditor) treenode.getUserObject()).isDependent())
					 {
					 	continue;
					 }
				 }

				 if ((treenode == null) || 
					((nl.item(i) instanceof Element) && (!loadXMLElementRecursively(tree, treenode, (Element) nl.item(i)))))
				 {
					 return false;
				 }
				 
				 /*
				  * If a tree node has dependents, update the number of dependent nodes
				  * depending on the new value specified in this tree node.
				  */
				 dispAndEditor submanageObj = (dispAndEditor) treenode.getUserObject();
				 if (submanageObj.hasDependents())
				 {
					String dependentName = submanageObj.getDependentStr();
					int newNumber = (new Integer (submanageObj.getValueString())).intValue();

					updateDependent(this.editPanel, tree, node, newNumber, dependentName);
				 }
			 }	
				 
			/*
			 * Update the number and displaying of all dependent children of this tree node.
			 */
			TreeModel model = tree.getModel();

			for (int i = 0; i < model.getChildCount(node); i++)
			{
				DefaultMutableTreeNode child = (DefaultMutableTreeNode) model.getChild(node, i);
				dispAndEditor childManageObj = (dispAndEditor) child.getUserObject();

				if (childManageObj.hasDependents())
				{
					String dependentName = 	childManageObj.getDependentStr();
					
					/*
					 * Get the jtree children and XMl tree children which has the given 
					 * name. Their number should be equal because earlier of this method
					 * the tree dependent nodes has been updated.
					 */
					Vector treeNodes = findChildrenWithName(tree, node, dependentName);
					Vector dependentElements = findXMLChildrenWithName(ele, dependentName);
					
					if (treeNodes.size() < dependentElements.size())
					{
						continue;
					}
					else
					{
						NodeList nlist = ele.getChildNodes();
						
						for (int j=0; j<dependentElements.size(); j++)
						{
							int eleIndex = ((Integer) dependentElements.elementAt(j)).intValue();
							Element childEle = (Element) nlist.item(eleIndex);
							
							int dependentIndex = ((Integer) treeNodes.elementAt(j)).intValue();
							DefaultMutableTreeNode treeDependent = (DefaultMutableTreeNode) model.getChild(node, dependentIndex);
							
							loadXMLElementRecursively(tree, treeDependent, childEle);
						}
					}
				}
			}
		}

		return true;
	}


	/**
	 * Check if a given XML Element and the subtree represented by it are compatible with
	 * the tree node and its subtree.
	 * 
	 * @param tree	A JTree instance.
	 * @param node	The tree node which represents the given XML Element.
	 * @param ele	The Element to be loaded.
	 * 
	 * @return	true if compatible. False otherwise.
	 */
	private boolean isCompatible(JTree tree, DefaultMutableTreeNode node, Element ele)
	{
		//System.out.println("current Element is "+ele.getNodeName());

		TreeModel data = tree.getModel();

		dispAndEditor manageObj = (dispAndEditor) node.getUserObject();

		if (data.isLeaf(node))
		{
			return true;
		}
		else if (!(getTreeNodeName(node).equals(ele.getNodeName())))
		{
			//System.out.println("name doesn't match : tree node "+getTreeNodeName(node)+", Element "+ele.getNodeName());

			return false;
		}
		else if (manageObj.isSimpleType())
		{
			if (node.getChildCount() != 1)
			{
				return false;
			}
		}
		else 
		{		
		   /*
			* For each child of the Element ele, find match tree node and check recursively.
			*/
			NodeList nl = ele.getChildNodes();

			for (int i=0; i<nl.getLength(); i++)
			{
				String nodeName;
				
				if (nl.item(i).getNodeType() == Node.TEXT_NODE)
				{
					nodeName = nl.item(i).getNodeValue();
				}
				else
				{
					nodeName = nl.item(i).getNodeName();
				}
			   
				DefaultMutableTreeNode treenode = null;
				
				if (manageObj instanceof ChoiceComponent)
				{
					Vector children = ((ChoiceComponent) manageObj).getChildren();
								
					for (int j=0; j<children.size(); j++)
					{
						//System.out.println("ChoiceComponent Elemen name is "+nodeName+", treenode is "+getTreeNodeName((DefaultMutableTreeNode) children.elementAt(j)));
						//System.out.println("ChoiceComponent Elemen name is "+nodeName+", treenode is "+ ((String) childrenNames.elementAt(j)));

						if (nodeName.equals(getTreeNodeName((DefaultMutableTreeNode) children.elementAt(j))))
						//if (nodeName.equals((String) childrenNames.elementAt(j)))
						{
							treenode = (DefaultMutableTreeNode) children.elementAt(j);
							break;
						}
					}
				}
				else
				{
					treenode = HypercastConfigurator.findChildWithName(tree, node, nodeName);
				}
				
				if ((treenode == null) || 
					((nl.item(i) instanceof Element) &&(!isCompatible(tree, treenode, (Element) nl.item(i)))))
				{
/*
					if (treenode == null)
						System.out.println("Doesn't find tree node for "+nodeName);
					else
						System.out.println("subtree for child "+nodeName+" is not compatible!");
*/					
					return false;
				}
			}
			   
		}
		
		return true;
	}


	/*
	 * Get the name of a tree node. It is defined for no method is found in 
	 * class DefaultMutableTreeNode which returns the name of a tree node.
	 */
	private String getTreeNodeName(DefaultMutableTreeNode node)
	{
		dispAndEditor manageObj = (dispAndEditor) node.getUserObject();
		if (manageObj != null)
		{
			return manageObj.getName();
		}
		else
		{
			return "";
		}
	}

}


/**
 * This interface defines the functions for displaying or
 * hidding the Swing untis for a tree node.
 */
interface dispAndEditor
{
	/**
	 * Gets the name of the Swing unit.
	 */
	public String getName();

	/**
	 * Displays the Swing unit on the edit pane.
	 */
	public void display();

	/**
	 * Hides the Swing unit which is displayed on the edit pane.
	 */
	public void hidden();

	/**
	 * Returns the value string of this object which is displayed in the corresponding
	 * Swing unit.
	 */
	public String getValueString();

	/**
	 * Returns the default value string of this object which is to be displayed in the corresponding
	 * Swing unit.
	 */
	public String getDefaultValueString();

	/**
	 * Sets the leafNodeDisplayHelper instance in this dispAndEditor object which
	 * represents a simpletype property node.
	 * @param lp a leafNodeDisplayHelper instance.
	 */
	public void setDispHelper(leafNodeDisplayHelper lp);
	
	/**
	 * Return true if the implentation of this interface is a dependent
	 * of another dispAndEditor object. False otherwise.
	 */
	public boolean isDependent();

	/**
	 * Return true if the implentation of this interface has other dispAndEditor
	 * objects as its dependents. False otherwise.
	 */
	public boolean hasDependents();

	/**
	 * Sets the isDependent flag in this dispAndEditor object. Sets to 
	 * true then this dispAndEditor object is a dependent of other
	 * dispAndEditor object.
	 * @param x	isDependent flag.
	 */
	public void setDependentFlag(boolean x);

	/**
	 * Return true if this dispAndEditor object represents a simple type
	 * node in the property tree. False otherwise.
	 */
	public boolean isSimpleType();

	/**
	 * Sets the simpletype flag in this dispAndEditor object.
	 * @param x	simpletype flag.
	 */
	public void setIsSimpleType(boolean x);

	/**
	 * Sets the name of dependent of this dispAndEditor object.
	 * @param dependentName	dependent name string.
	 */	
	public void setDependentStr(String dependentName);

	/**
	 * Returns the dependent name string of this dispAndEditor object.
	 */
	public String getDependentStr();

	/**
	 * Resets the value of this dispAndEditor object to default value.
	 */
	public void resetValue();

	/**
	 * Resets the value of this dispAndEditor object to the given one.
	 */
	public void resetValue(String newValueStr);
	
	/**
	 * Sets the Element of the shcema file which corresponds to
	 * this dispAndEditor object.
	 * @param ele	Element in the Schema Document.
	 */	
	public void setSchemaElement(Element ele);

	/**
	 * Returns the Element of the Schema Document which corresponds to
	 * this dispAndEditor object.
	 */
	public Element getSchemaElement();

	/**
	 * Sets the HypercastConfigurator object to which this dispAndEditor object
	 * belongs.
	 */	
	public void setHyperCastConf(HypercastConfigurator hc);

	/**
	 * Returns the HypercastConfigurator object to which this dispAndEditor object
	 * belongs.
	 */
	public HypercastConfigurator setHyperCastConf();

}

/**
 * This class defines a Swing unit, Label, which is extended by all other
 * Swing unit classes.
 */
class LabelComponent implements dispAndEditor
{
	/**
	 * The name of the tree node.
	 */
	String nodeName;
	
	/**
	 * A JPanel instance for editing a tree node.
	 */
	JPanel editpane;

	/**
	 * A JTree instance.
	 */
	JTree jtree;

	/**
	 * Tree node to which this object is bound.
	 */
	DefaultMutableTreeNode treeNode;

	/**
	 * The Label for displaying.
	 */	
	JLabel label;

	/**
	 * Tips which is displayed when click on the Label of this Component. 
	 */
	String tip = "";

	/**
	 * a leafNodeDisplayHelper instance which helps to get 
	 * representation of value of the tree node to which 
	 * this object is bound.
	 */
	leafNodeDisplayHelper dispHelper;

	/**
	 * Flag showing if the tree node to whicn this object is bound is
	 * a dependent.
	 */
	boolean isDependent = false;

	/**
	 * The name of dependent node. The number of dependent node in the jtree
	 * is decided by the value in the Swing unit of this object.
	 */
	String dependentName = null;

	/**
	 * Flag showing if the tree node represents a simpleType ELement in the 
	 * schema document.
	 */
	boolean isSimpleType = false;
	
	/**
	 * Default value used to initilize a Swing unit which extends this class.
	 */
	String defaultValueStr = "";

	/**
	 * Element in the Schema Document which corresponds to this dispAndEditor object.
	 * It is used when handling dependenct attributes.
	 */	
	Element schemaEle = null;

	/**
	 * the HypercastConfigurator object to which this dispAndEditor object
	 * belongs.	
	 */
	HypercastConfigurator hconf = null;
		
	/**
	 * The position to display this Label in the edit panel.
	 */
	int labelX = 100, labelY = 20;
	
	/**
	 * Constructor.
	 *  
	 * @param pane	JPanel instance for displaying Swing units.
	 * @param jt	JTree instance which defines a tree consisting of nodes in the configuration file.
	 * @param treenode	The node on the tree which represents an attribute. This object (a Swing Label)
	 *  defines the label for the tree node.
	 * @param name	The name displayed on the Swing Label unit.
	 */
	public LabelComponent(JPanel pane, JTree jt, DefaultMutableTreeNode treenode, String name)
	{
		editpane = pane;
		jtree = jt;
		treeNode = treenode;
		nodeName = name;

		label = new JLabel(nodeName);
		label.setVisible(false);
		
		if (!tip.equals(""))
		{
			label.setToolTipText(tip);
		}

		label.setBounds(labelX, labelY, HypercastConfigurator.LABEL_WIDTH, HypercastConfigurator.LABEL_HEIGHT);
		
		editpane.add(label);
	}

	/** Return x coordinate of the Swing unit. */
	protected int getX()
	{
		return labelX;
	}

	/** Return y coordinate of the Swing unit. */
	protected int getY()
	{
		return labelY;
	}

	/** Return the name of the Swing unit. */
	public String getName()
	{
		return nodeName;
	}

	/**
	 * Resets the value of this object. It does nothing for a Label component.
	 */
	public void resetValue() {}

	/**
	 * Resets the value of this object with the given string.  It does nothing for a Label component.
	 */
	public void resetValue(String newValueStr) {}

	/** Enable display flag for the Swing unit defined in this object. */
	public void display()
	{
		label.setVisible(true);
	}
	
	/** Disable display flag for the Swing unit defined in this object. */
	public void hidden()
	{
		label.setVisible(false);
	}

	/** Return the name of the Swing unit defined in this object. */
	public String toString()
	{
		return nodeName;
	}

	/**
	 * Returns the value string of this object which is displayed in the corresponding
	 * Swing unit.
	 * 
	 * @see dispAndEditor#getValueString() 
	 */
	public String getValueString()
	{
		return nodeName;
	}

	/**
	 * Returns the default value string of this object which is to be displayed in the corresponding
	 * Swing unit.
	 * 
	 * @see dispAndEditor#getDefaultValueString() 
	 */
	public String getDefaultValueString()
	{
		return defaultValueStr;
	}


	/**
	 * Sets the leafNodeDisplayHelper instance in this object which
	 * represents a simpletype property node.
	 * 
	 * @see dispAndEditor#setDispHelper(leafNodeDisplayHelper) 
	 */
	public void setDispHelper(leafNodeDisplayHelper lp)
	{
		dispHelper = lp;
	}
	
	/**
	 * Return true if the implentation of this interface is a dependent
	 * of another dispAndEditor object. False otherwise.
	 * 
	 * @see dispAndEditor#isDependent() 
	 */
	public boolean isDependent()
	{
		return isDependent;
	}

	/**
	 * Return true if the implentation of this interface has other dispAndEditor
	 * objects as its dependents. False otherwise.
	 * 
	 * @see dispAndEditor#hasDependents() 
	 */
	public boolean hasDependents()
	{
		if (dependentName != null)
		{
			return true;
		}
		else
		{
			return false;
		}
	}


	/**
	 * Sets the isDependent flag in this object. Sets to 
	 * true then this object is a dependent of other
	 * dispAndEditor object.
	 * 
	 * @see dispAndEditor#setDependentFlag(boolean) 
	 */
	public void setDependentFlag(boolean x)
	{
		isDependent = x;
	}

	/**
	 * Sets the name of dependent of this object.
	 * 
	 * @see dispAndEditor#setDependentStr(String) 
	 */	
	public void setDependentStr(String depName)
	{
		dependentName = depName;
	}

	/**
	 * Returns the dependent name string of this object.
	 * 
	 * @see dispAndEditor#getDependentStr() 
	 */
	public String getDependentStr()
	{
		return dependentName;
	}

	/**
	 * Return true if this object represents a simple type
	 * node in the property tree. False otherwise.
	 * 
	 * @see dispAndEditor#isSimpleType() 
	 */
	public boolean isSimpleType()
	{
		return isSimpleType;
	}
	
	/**
	 * Sets the simpletype flag in this object.
	 * 
	 * @see dispAndEditor#setIsSimpleType(boolean) 
	 */
	public void setIsSimpleType(boolean x)
	{
		isSimpleType = x;
	}


	/**
	 * Sets the Element of the shcema file which corresponds to
	 * this object.
	 * 
	 * @see dispAndEditor#setSchemaElement(Element)
	 * 
	 */	
	public void setSchemaElement(Element ele)
	{
		schemaEle = ele;
	}

	/**
	 * Returns the Element of the Schema Document which corresponds to
	 * this object.
	 * 
	 * @see dispAndEditor#getSchemaElement() 
	 */
	public Element getSchemaElement()
	{
		return schemaEle;
	}

	/**
	 * Sets the HypercastConfigurator object to which this object
	 * belongs.
	 * 
	 * @see dispAndEditor#setHyperCastConf(HypercastConfigurator) 
	 */	
	public void setHyperCastConf(HypercastConfigurator hc)
	{
		hconf = hc;
	}

	/**
	 * Returns the HypercastConfigurator object to which this object
	 * belongs.
	 * 
	 * @see dispAndEditor#setHyperCastConf() 
	 */
	public HypercastConfigurator setHyperCastConf()
	{
		return hconf;
	}

}


/**
 * This class defines a Swing unit which defines choice function on a set of
 * children.
 */
class ChoiceComponent extends LabelComponent
implements ActionListener
{
	/**
	 * The vctor of children of this tree node to be selected.
	 */
	Vector children = new Vector();

	/**
	 * The vctor of names of the children of this tree node to be selected.
	 */
	Vector childrenNames = new Vector();
		
	/**
	 * The JComboBox instance for choosing name from a list of candidates.
	 * It is for Element which defines choice on its children.
	 */	
	JComboBox combox;
	
	/**
	 * Constructor.
	 *  
	 * @param pane	JPanel instance for displaying Swing units.
	 * @param jt	JTree instance which defines a tree consisting of nodes in the configuration file.
	 * @param treenode	The node on the tree which represents an attribute. This object defines 
	 *  the label for the tree node.
	 * @param name	The name displayed on the Swing Label unit attached to this Swing unit.
	 * @param childrenVec	The vector contains all children for choosing.
	 * @param childrennames	The vector contains the names of all the children for choosing.
	 * @param defaultvalue	The default of the choice on the children.
	 * 
	 */
	public ChoiceComponent(JPanel pane, JTree jt, DefaultMutableTreeNode treenode, String name, Vector childrenVec, Vector childrennames, String defaultvalue)
	{
		super (pane, jt, treenode, name);
		
		children = childrenVec;
		
		childrenNames = childrennames;

		defaultValueStr = defaultvalue;

		combox = new JComboBox(childrenNames);
		combox.setEditable(false);

		resetValue(defaultValueStr);

		int x = super.getX();
		int y = super.getY()+HypercastConfigurator.LABEL_HEIGHT;

		combox.setBounds(x, y, HypercastConfigurator.COMBOX_WIDTH, HypercastConfigurator.COMBOX_HEIGHT);

		editpane.add(combox);
		
		combox.addActionListener(this);
	}

	/** Reset the the selected item to default value. */
	public void resetValue()
	{
		combox.setSelectedItem(defaultValueStr);
		combox.setVisible(false);
	}

	/** Reset the selected item to the item having given string as name. */
	public void resetValue(String newValueStr)
	{
		for (int i=0; i<childrenNames.size()-1; i++)
		{
			if (newValueStr.equals((String) childrenNames.elementAt(i)))
			{
				combox.setSelectedItem(newValueStr);
				combox.setVisible(false);
				return;
			}
		}
		
		/*
		 * If newValueStr is not a valid value, use default value to reset.
		 */
		resetValue();
	}

	/** Display this Swing unit. */
	public void display()
	{
		super.display();
		
		combox.setVisible(true);
	}
	
	/** Hidden this Swing unit. */
	public void hidden()
	{
		super.hidden();
		
		combox.setVisible(false);
	}

	/**
	 * Return the children vector.
	 */
	public Vector getChildren()
	{
		return children;
	}

	/**
	 * Set the children vector.
	 */
	public void setChildren(Vector childrenvec)
	{
		children = childrenvec;
	}
	
	/**
	 * Return the children name vector.
	 */
	public Vector getChildrenNames()
	{
		return childrenNames;
	}

	/**
	 * Return the node name. It is called by the JTree instance to get 
	 * the representation of a tree node in the tree.
	 */
	public String getValueString()
	{
		return nodeName;
	}
	
	/**
	 * Handle the event on Swing unit contained in the variable "Editor".
	 *
	 * @param e ActionEven.
	 */
	public void actionPerformed(ActionEvent e) 
	{
			String selectedItem = (String)(((JComboBox)e.getSource()).getSelectedItem());

			//System.out.println("selected "+selectedItem);
						
			for (int i=0; i<children.size(); i++)
			{
				DefaultMutableTreeNode candidateChild = (DefaultMutableTreeNode) children.elementAt(i);
				
				if (candidateChild.toString().equals(selectedItem))
				{

					if (treeNode.getChildCount() == 1)
					{
						//System.out.println("find candidate child "+candidateChild.toString()+", it will exchange with "+treeNode.getChildAt(0).toString());

						treeNode.remove(0);
						treeNode.add(candidateChild);
					}
					
					/*
					 * Redraw the tree node.
					 */
					((DefaultTreeModel) jtree.getModel()).reload(candidateChild);
					
					//System.out.println("treeNode's child become "+treeNode.getChildAt(0).toString());
					
					break;
				}
			}
	}


}

/**
 * This class defines a Swing unit which defines an area for typing and editing text.
 */
class TextFieldComponent extends LabelComponent
implements FocusListener, KeyListener
{
	/**
	 * The JTextField instance for editing a string format attribute.
	 */	
	JTextField field;

	/**
	 * Constructor.
	 *  
	 * @param pane	JPanel instance for displaying Swing units.
	 * @param jt	JTree instance which defines a tree consisting of nodes in the configuration file.
	 * @param treenode	The node on the tree which represents an attribute. This object defines 
	 *  the label for the tree node.
	 * @param name	The name displayed on the Swing Label unit attached to this Swing unit.
	 * @param defaultvalue	The default value displayed in the text editing area.
	 * 
	 */
	public TextFieldComponent(JPanel pane, JTree jt, DefaultMutableTreeNode treenode, String name, String defaultvalue)
	{
		super (pane, jt, treenode, name);

		defaultValueStr = defaultvalue;

		field = new JTextField();

		resetValue(defaultvalue);

		int x = super.getX();
		int y = super.getY()+HypercastConfigurator.LABEL_HEIGHT;

		field.setBounds(x, y, HypercastConfigurator.TEXTFIELD_WIDTH, HypercastConfigurator.TEXTFIELD_HEIGHT);

		editpane.add(field);

		field.addFocusListener(this);
		field.addKeyListener(this);
	}

	/** Reset the text to default value in the editing area. */
	public void resetValue()
	{
		field.setText(defaultValueStr);
		field.setVisible(false);
	}

	/** Reset the text to given string in the editing area. */
	public void resetValue(String newValueStr)
	{
		field.setText(newValueStr);
		field.setVisible(false);
	}

	/** Display this Swing unit. */
	public void display()
	{
		super.display();
		
		field.setVisible(true);
	}
	
	/** Hidden this Swing unit. */
	public void hidden()
	{
		super.hidden();
		
		field.setVisible(false);
	}

	/** Return the text in the editing area.*/
	public String getValueString()
	{
		if (field != null)
		{
			return field.getText();
		}
		else
		{
			return "";
		}
	}

	/**
	 * <p>Another callback for TextComponents; we don't care when a component
	 * gains focus, so this does nothing, it's just needed for compatibility.<p>
	 *
	 * @param e FocusEvent
	 */
	public void focusGained(FocusEvent e){}


	/**
	 * <p>The callback for TextComponents; sets the appropriate ones visible.</p>
	 *
	 * @param e FocusEvent, data updates when boxes lose focus
	 */
	public void focusLost(FocusEvent e) 
	{
		((DefaultTreeModel) jtree.getModel()).nodeChanged(treeNode);

	}

	/** Handle the key typed event from the text field. */
	public void keyTyped(KeyEvent e) {
		((DefaultTreeModel) jtree.getModel()).nodeChanged(treeNode);
	}

	/** Handle the key pressed event from the text field. */
	public void keyPressed(KeyEvent e) {
	}

	/** Handle the key released event from the text field. */
	public void keyReleased(KeyEvent e) {
	}

}


/**
 * This class defines a Swing unit which defines a spinner for choosing integer value.
 */
  class SpinnerComponent extends LabelComponent
  implements ChangeListener
  {
	  /**
	   * The minimum and maximum value which can be selected in the spinner.
	   */
	  int min=1, max = 1;
	
	  /**
	   * The JSpinner instance for editing a integer attribute with maximum
	   * and minimum values.
	   */	
	  JSpinner spinner;
	
	/**
	 * Constructor.
	 *  
	 * @param pane	JPanel instance for displaying Swing units.
	 * @param jt	JTree instance which defines a tree consisting of nodes in the configuration file.
	 * @param treenode	The node on the tree which represents an attribute. This object
	 *  defines the label for the tree node.
	 * @param name	The name displayed on the Swing Label unit attached to this Swing unit.
	 * @param defaultvalue	The default value displayed in the spinner.
	 * @param minVal	The minimal value supported in the spinner.
	 * @param maxVal	The maximal value supported in the spinner.
	 * 
	 */
	  public SpinnerComponent(JPanel pane, JTree jt, DefaultMutableTreeNode treenode, String name, String defaultvalue, String minVal, String maxVal)
	  {
			super (pane, jt, treenode, name);

			defaultValueStr = defaultvalue;

			Integer val;
			if (defaultvalue != null)
			{
				val = new Integer(defaultvalue);
			}
			else 
			{
				val = new Integer(1);
			}
	
			min = (new Integer(minVal)).intValue();
			max = (new Integer(maxVal)).intValue();
	
			spinner = new JSpinner(new SpinnerNumberModel(val.intValue(), min, max, 1));
	
			spinner.setVisible(false);
	
			int x = super.getX();
			int y = super.getY()+HypercastConfigurator.LABEL_HEIGHT;

			spinner.setBounds(x, y, HypercastConfigurator.SPINNER_WIDTH, HypercastConfigurator.SPINNER_HEIGHT);
	
			editpane.add(spinner);
			
			spinner.addChangeListener(this);
	  }

	  /** Reset the value displayed in the spinner to default value. */
	  public void resetValue()
	  {
		spinner.setValue(new Integer(defaultValueStr));
		spinner.setVisible(false);
	  }

	  /** Reset the value diaplyed in the spinner to given string */
	  public void resetValue(String newValueStr)
	  {
		int newValue = 0;

		try
		{
			newValue = Integer.parseInt(newValueStr);
		}
		catch (Exception e)
		{
			resetValue();
			return;
		}

		if ((newValue < min) || (newValue > max))
		{
			resetValue();
		}
		else
		{
			spinner.setValue(new Integer(newValue));
			spinner.setVisible(false);
		}
	  }

	  /** Return the minimal value supported by the spinner. */
	  public String getMinValue()
	  {
		  return ""+min;
	  }

	/** Return the maximal value supported by the spinner. */
	  public String getMaxValue()
	  {
		  return ""+max;
	  }

	  /** Display this Swing unit */
	  public void display()
	  {
		  super.display();
		
		  spinner.setVisible(true);
	  }
	
	  /** Hidden this Swing unit. */
	  public void hidden()
	  {
		  super.hidden();
		
		  spinner.setVisible(false);
	  }

	/**
	 * Handle the event on Swing unit "spinner".
	 *
	 * @param e ChangeEvent instance.
	 */
	public void stateChanged(ChangeEvent e) 
	{
		((DefaultTreeModel) jtree.getModel()).nodeChanged(treeNode);
		
		/*
		 * If this obejct has dependent, update the number of dependent nodes.
		 */
		if (dependentName != null)
		{
			//System.out.println("Spinner: update dependent.");
			
			hconf.updateDependent(editpane, jtree, (DefaultMutableTreeNode) treeNode.getParent().getParent(), ((Integer) spinner.getValue()).intValue(), dependentName);
		}
	}

	/** Return the value in string format displayed in the spinner. */
	public String getValueString()
	{
		if (spinner != null)
		{
			return ((Integer) spinner.getValue()).toString();
		}
		else
		{
			return "";
		}
	}
	
  }


/**
 * This class defines a Swing unit which defines a combox for choosing 
 * value from a list of values.
 */
  class ComboxComponent extends LabelComponent
  implements ActionListener
  {
	  /**
	   * The vector of valid values for selecting.
	   */
	  Vector valueVec = new Vector();

	  /**
	   * The JComboBox instance for choosing value from a list of candidates.
	   */	
	  JComboBox combox;

	/**
	 * Constructor.
	 *  
	 * @param pane	JPanel instance for displaying Swing units.
	 * @param jt	JTree instance which defines a tree consisting of nodes in the configuration file.
	 * @param treenode	The node on the tree which represents an attribute. This object 
	 *  defines the label for the tree node.
	 * @param name	The name displayed on the Swing Label unit attached to this Swing unit.
	 * @param valuelist The value list supported by this combox object.
	 * @param selectedValue	The default selected value.
	 * 
	 */
	  public ComboxComponent(JPanel pane, JTree jt, DefaultMutableTreeNode treenode, String name, Vector valuelist, String selectedValue)
	  {
		  super (pane, jt, treenode, name);

		  defaultValueStr = selectedValue;

		  valueVec = valuelist;

		  combox = new JComboBox(valuelist);
		  combox.setEditable(false);
		  combox.setSelectedItem(selectedValue);
		  combox.setVisible(false);

		  int x = super.getX();
		  int y = super.getY()+HypercastConfigurator.LABEL_HEIGHT;

		  combox.setBounds(x, y, HypercastConfigurator.COMBOX_WIDTH, HypercastConfigurator.COMBOX_HEIGHT);

		  editpane.add(combox);

		  combox.addActionListener(this);
	  }

	  /**
	   * Return the valid value vector.
	   */
	  public Vector getValueList()
	  {
		  return valueVec;
	  }


		public void resetValue()
		{
			combox.setSelectedItem(defaultValueStr);
			combox.setVisible(false);
		}
	
		public void resetValue(String newValueStr)
		{
			for (int i=0; i<valueVec.size(); i++)
			{
/*
				if (HypercastConfigurator.debug)
					System.out.println("ComboxComponent: valid element="+(String) valueVec.elementAt(i));
*/

				if (newValueStr.equals((String) valueVec.elementAt(i)))
				{
/*
					if (HypercastConfigurator.debug)
						System.out.println("ComboxComponent: found valid element whse name is "+newValueStr);
*/

					combox.setSelectedItem(newValueStr);
					combox.setVisible(false);
					return;
				}
			}

/*			
			if (HypercastConfigurator.debug)
				System.out.println("ComboxComponent: don't find valid element, set selected element to defaultValue="+defaultValueStr+".");
*/

			/*
			 * If newValueStr is not a valid value, use default value to reset.
			 */
			resetValue();
		}

		/** Display this Swing unit. */
		public void display()
		{
			super.display();
			
			combox.setVisible(true);
		}
		
		/** Hidden this Swing unit. */
		public void hidden()
		{
			super.hidden();
			
			combox.setVisible(false);
		}

		/**
		 * Handle the event on Swing unit contained in the variable "Editor".
		 *
		 * @param e ActionEven.
		 */
		public void actionPerformed(ActionEvent e) 
		{
			((DefaultTreeModel) jtree.getModel()).nodeChanged(treeNode);
		}
	
		/** Return the selected value in string format. */
		public String getValueString()
		{
			if (combox != null)
			{
				return (String)(combox.getSelectedItem());
			}
			else
			{
				return "";
			}
		}

  }


/**
 * This class defines a Swing unit which defines a combination of spinners for choosing 
 * a set of integer values.
 */
	class GroupComponent extends LabelComponent
	implements ChangeListener
	{
		/**
		 * The vector which contains a set of Swing display units for editing
		 * a Element with complex value format, such as INETV4AndOnePort and
		 * INETV4AndTwoPorts.
		 */	
		Vector btns;
		
		/**
		 * Constructor.
		 *  
		 * @param pane	JPanel instance for displaying Swing units.
		 * @param jt	JTree instance which defines a tree consisting of nodes in the configuration file.
		 * @param treenode	The node on the tree which represents an attribute. This object 
		 *  defines the label for the tree node.
		 * @param name	The name displayed on the Swing Label unit attached to this Swing unit.
		 * @param defaultvalue	Thestring contains the default selected value for each spinner.
		 * 
		 */
		public GroupComponent(JPanel pane, JTree jt, DefaultMutableTreeNode treenode, String name, String defaultvalue)
		{
			super (pane, jt, treenode, name);

			defaultValueStr = defaultvalue;

			//String buttons = (configprops.getProperty(key + ".Default")).trim();
			/*
			 * class the group in two catergories
			 * 1. IP + Port 
			 * 2. IP + Port + Port
			 */
			
			String[] paFields = defaultvalue.trim().split(":");

			int x = super.getX();
			int y = super.getY()+HypercastConfigurator.LABEL_HEIGHT;
	
			/*
			 * create the group of spinner, label and textfield
			 */
			btns = new Vector();
			int pos = 0;				
			JSpinner spinner;
			JLabel lab;
	
			for (int i=0; i<paFields.length; i++ )
			{
				String buttons = paFields[i];
				while ((pos = buttons.indexOf('.')) != -1) 
				{
					spinner = new JSpinner(new SpinnerNumberModel((new Integer (buttons.substring(0, pos))).intValue(), 0, 255, 1));
					spinner.setBounds(x, y, HypercastConfigurator.SPINNERFORNUMBER_WIDTH, HypercastConfigurator.COMBOX_HEIGHT);
					spinner.addChangeListener(this);
					btns.add(spinner);
					x += HypercastConfigurator.SPINNERFORNUMBER_WIDTH;
	
					lab = new JLabel(".");
					lab.setBounds(x, y, 10, HypercastConfigurator.LABEL_HEIGHT);
					btns.add(lab);
					x += 10;
				
					buttons = buttons.substring(pos+1);
				}
				
				if (i == 0)
				{
					spinner = new JSpinner(new SpinnerNumberModel((new Integer (buttons)).intValue(), 0,255, 1));
				}
				else
				{
					spinner = new JSpinner(new SpinnerNumberModel((new Integer (buttons)).intValue(), 0,65535, 1));
				}

				spinner.setBounds(x, y, HypercastConfigurator.SPINNERFORNUMBER_WIDTH, HypercastConfigurator.COMBOX_HEIGHT);
				spinner.addChangeListener(this);
				btns.add(spinner);
				x += HypercastConfigurator.SPINNERFORNUMBER_WIDTH;
				
				if (i < paFields.length-1)
				{
					lab = new JLabel(":");
					//lab.setAlignmentX(Label.CENTER);
					lab.setBounds(x, y, 10, HypercastConfigurator.LABEL_HEIGHT);
					btns.add(lab);
					x += 10;
				}
			}
	
			/*
			 * add to the panel
			 */
			for (int i=0; i<btns.size(); i++) 
			{
				Object unit = btns.elementAt(i);
				
				if (unit instanceof JSpinner)
				{
					editpane.add((JSpinner) unit);
					((JSpinner) unit).setVisible(false);
				}
				else if (unit instanceof JLabel)
				{
					editpane.add((JLabel) unit);
					((JLabel) unit).setVisible(false);
				}
			}
		}

		/** Reset the selected value for each spinner to default value. */
		public void resetValue()
		{
			resetValue(defaultValueStr);
		}
	
		/** Reset the selected value for each spinner to the values contained in the given string. */
		public void resetValue(String newValueStr)
		{
			String newStr = newValueStr;
			if (!isCompatible(defaultValueStr, newValueStr))
			{
				newStr = defaultValueStr;
			}

			//System.out.println("new String for spinner reset is "+newStr);
			
			Vector valueFields = getVauleFields(newStr);

			//System.out.println("Length of new value vector is "+valueFields.size());
			
			int j = 0;
			
			for (int i=0; i<btns.size(); i++) 
			{
				if (j >= valueFields.size())
				{
					return;
				}

				Object unit = btns.elementAt(i);
				
				if (unit instanceof JSpinner)
				{
					((JSpinner) unit).setValue((Integer) valueFields.elementAt(j));
					j++;
					
					((JSpinner) unit).setVisible(false);
				}
				else if (unit instanceof JLabel)
				{
					((JLabel) unit).setVisible(false);
				}
			}
			
		}

		/** Check if the values contained in the given string is compatible to the default value string.*/
		private boolean isCompatible(String defaultStr, String newValueStr)
		{
			/*
			 * Once try to use String.split() function but found it doesn't
			 * work for delimiter ".", so switch to use StringTokenizer.
			 */

			StringTokenizer defaultFields = new StringTokenizer(defaultStr, ":");
			StringTokenizer newvalueFields = new StringTokenizer(newValueStr, ":");
			
			if (defaultFields.countTokens() != newvalueFields.countTokens())
			{
				return false;
			}

			while (defaultFields.hasMoreTokens())
			{
				StringTokenizer defaultSubFields = new StringTokenizer(defaultFields.nextToken(), ".");
				StringTokenizer newvalueSubFields = new StringTokenizer(newvalueFields.nextToken(), ".");
				
				if (defaultSubFields.countTokens() != newvalueSubFields.countTokens())
				{
					return false;
				}
			}
			return true;
		}

		
		/** Return the selected values of all spinner as a string delimited with ":". */
		private Vector getVauleFields(String valueStr)
		{
			Vector valueVec = new Vector();
			
			StringTokenizer valueFields = new StringTokenizer(valueStr, ":");

			int i = 0;			
			while (valueFields.hasMoreTokens())
			{
				StringTokenizer valueSubFields = new StringTokenizer(valueFields.nextToken(), ".");

				while (valueSubFields.hasMoreTokens())
				{
					int newValue = 0;
					try
					{
						newValue = Integer.parseInt(valueSubFields.nextToken());
					}
					catch (Exception e)
					{
						newValue = 1;
					}
	
					if ((newValue < 0) || ((i == 0) && (newValue > 255)))
					{
						newValue = 1;
					}
					
					valueVec.addElement(new Integer(newValue));			
				}
				
				i++;
			}
			
			return valueVec;
		}		


		/** Display this Swing unit. */
		public void display()
		{
			super.display();
		
			for (int i=0; i<btns.size(); i++) 
			{
				Object unit = btns.elementAt(i);
				
				if (unit instanceof JSpinner)
				{
					((JSpinner) unit).setVisible(true);
				}
				else if (unit instanceof JLabel)
				{
					((JLabel) unit).setVisible(true);
				}
			}
		}
	
		/** Hidden this Swing unit. */
		public void hidden()
		{
		  	super.hidden();
		
			for (int i=0; i<btns.size(); i++) 
			{
				Object unit = btns.elementAt(i);
					
				if (unit instanceof JSpinner)
				{
					((JSpinner) unit).setVisible(false);
				}
				else if (unit instanceof JLabel)
				{
					((JLabel) unit).setVisible(false);
				}
			}
		}

		/**
		 * Handle the event on Swing unit vector "btns".
		 *
		 * @param e ChangeEvent instance.
		 */
		public void stateChanged(ChangeEvent e) 
		{
			((DefaultTreeModel) jtree.getModel()).nodeChanged(treeNode);
		}
	
		/** Return the values in the spinners in string format. */
		public String getValueString()
		{
			if (btns != null)
			{
				String value = "";
				for (int i=0; i<btns.size(); i++) 
				{
					Object unit = btns.elementAt(i);
						
					if (unit instanceof JSpinner)
					{
						value = value + ((Integer) ((JSpinner) unit).getValue()).toString();
					}
					else if (unit instanceof JLabel)
					{
						value = value + ((JLabel) unit).getText();
					}
				}
				
				return value;
			}
			else
			{
				return "";
			}
		}
	}

/** This class defines an object to help other dispAndEditor object to create value string. */
class leafNodeDisplayHelper
{
	dispAndEditor parent;
	
	/** Constructor */
	public leafNodeDisplayHelper(dispAndEditor p)
	{
		parent = p;
	}

	/** Return the value of the parent object in string format. */
	public String toString()
	{
		return parent.getValueString();
	}
}
 
