/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.util;

import java.io.File;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import java.io.InputStream;
import java.util.Vector;
import java.io.IOException;
import java.text.ParseException;
import java.io.FileNotFoundException;


/**
 * This class defines functions to create a XML Document
 * from a schema file. The created XML Document is built 
 * based on the default values defined in the scema file. 
 *
 * @author HyperCast Team
 * @version 2005 (version 3.0)
 */
public class createPrivatePropertyDocFromSchemaFile
{
	/**
	 * debug flag
	 */
	final static boolean debug = false;
	
	/** The schema file. */
	private File SchemaFile = null;

	/** The schema file name. */
	private String schemaFileName = null;

	/** The Document instance created from the opened schema file. */
	private Document openedSchemaFile = null;

	/** The default schema file name. It is used when no schema file
	 * is specified to create Document.
	 */
	private final String defaultSchemaFileName = "hypercast_private.xsd";

    /**
     * Constructor.
     */
    public createPrivatePropertyDocFromSchemaFile(String schemafile)
    throws IOException, ParseException
	{
		schemaFileName = schemafile;

		SchemaFile = new File(schemafile);

		if (schemaFileName.toLowerCase().endsWith(".xsd"))
		{
			/*
			 * Open schema file.
			 */
			try
			{ 
				openedSchemaFile = XmlUtil.createDocument(schemafile);
			}
			catch (final IOException ioe)
			{
				throw new IOException("Cannot open schema file "+schemaFileName+": "+ioe);
			}
			catch (final ParseException ioe)
			{
				throw new ParseException("Failed in parsingschema  file "+schemaFileName+": "+ioe, 0);
			}

			if (openedSchemaFile == null)
			{
				throw new IOException("Poor formated schema file: hypercast.xsd");
			}
		}
		else
		{		
		}
    }

	/**
	 * Constructor.
	 */
	public createPrivatePropertyDocFromSchemaFile()
	throws IOException
	{
		schemaFileName = defaultSchemaFileName;

		InputStream fis = this.getClass().getResourceAsStream( schemaFileName );
				
		if (fis == null)
		{
			//System.out.println("Connot find default schema file "+schemaFileName+".");
			return;
		}
				
		/*
		 * Open schema file.
		 */
		try
		{ 
			openedSchemaFile = XmlUtil.createDocument(fis, schemaFileName);
		}
		catch (final IOException ioe)
		{
			throw new IOException("Cannot open schema file "+schemaFileName+": "+ioe);
		}
				
		if (openedSchemaFile == null)
		{
			throw new IOException("Poor formated schema file: "+schemaFileName);
		}

		//System.out.println("Schema document is created from "+schemaFileName+".");

	}

	/**
	 * Build the Document object for all attribute Elements (based on their default values).
	 */
	public Document getPrivateProperties()
		throws IOException
	{
		if (openedSchemaFile == null)
		{
		   //System.out.println("openedSchemaFile is null.");

			schemaFileName = defaultSchemaFileName;

			InputStream fis = getClass().getResourceAsStream( schemaFileName );
				
			if (fis == null)
			{
				//System.out.println("open schema file "+openedSchemaFile+" failed.");

				throw new IOException("Connot find default schema file "+schemaFileName+".");
			}
				
			/*
			 * Open schema file.
			 */
			try
			{ 
				openedSchemaFile = XmlUtil.createDocument(fis, schemaFileName);
			}
			catch (final IOException ioe)
			{
				throw new IOException("Cannot open schema file "+schemaFileName+": "+ioe);
			}
				
			if (openedSchemaFile == null)
			{
				throw new IOException("Poor formated schema file: hypercast.xsd");
			}
		}

		return getPrivateProperties(openedSchemaFile);
	}

	/**
	 * Build the Document object for all attribute Elements (based on their default values) 
	 * in the given schema file.
	 */
	public static Document getPrivateProperties(String schemafile)
		throws FileNotFoundException, ParseException, IOException
	{
		Document schemaDoc = null;

		if (schemafile.toLowerCase().endsWith(".xsd"))
		{
			/*
			 * Open schema file.
			 */
			schemaDoc = XmlUtil.createDocument(schemafile);

			if (schemaDoc == null)
			{
				throw new IOException("Poor formated schema file: "+schemafile);
			}
		}
		else
		{		
			System.err.println("The file "+schemafile+" is not a schema file.");
		}
		
		return getPrivateProperties(schemaDoc);
	}
	
    /**
     * Build the Document object for all attribute Elements (based on their default values) 
     * in the given schema file (in Document form).
     */
    public static Document getPrivateProperties(Document schemaDoc)
    {
		Document doc = XmlUtil.createDocument();
		
		if (schemaDoc == null)
		{
			return doc;
		}
		
		/*
		 * Find the property Element in schema file. The way to find
		 * property Element is hypercast specific.
		 */
		Element propertyElement = createPrivatePropertyDocFromSchemaFile.findPropertyElement(schemaDoc, new String("Private"));

		if (propertyElement == null)
		{
			System.err.println("Poor formated schema file: " + schemaDoc.getNodeName());
			return null;						
		}

		if (debug)
			System.out.println("find private property Element in the given Document: "+propertyElement.getAttribute("name"));

		String rootNodeName = propertyElement.getAttribute("name");
		Element rootNode = doc.createElement(rootNodeName);
		doc.appendChild(rootNode);

		recursiveGetDefaultElement(doc, propertyElement, rootNode);

		return doc;
    }
    
	/**
	 * Recursively process all attribute Elements in the schema file and create new XML
	 * ELement for each os them.
	 * 
	 * @param doc	the Document to be created.
	 * @param xmlElement an Element in the schema DOM which represents an attribute.
	 * @param newXmlElement current Element in the created XMLDocument.
	 */
	private static void recursiveGetDefaultElement(Document doc, Element xmlElement, Element newXmlElement)
	{
		if (debug)
			System.out.println("Process Element: "+xmlElement.getAttribute("name"));

		/*
		 * Handle complexType Element.
		 */
		if (createPrivatePropertyDocFromSchemaFile.isComplexElement(xmlElement))
		{
			processComplexElement(doc, xmlElement, newXmlElement);
		}
		/*
		 * Otherwise it is simpleType Element
		 */
		else
		{
			processSimpleElement(doc, xmlElement, newXmlElement);
		}
	}


	/**
	 * Process the given complexType Element e and recursively process its children.
	 * 
	 * @param doc	the Document to be created.
	 * @param e an Element in the schema DOM which represents an attribute.
	 * @param newe current Element in the created XMLDocument.
	 */
	private static void processComplexElement(Document doc, Element e, Element newe)
	{
		String name = e.getAttribute("name");

		if (debug)
			System.out.println("Process complexType Element: "+name);
			
		/*
		 * Get all children of this Element in the schema file.
		 */
		Vector children = createPrivatePropertyDocFromSchemaFile.getChildrenOfComplexElement(e);
		
		/*
		 * Only when given compleType Element has choice on its children, the default
		 * value is created for it.
		 */
		if (createPrivatePropertyDocFromSchemaFile.hasChoiceOnChildren(e))
		{
			String defaultValue = e.getAttribute("default");

			Element defaultChild = null;
			
			if ((defaultValue == "") || (defaultValue == null))
			{
				defaultChild = (Element) children.elementAt(0);//the first child name because default value.
			}
			else
			{
				for (int i=0; i<children.size(); i++)
				{
					Element child = (Element) children.elementAt(i);
					String nodeName = child.getAttribute("name");
					
					if (nodeName.equals(defaultValue))
					{
						defaultChild = child;
						break;
					}
				}
			}

			if (defaultChild != null)
			{
				Element ele = doc.createElement(defaultChild.getAttribute("name"));
	
				if (debug)
					System.out.println("Entry is added to Document: "+ele.getNodeName());
	
				newe.appendChild(ele);

				recursiveGetDefaultElement(doc, defaultChild, ele);
			}
			else
			{
				System.out.println("The complexType Element "+name+" doesn't have child.");
			}

			if (debug)
				System.out.println("Entry is added to Document: "+defaultChild.getAttribute("name"));
		}
		else
		{

			/*
			 * This loop is to process this node's children recursively.
			 */
			for (int i=0; i<children.size(); i++)
			{
				Element child = (Element) children.elementAt(i);
					
				if (child == null)
				{
					continue;
				}
				
				String nodeName = child.getAttribute("name");
				Element ele = doc.createElement(child.getAttribute("name"));
				newe.appendChild(ele);
	
				if (debug)
					System.out.println("Loop on the children of "+name);
	
				recursiveGetDefaultElement(doc, child, ele);
			}
		}
	}

	/**
	 * Process the given simpleType Element by adding Text Node which contains the default value.
	 * 
	 * @param doc	the Document to be created.
	 * @param e an Element in the schema DOM which represents an attribute.
	 * @param newe current Element in the created XMLDocument.
	 */
	private static void processSimpleElement(Document doc, Element e, Element newe)
	{
		String name = e.getAttribute("name");

		if (debug)
			System.out.println("Process simpleType Element: "+name);

		if (createPrivatePropertyDocFromSchemaFile.getRestrictionElement(e) == null)
		{
			return;
		}

		String defaultValue = e.getAttribute("default");

		if (defaultValue != null)
		{
			if (debug)
				System.out.println("Loop on the children of "+e.getNodeName());

			Text textNode = doc.createTextNode(defaultValue);
			
			newe.appendChild(textNode);
		}
	}
 
	/**
	 * Find the property Element in the given Document. This Element is
	 * the starting point for building displaying tree.
	 * This method assume the schema file has fixed format to reach the 
	 * property Element.
	 * 
	 * @param doc	A Document instance.
	 * @param rootElementName	The Element name of the root for all attributes.
	 * 
	 * @return	the Element which has attribute "name = rootElemenName".
	 */
	public static Element findPropertyElement(Document doc, String rootElementName)
	{
		Element propertyElement = null;

		//get document's children
		NodeList nl = doc.getChildNodes();

		Element schemaElement = null;

		/*
		 * Find the Element whose name contains "schema".
		 */
		for (int i=0; i<nl.getLength(); i++)
		{
			Node node = nl.item(i);
			//System.out.println("node with name "+node.getNodeName());

			if (node.getNodeName().indexOf("schema") >= 0)
			{
				schemaElement = (Element) node;
				break;
			}
		}

		if (schemaElement == null)
		{
			//throw new IllegalArgumentException("Poorly structured schame file!");
			System.err.println("Poorly structured schame file!");
			return null;
		}

		//get schema Element's children
		nl = schemaElement.getChildNodes();

		for (int i=0; i<nl.getLength(); i++)
		{
			Node node = nl.item(i);
			//System.out.println("element with name "+node.getNodeName());
			
			/*
			 * There are some comments line in the file which become Node
			 * not Element. Skip all Nodes which are not instances of Element.
			 */
			if (!(node instanceof Element))
			{
				continue;
			}

			String attrValue =((Element) node).getAttribute("name");

			if ((attrValue != null) && (attrValue.equals(rootElementName)))
			{
				propertyElement = (Element) node;
				break;
			}
		}

		return propertyElement;
	}

	/**
	 * Check if this Element has a "complex" child.
	 */
	public static boolean isComplexElement(Element e)
	{
		NodeList firstnl = e.getChildNodes();

		for (int i=0; i<firstnl.getLength(); i++)
		{
			Node node = firstnl.item(i);
			String name = firstnl.item(i).getNodeName();
		
			if (name.indexOf("annotation") >= 0)
			{
				continue;
			}
			else if (name.indexOf("complexType") >= 0)
			{
				return true;
			}
			else
			{
				return false;
			}
		}
		return false;
	}

	/**
	 * Get children of this Element.
	 */
	public static Vector getChildrenOfComplexElement(Element e)
	{
		Vector p = new Vector();
		
		NodeList firstnl = e.getChildNodes();

		Node complexTypeChild = null;
		for (int i=0; i<firstnl.getLength(); i++)
		{
			Node node = firstnl.item(i);
			String name = firstnl.item(i).getNodeName();
				
			if (name.indexOf("annotation") >= 0)
			{
				continue;
			}
			else if (name.indexOf("complexType") >= 0)
			{
				complexTypeChild = node;
				break;
			}
		}

		if (complexTypeChild == null)
		{
			return p;
		}

		NodeList secnl = complexTypeChild.getChildNodes();
		NodeList thirdnl = null;

		if (secnl.getLength() == 0)
		{
			return p;
		}
		else if ((secnl.getLength() == 1) &&
				  ((secnl.item(0).getNodeName().equals("xs:sequence"))
					|| (secnl.item(0).getNodeName().equals("xs:choice"))))
		{
			thirdnl = secnl.item(0).getChildNodes();
		}
		else
		{
			thirdnl = secnl;
		}

		/*
		 * It is the Element with name "sequence" or "choice".
		 */
		for (int i=0; i<thirdnl.getLength(); i++)
		{
			Node node = thirdnl.item(i);

			/*
			 * There are some comments line in the file which become Node
			 * not Element. Skip all Nodes which are not instances of Element.
			 */
			if (!(node instanceof Element))
			{
				continue;
			}

			p.addElement((Element) node);
		}

		return p;
	}

	/**
	 * Check if this Element has a sequence of children.
	 */
	public static boolean hasChoiceOnChildren(Element e)
	{
		NodeList firstnl = e.getChildNodes();
		if (firstnl.getLength() != 1)
		{
			return false;
		}
		
		NodeList secnl = firstnl.item(0).getChildNodes();
		if (secnl.getLength() != 1)
		{
			return false;
		}
		else
		{
			String name = secnl.item(0).getNodeName();
			if (name.indexOf("choice") >= 0)
			{
				return true;
			}
			else
			{
				return false;
			}
		}
	}

	/**
	 * Get the Element which defines restrition for given simpleType Element.
	 * 
	 * @param e	Element in the configuration document.
	 * @return	the restrition definition Element.
	 */
	public static Element getRestrictionElement(Element e)
	{
		NodeList nl = e.getChildNodes();

		Node simpleTypeElement = null;

		/*
		 * Find a child which has name "simpleType".
		 */
		for (int i=0; i<nl.getLength(); i++)
		{
			Node node = nl.item(i);

			if (node.getNodeName().indexOf("simpleType") >= 0)
			{
				simpleTypeElement = node;
				break;
			}
		}
		
		/*
		 * simpleTypeElement's solo child is restrition Element.
		 */
		if ((simpleTypeElement == null) || (simpleTypeElement.getChildNodes().getLength() != 1))
		{
			return null;
		}

		Element restritionElement = (Element) simpleTypeElement.getChildNodes().item(0);
		if (restritionElement.getNodeName().indexOf("restriction") < 0)
		{
			return null;
		}
		else
		{
			return restritionElement;
		}
	}

}
